using System;
using System.Collections;
using BBN.RDF;
using BBN.RDF.Common;
using BBN.RDF.Vocabulary;
using BBN.DAML.Vocabulary;
using BBN.DAML;

namespace BBN.DAML
{


	/// <summary>
	/// 
	/// </summary>
	public abstract class DAMLCommonImpl : ResourceImpl, DAMLCommon
	{

		// Instance variables
		//////////////////////////////////

		/// <summary>
		/// Literal accessor for label property
		/// </summary>
		private LiteralAccessor m_propLabel = null;

		/// <summary>
		/// Literal accessor for comment property
		/// </summary>
		private LiteralAccessor m_propComment = null;

		/// <summary>
		/// Property accessor for equivalentTo
		/// </summary>
		private PropertyAccessor m_propEquivalentTo = null;

		/// <summary>
		/// The vocabulary that corresponds to the namespace this DAML value was declared in
		/// </summary>
		private DAMLVocabulary m_vocabulary = null;

		/// <summary>
		/// Property accessor for RDF:type
		/// </summary>
		private PropertyAccessor m_propType = null;

		
		// Constructors
		//////////////////////////////////

		/// <summary>
		/// Constructor, takes the URI this value, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="uri">The URI of the DAML value, or null for an anonymous node.</param>
		/// <param name="model">Reference to the DAML model that will contain statements about this DAML value.</param>
		/// <param name="vocabulary">The vocabulary for this value (defines the namespace).  Can be null, in which case the vocabulary defaults to the most recent.</param>
		public DAMLCommonImpl( string uri, DAMLModel model, DAMLVocabulary vocabulary ) : base(uri)
		{
			this.model = model;
			m_vocabulary = vocabulary;
			m_propLabel = new LiteralAccessorImpl( GetVocabulary().label, this );
			m_propComment = new LiteralAccessorImpl( GetVocabulary().comment, this );
		}

		/// <summary>
		/// Constructor, takes the name and namespace for this value, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="ns">The namespace the class inhabits, or null</param>
		/// <param name="name">The name of the DAML value</param>
		/// <param name="model">Reference to the DAML model that will contain statements about this DAML class.</param>
		/// <param name="vocabulary">The vocabulary for this value (defines the namespace).  Can be null, in which case the vocabulary defaults to the most recent.</param>
		public DAMLCommonImpl( string ns, string name, DAMLModel model, DAMLVocabulary vocabulary ) : base(ns + name)
		{
			this.model = model;
			m_vocabulary = vocabulary;
			m_propLabel = new LiteralAccessorImpl( GetVocabulary().label, this );
			m_propComment = new LiteralAccessorImpl( GetVocabulary().comment, this );
		}


		// External signature methods
		//////////////////////////////////

		/// <summary>
		/// Answer the underlying model
		/// </summary>
		/// <returns>A DAML model</returns>
		public virtual DAMLModel GetDAMLModel() 
		{
			return (DAMLModel) GetModel();
		}


		/// <summary>
		/// Set the RDF type property for this node in the underlying model, replacing any
		/// existing type property.  To add a second or subsequent type statement to a resource,
		/// use {@link #setRDFType( com.hp.hpl.mesa.rdf.jena.model.Resource, boolean )
		/// setRDFType( Resource, false ) }.
		/// </summary>
		/// <param name="rdfClass">The RDF resource denoting the new value for the rdf:type property,
		/// which will replace any existing type property.</param>
		public virtual void SetRDFType( Resource rdfClass ) 
		{
			SetRDFType( rdfClass, true );
		}


		/// <summary>
		/// Add an RDF type property for this node in the underlying model. If the replace flag
		/// is true, this type will replace any current type property for the node. Otherwise,
		/// the type will be in addition to any existing type property.  Note that for most normal
		/// uses, a DAML resource should have at most one rdf:type property.  One exception to this,
		/// in the March 2001 release, is when DatatypeProperties are marked as unique, unambiguous
		/// or transitive.  This is achieved by the use of two rdf type properties.
		/// </summary>
		/// <param name="rdfClass">The RDF resource denoting the class that will be new value for the rdf:type property</param>
		/// <param name="replace">If true, the given class will replace any existing type property for this
		/// value, otherwise it will be added as an extra type statement.</param>
		public virtual void SetRDFType( Resource rdfClass, bool replace ) 
		{
			// check first that we have a model - classes in the vocabulary don't have a model
			// but still invoke setRDFType when constructing
			if (GetModel() != null) 
			{
				if (replace) 
				{
					ReplaceProperty( BBN.RDF.Vocabulary.RDF.type, rdfClass );
				}
				else 
				{
					try 
					{
						AddProperty( BBN.RDF.Vocabulary.RDF.type, rdfClass );
					}
					catch (System.Exception e) 
					{
						//Log.severe( "RDF exception " + e, e );
						Console.Error.WriteLine( "RDF exception " + e, e );
						throw new System.Exception( "RDF Exception " + e );
					}
				}
			}
		}

		/**
		 * Answer true if this DAML value is a member of the class denoted by the given URI.
		 *
		 * @param classURI String denoting the URI of the class to test against
		 * @return true if it can be shown that this DAML value is a member of the class, via
		 *         <code>rdf:type</code>.
		 */
		public virtual bool HasRDFType( string classURI ) 
		{
			// do we know a daml class that matches this uri?
			DAMLCommon damlVal = ((DAMLModel) GetModel()).GetDAMLValue( classURI );

			if (damlVal != null  &&  damlVal is DAMLClass) 
			{
				// we know a class that matches the URI, so test against it
				return HasRDFType( (DAMLClass) damlVal );
			}
			else 
			{
				// last resort, make a temporary resource to hold the uri
				return HasRDFType( new ResourceImpl( classURI, GetModel() ) );
			}
		}


		/**
		 * Answer true if this DAML value is a member of the class denoted by the
		 * given DAML class object.  This will traverse the class hierarchy, until
		 * every class and super-class for this
		 * DAML value has been examined.  Depending on the depth of the hierarchy,
		 * this may be an expensive operation. Cycles are detected, however, so it
		 * is guaranteed to terminate.
		 *
		 * @param damlClass Denotes a class to which this value may belong
		 * @return true if the value is a member of the class (or one of its sub-classes)
		 *         via <code>rdf:type</code>.
		 */
		public virtual bool HasRDFType( Resource damlClass ) 
		{
			// see if the given class is in the iteration of this object's types
			bool found = false;
			string damlClassURI = damlClass.GetURI();

			for (IEnumerator i = GetRDFTypes( true );  !found  &&  i.MoveNext();  ) 
			{
				Resource c = (Resource) i.Current;

				// this class could be equal to the target class, or could be a sub-class of it under the
				// built-in knowledge of the DAML hierarchy
				found = c.Equals( damlClass ) ||
					DAMLHierarchy.GetInstance().IsDAMLSubClassOf( c.GetURI(), damlClassURI );
			}

			return found;
		}


		/**
		 * Answer an iterator over all of the types to which this class belongs. Optionally,
		 * generate a closure by considering the closure of the set of classes over the
		 * class hierarchy (e.g. if 'fido' is the resource, the non-closed set of fido's
		 * classes might be
		 * <code>{Dog, Vaccinated}</code>,
		 * i.e. the set of classes for which rdf:type statements exist for fido,
		 * while the closed set might be
		 * <code>{Dog, Vaccinated, Mammal, Pet, Vertebrate, Thing, MedicallyCertified}</code>
		 *
		 * @param closed If true, generate the closed set by considering the super-classes of
		 *               the known classes of this value.
		 * @return an iterator over the set of this value's classes (note: it is a set, so
		 *         each class will only appear once)
		 */
		public virtual IEnumerator GetRDFTypes( bool closed ) 
		{
			Hashtable types = new Hashtable();

			// first we get all of the values for rdf:type or daml:type
			for (IEnumerator i = Prop_Type.GetAll( false );  i.MoveNext(); ) 
			{
				// add each type to the type set
				Resource r = (Resource) i.Current;
				types[ r ] = true;

				// for the closed set of types, we also add the super-types of each type
				if (closed && r is DAMLClass) 
				{
					for (IEnumerator j = ((DAMLClass) r).GetSuperClasses();  j.MoveNext(); ) 
					{
						types[ j.Current ] = true;
					}
				}
			}

			// check in case we have a default type to add
			if (closed && GetDefaultType() != null) 
			{
				types[ GetDefaultType() ] = true;
			}

			return types.Keys.GetEnumerator();
		}


		/**
		 * Answer the value of a given RDF property for this DAML value, or null
		 * if it doesn't have one.  The value is returned as an RDFNode, from which
		 * the value can be extracted for literals.  If there is more than one RDF
		 * statement with the given property for the current value, it is not defined
		 * which of the values will be returned.
		 *
		 * @param property An RDF property
		 * @return An RDFNode whose value is the value, or one of the values, of the
		 *         given property. If the property is not defined, or an error occurs,
		 *         returns null.
		 */
		public virtual RDFNode GetPropertyValue( Property property ) 
		{
			try 
			{
				return GetProperty( property ).GetObject();
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception while getting property " + property + " value was: " + e);
				throw new System.Exception("RDF error when getting values for property " + property + ": " + e );
			}
		}




		/**
		 * Answer an iterator over the set of all values for a given RDF property. Each
		 * value in the iterator will be an RDFNode, representing the value (object) of
		 * each statement in the underlying model.
		 *
		 * @param property The property whose values are sought
		 * @return An Iterator over the values of the property, each of which will be an
		 *         {@link com.hp.hpl.mesa.rdf.jena.model.RDFNode RDFNode}.
		 */
		public virtual NodeEnumerator GetPropertyValues( Property property ) 
		{
			return new NodeEnumeratorImpl( new PropertyEnumerator( this, property, null, false, false, true) );
		}


		/**
		 * Set the value of the given property of this DAML value to the given
		 * value, encoded as an RDFNode.  Maintains the invariant that there is
		 * at most one value of the property for a given DAML object, so existing
		 * property values are first removed.  To add multiple properties to a
		 * given DAML object, use
		 * {@link com.hp.hpl.mesa.rdf.jena.model.Resource#addProperty( com.hp.hpl.mesa.rdf.jena.model.Property, com.hp.hpl.mesa.rdf.jena.model.RDFNode ) addProperty}.
		 *
		 * @param property The property to update
		 * @param value The new value of the property as an RDFNode, or null to
		 *              effectively remove this property.
		 */
		public virtual void SetPropertyValue( Property property, RDFNode value ) 
		{
			if (value == null) 
			{
				RemoveAll( property );
			}
			else 
			{
				ReplaceProperty( property, value );
			}
		}


		/**
		 * Remove the specific RDF property-value pair from this DAML resource.
		 *
		 * @param property The property to be removed
		 * @param value The specific value of the property to be removed
		 */
		public virtual void RemoveProperty( Property property, RDFNode value ) 
		{
			RemoveProperties( GetEquivalenceClass( property ), GetEquivalenceClass( value ) );
		}

		/// <summary>
		/// Remove all the values for a given property on this DAML resource.
		/// </summary>
		/// <param name="prop">The RDF resource that defines the property to be removed</param>
		public virtual void RemoveAll( Property prop ) 
		{
			RemoveAll( GetEquivalenceClass( prop ) );
		}



		/// <summary>
		/// Replace the value of the named property with the given value.  All existing
		/// values, if any, for the property are first removed.
		/// </summary>
		/// <param name="prop">The RDF property to be updated</param>
		/// <param name="value">The new value.</param>
		public virtual void ReplaceProperty( Property prop, RDFNode value ) 
		{
			try 
			{
				// if there is an existing property, remove it
				RemoveAll( prop );

				// now set the new value
				AddProperty( prop, value );
			}
				// RDFException
			catch (System.Exception e) 
			{
				//Log.severe( "RDF exception while replacing value of DAML property: " + e, e );
				Console.Error.WriteLine( "RDF exception while replacing value of DAML property: " + e);
				throw new System.Exception( "RDF exception while replacing value of DAML property: " + e );
			}
		}


		/**
		 * Answer the number of values a given property has with this value as subject.
		 *
		 * @param property The property to be tested
		 * @return The number of statements with this value as subject and the given
		 *         property as relation.
		 */
		public virtual int GetNumPropertyValues( Property property ) 
		{
			try 
			{
				int count = 0;

				// lookup the super-classes from this model
				for (NodeEnumerator i = GetPropertyValues( property );  i.MoveNext(); ) 
				{
					count++;
				}

				return count;
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "Exception while listing values: " + e);
				throw new System.Exception( "RDF failure while listing values: " + e );
			}
		}

		/**
		 * Answer an iterator over a set of resources that are the objects of statements
		 * with subject this DAML object and predicate the given property. Respects DAML
		 * semantics of equivalence, transitivity and the property hierarchy.
		 *
		 * @param property The property whose values are sought
		 * @param closed If true, and the given property is transitive, generate the
		 *               closure over the given property from this value.
		 * @return An iterator of resources that are the objects of statements whose
		 *         subject is this value (or one of its equivalents) and whose predicate
		 *         is <code>property</code> or one of its equivalents
		 */
		public virtual IEnumerator GetAll( Property property, bool closed ) 
		{
			// get an iterator over all of the values of this property, including equivalent values
			PropertyEnumerator iter = new PropertyEnumerator( GetEquivalentValues(), property,
				GetPropertyInverse( property ), (closed && IsTransitive( property )), false, true );
			return iter;
		}


		/// <summary>
		/// Answer the DAML+OIL vocabulary that corresponds to the namespace that this value
		/// was declared in.
		/// </summary>
		/// <returns>a vocabulary object</returns>
		public virtual DAMLVocabulary GetVocabulary() 
		{
			if (m_vocabulary == null) 
			{
				// need to establish the vocabulary for this object
				// TBD: Implement or bypass a Vocabulary Manager.
				//m_vocabulary = VocabularyManager.GetVocabulary( this );
			}

			return m_vocabulary;
		}


		/**
		 * Answer a key that can be used to index collections of this DAML value for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public abstract Object GetKey();


		/// <summary>
		/// Answer an iterator over all of the DAML objects that are equivalent to this
		/// value under the <code>daml:equivalentTo</code> relation.  The common method
		/// just tests this relation, <code>GetEquivalentValues()</code> in sub-classes,
		/// such as {@link com.hp.hpl.jena.daml.DAMLClass}, may extend this with specific
		/// additional semantics such as <code>daml:sameClassAs</code>.
		/// </summary>
		/// <returns>an iterator ranging over every equivalent DAML value - each value of
		/// the iteration will be a damlCommon object</returns>
		public virtual IEnumerator GetEquivalentValues() 
		{
			// equivalentTo is transitive, and is its own inverse
			// we also need to exploit background knowledge about the equivalence of DAML relations
			return new ConcatenatedEnumerator(
				new PropertyEnumerator( this, GetVocabulary().equivalentTo, GetVocabulary().equivalentTo, true, true, false ),
				DAMLHierarchy.GetInstance().GetEquivalentValues( this ) );
		}

		/**
		 * Answer the set of equivalent values to this value, but not including the
		 * value itself.  The iterator will range over a set: each element occurs only
		 * once.
		 *
		 * @return An iteration ranging over the set of values that are equivalent to this
		 *         value, but not itself.
		 */
		public virtual IEnumerator GetEquivalenceSet() 
		{
			Hashtable s = new Hashtable();

			// eliminate duplicates from the equivalent values iteration
			for (IEnumerator i = GetEquivalentValues();  i.MoveNext();  ) 
			{
				Object o = i.Current;

				if (!Equals( o )) 
				{
					s[ o ] = true;
				}
			}

			return s.Keys.GetEnumerator();
		}


		/**
		 * Answer an iterator that contains exactly this value.  This is useful as
		 * we often use Iterators to stand for sets or collections, and two such iterators
		 * can be appended to form the union of the collections.
		 *
		 * @return an iterator whose sole value is this DAML value
		 */
		public virtual IEnumerator GetSelfEnumerator() 
		{
			ArrayList l = new ArrayList();
			l.Add( this );
			return l.GetEnumerator();
		}


		/**
		 * Return a readable representation of the DAML value
		 *
		 * @return a string denoting this value
		 */
		public virtual string toString() 
		{
			// get the public name for this value type (e.g. we change DAMLClassImpl -> DAMLClass)
			string cName = GetType().FullName;
			int i = cName.IndexOf( "Impl" );
			int j = cName.LastIndexOf( "." ) + 1;
			cName = (i > 0) ? cName.Substring( j, i ) : cName.Substring( j );

			// now format the return string
			return (IsAnon()) ?
				("<Anonymous " + GetIDTranslation( GetId() ) + " " + cName + "@" + GetHashCode().ToString("X") + ">") :
				("<" + cName + " " + GetURI() + ">");
		}


		/**
		 * Remove the DAML object from the model.  All of the RDF statements with this
		 * DAML value as its subject will be removed from the model, and this object will
		 * be removed from the indexes.  It will be the responsibility of client code to
		 * ensure that references to this object are removed so that the object itself
		 * can be garbage collected.
		 */
		public virtual void Remove() 
		{
			try 
			{
				// first remove all of the statements corresponding to this object
				for (StmtEnumerator i = ListProperties();  i.MoveNext();  i.Current.Remove() );

				// now remove this object from the DAML indexes
				((DAMLModelImpl) GetModel()).Unindex( this );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception while removing object from model: " + e);
			}
		}


		/**
		 * Port this DAML value to another model.  The intent of this operation is to
		 * create a new value that has its getModel() pointing to the target model.
		 *
		 * @param m The model to port to
		 * @return The ported value, as a resource
		 */
		public override Resource Port( Model m )
		{
			if (GetModel() == m) 
			{
				return this;
			}
			else if (GetModel() != null  && m is DAMLModel) 
			{
				// port a DAML value to another DAML model
				try 
				{
					// since we don't know which class is being constructed, use reflection
					// Get the constructor for the current class that has these parameter types...
					//Constructor con = getClass().getConstructor( new Class[] {String.class, DAMLModel.class, DAMLVocabulary.class} );
					// Create a new instance using the constructor and pass in the arguments...
					//return (Resource) con.newInstance( new Object[] {getURI(), (DAMLModel) m, getVocabulary()} );
					object[] parameters = {GetURI(), (DAMLModel)m, GetVocabulary()};
					Resource resource = (Resource)Activator.CreateInstance(GetType(), parameters );
					return resource;
				}
				catch (Exception e) 
				{
					Console.WriteLine( "Failed to construct new instance of DAML value during port: " + e );
					throw e;
				}
			}
			else 
			{
				// rely on the default implementation
				return base.Port( m );
			}
		}



		// Properties
		/////////////

		/**
		 * Accessor for the property of the label on the value, whose value
		 * is a literal (string).
		 *
		 * @return Literal accessor for the label property
		 */
		public virtual LiteralAccessor Prop_Label 
		{
			get
			{
				return m_propLabel;
			}
		}


		/**
		 * Accessor for the property of the comment on the value, whose value
		 * is a literal (string).
		 *
		 * @return Literal accessor for the comment property
		 */
		public virtual LiteralAccessor Prop_Comment 
		{
			get 
			{
				return m_propComment;
			}
		}


		/**
		 * Property accessor for the 'equivalentTo' property of a DAML value. This
		 * denotes that two terms have the same meaning. The spec helpfully
		 * says: <i>for equivalentTo(X, Y), read X is an equivalent term to Y</i>.
		 *
		 * @return Property accessor for 'equivalentTo'.
		 */
		public virtual PropertyAccessor Prop_EquivalentTo 
		{
			get
			{
				if (m_propEquivalentTo == null) 
				{
					m_propEquivalentTo = new PropertyAccessorImpl( GetVocabulary().equivalentTo, this );
				}

				return m_propEquivalentTo;
			}
		}



		/**
		 * Property accessor for the 'rdf:type' property of a DAML value.
		 *
		 * @return Property accessor for 'rdf:type'.
		 */
		public virtual PropertyAccessor Prop_Type 
		{
			get
			{
				if (m_propType == null) 
				{
					m_propType = new PropertyAccessorImpl( BBN.RDF.Vocabulary.RDF.type, this );
				}

				return m_propType;
			}
		}



		// Internal implementation methods
		//////////////////////////////////

		/**
		 * For debugging, a more concise rendering of the UID
		 */
		//protected static java.util.HashMap s_UIDs = new java.util.HashMap();
		protected static Hashtable s_UIDs = new Hashtable();
		protected static int s_count = 0;
		protected static string GetIDTranslation( AnonId uid ) 
		{
			string id = (string) s_UIDs[ uid ];
			if (id == null) 
			{
				id =  s_count++.ToString() ;
				s_UIDs[ uid ] = id;
			}
			return id;
		}

		/// <summary>
		/// Answer the equivalence class, as an iterator, for a given resource
		/// </summary>
		/// <param name="n">The resource</param>
		/// <returns>An iteration of the resource and its equivalents</returns>
		protected IEnumerator GetEquivalenceClass( RDFNode n ) 
		{
			if (n is DAMLCommon) 
			{
				// if it's a DAML object, we can calculate the equivalence classs
				return ((DAMLCommon) n).GetEquivalentValues();
			}
			else 
			{
				// otherwise return the singleton set
				ArrayList single = new ArrayList();
				single.Add( n );
				return single.GetEnumerator();
			}
		}


		/**
		 * Remove the properties defined by the given equivalence classes, if they exist.
		 * Specifically, remove every statement for which the the subject is this DAML value,
		 * the predicate is one of the given properties, and the value is one of the given values.
		 *
		 * @param preds An iterator over a set of properties
		 * @param values A set of values to match against
		 */
		protected void RemoveProperties( IEnumerator preds, IEnumerator values ) 
		{
			// we have to clone one of the sets in order to do the cross-product
			ArrayList predSet = new ArrayList();
			while (preds.MoveNext()) 
			{
				predSet.Add( preds.Current );
			}

			// now, go through every element of the cross-product, and remove that statement
			while (values.MoveNext()) 
			{
				RDFNode n = (RDFNode) values.Current;

				for (IEnumerator i = predSet.GetEnumerator();  i.MoveNext(); ) 
				{
					Property p = (Property) i.Current;

					try 
					{
						for (StmtEnumerator j = GetModel().ListStatements( new SelectorImpl( this, p, n ) );  j.MoveNext(); ) 
						{
							j.Current.Remove();
						}
					}
					catch (System.Exception e) 
					{
						Console.Error.WriteLine( "Possible RDF error when zapping from model: " + e);
					}
				}
			}
		}

		/// <summary>
		/// Remove all statements whose predicate is in the given iteration, and whose
		/// subject is this resource.
		/// </summary>
		/// <param name="preds">An iterator over a set of properties</param>
		protected void RemoveAll( IEnumerator preds ) 
		{
			try 
			{
				while (preds.MoveNext()) 
				{
					Property p = (Property) preds.Current;

					for (StmtEnumerator i = GetModel().ListStatements( new SelectorImpl( this, p, (RDFNode) null ) );  i.MoveNext(); ) 
					{
						i.Current.Remove();
					}
				}
			}
				// RDFException
			catch (System.Exception e) 
			{
				//Log.severe( "Possible RDF error when zapping from model: " + e, e );
				Console.Error.WriteLine( "Possible RDF error when zapping from model: " + e);
			}
		}




		/**
		 * Answer the inverse of a property, if we know what it is
		 *
		 * @param property An RDF or DAML property
		 * @return The inverse property of property, or null
		 */
		protected Property GetPropertyInverse( Property property ) 
		{
			// can only check the inverses of DAML object properties
			if (property is DAMLObjectProperty) 
			{
				// Note: we don't use the prop_inverseOf property accessor here, since that
				// would cause infinite recursion, as the property accessor needs to know what
				// the property inverse is

				DAMLProperty dProperty = (DAMLProperty) property;

				// lookup the inverse of this property, if it has one
				try 
				{
					Statement inv = dProperty.GetProperty( dProperty.GetVocabulary().inverseOf );
					return (Property) inv.GetObject();
				}
				catch (System.Exception) 
				{
					// can ignore - just means not present
				}
			}

			return null;
		}


		/**
		 * Answer true if the given property is transitive
		 *
		 * @param property A property definition
		 * @return True if the property is defined to be transitive in the ontology,
		 *         or is known to be a transitive DAML property.
		 */
		protected bool IsTransitive( Property property ) 
		{
			return DAMLHierarchy.GetInstance().IsTransitiveProperty( property )  ||
				((property is DAMLObjectProperty)  &&
				((DAMLObjectProperty) property).Transitive);
		}


		/**
		 * Answer a value that will be a default type to include in an iteration of
		 * the value's rdf types.  Typically there is no default (null), but for an
		 * instance we want to ensure that the default type is daml:Thing.
		 *
		 * @return The default type or null
		 */
		protected virtual Resource GetDefaultType() 
		{
			return null;
		}


	}
}
