using System;
using System.Collections;
using BBN.DAML.Vocabulary;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// A DAML data instance is an instance of a DAML Dataype (compared to an DAML instance which is
	/// an instance of a DAML Class; Class and Datatype are disjoint).
	/// </summary>
	public class DAMLDataInstanceImpl : DAMLCommonImpl, DAMLDataInstance
	{

		// Instance variables
		//////////////////////////////////

		/** Property accessor for sameIndividualAs */
		protected PropertyAccessor m_propsameIndividualAs = null;



		// Constructors
		//////////////////////////////////

		/**
		 * Constructor, takes the name and namespace for this instance, and the underlying
		 * model it will be attached to.  Note that it is assumed that the RDF store
		 * will contain a statement of the class to which this instance belongs.
		 *
		 * @param namespace The namespace the instance inhabits, or null
		 * @param name The name of the instance
		 * @param store The RDF store that contains the RDF statements defining the properties of the instance
		 * @param vocabulary Reference to the DAML vocabulary used by this instance.
		 */
		public DAMLDataInstanceImpl( string ns, string name, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( ns, name, store, vocabulary )
		{
		}


		/**
		 * Constructor, takes the URI for this instance, and the underlying
		 * model it will be attached to.  Note that it is assumed that the RDF store
		 * will contain a statement of the class to which this instance belongs.
		 *
		 * @param uri The URI of the instance
		 * @param store The RDF store that contains the RDF statements defining the properties of the instance
		 * @param vocabulary Reference to the DAML vocabulary used by this instance.
		 */
		public DAMLDataInstanceImpl( string uri, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( uri, store, vocabulary )
		{
		}




		// External signature methods
		//////////////////////////////////


		/**
		 * Answer a key that can be used to index collections of this DAML instance for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public override Object GetKey() 
		{
			return DAML_OIL._Thing.GetURI();
		}


		/**
		 * Property accessor for <code>daml:sameIndividualAs</code> property on a DAML instance.
		 *
		 * @return a property accessor
		 */
		public PropertyAccessor Prop_SameIndividualAs
		{
			get
			{
				if (m_propsameIndividualAs == null) 
				{
					m_propsameIndividualAs = new PropertyAccessorImpl( GetVocabulary().sameIndividualAs, this );
				}

				return m_propsameIndividualAs;
			}
		}


		/**
		 * Return an iterator over all of the instances that are the same as this one,
		 * by generating the transitive closure over the <code>daml:samePropertyAs</code>
		 * property.
		 *
		 * @return an iterator whose values will all be DAMLInstance objects
		 */
		public IEnumerator GetSameInstances() 
		{
			return new PropertyEnumerator( this, GetVocabulary().sameIndividualAs,
				GetVocabulary().sameIndividualAs, true, true, true );
		}


		/**
		 * Answer the Datatype translator for values encoded by the datatype of this
		 * instance.
		 *
		 * @return the datatype translator defined by the DAMLDatatype that is the rdf:type
		 *         of this instance, or null if it is not defined.
		 */
		public DatatypeTranslator GetTranslator() 
		{
			// get the RDF type of this value; should be only 1
			IEnumerator i = GetRDFTypes( false );

			if (i.MoveNext()) 
			{
				// get the type resource
				Resource r = (Resource) i.Current;

				// should be a datatype
				if (r is DAMLDatatype) 
				{
					if (i.MoveNext()) 
					{
						Console.Error.WriteLine( "DAMLInstance " + this + " has more than one rdf:type, only returning first one" );
					}

					return (DatatypeTranslator) ((DAMLDatatype) r).GetTranslator();
				}
				else 
				{
					Console.Error.WriteLine( "DAMLInstance has rdf:type that is not a DAMLDatatype" );
				}
			}

			return null;
		}


		/**
		 * Answer the value of this instance as a Java object, translated from the
		 * serialised RDF representation by the Dataype's translator.
		 *
		 * @return the value of this instance, or null if either the translator or the
		 *         serialised value is defined
		 */
		/**
		 * Set the value of this instance to the given Java value, which will be
		 * serialised into the RDF graph by the datatype's translator.
		 *
		 * @param value A Java value whose serialisation will be made the value of this
		 *              data instance by setting the <code>rdf:value</code> relation.
		 */
		public Object Value 
		{
			get 
			{
				try 
				{
					return GetTranslator().Deserialize( GetProperty( BBN.RDF.Vocabulary.RDF.Value ).GetObject() );
				}
				catch (Exception e) 
				{
					Console.Error.WriteLine( "RDF exception while setting value of data instance: " + e);
					return null;
				}
			}

			set
			{
				ReplaceProperty( BBN.RDF.Vocabulary.RDF.Value, GetTranslator().Serialize( value, GetDAMLModel() ) );
			}


		}
	}
}
