using System;
using System.Collections;
using BBN.RDF.Vocabulary;
using BBN.DAML.Vocabulary;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class DAMLHierarchy
	{
		// Static variables
		//////////////////////////////////

		/** The singleton instance */
		private static DAMLHierarchy s_instance = new DAMLHierarchy();

		// Instance variables
		//////////////////////////////////

		/**
		 * Map of DAML and RDF standard classes that are super-classes of other DAML standard classes.
		 * Assumes each class has at most one super-class.
		 */
		public Resource[,] DAML_STANDARD_CLASS_HIERARCHY = new Resource[,] {
																				 // sub-class                           super-class
		{DAML_OIL._Class,                       RDFS.Class},
		{DAML_OIL._Datatype,                    RDFS.Class},
		{DAML_OIL._Restriction,                 DAML_OIL._Class},
		{DAML_OIL._ObjectProperty,              BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL._DatatypeProperty,            BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL._TransitiveProperty,          DAML_OIL._ObjectProperty},
		{DAML_OIL._UnambiguousProperty,         DAML_OIL._ObjectProperty},
		{DAML_OIL._UniqueProperty,              BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL._List,                        BBN.RDF.Vocabulary.RDF.Seq},

		{DAML_OIL_2000_12._Disjoint,            DAML_OIL_2000_12._List},
		{DAML_OIL_2000_12._TransitiveProperty,  BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL_2000_12._UniqueProperty,      BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL_2000_12._UnambiguousProperty, BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL_2000_12._List,                BBN.RDF.Vocabulary.RDF.Seq}
																			 };


		/**
		 * Known equivalences between DAML and RDFS values
		 */
		public Resource[,] DAML_STANDARD_EQUIVALENCES = new Resource[,] {
		{DAML_OIL._subClassOf,                   RDFS.subClassOf},
		{DAML_OIL._Literal,                      RDFS.Literal},
		{DAML_OIL._Property,                     BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL._type,                         BBN.RDF.Vocabulary.RDF.type},
		{DAML_OIL._value,                        BBN.RDF.Vocabulary.RDF.Value},
		{DAML_OIL._subPropertyOf,                RDFS.subPropertyOf},
		{DAML_OIL._domain,                       RDFS.domain},
		{DAML_OIL._range,                        RDFS.range},
		{DAML_OIL._label,                        RDFS.label},
		{DAML_OIL._comment,                      RDFS.comment},
		{DAML_OIL._seeAlso,                      RDFS.seeAlso},
		{DAML_OIL._isDefinedBy,                  RDFS.isDefinedBy},

		{DAML_OIL_2000_12._subPropertyOf,        RDFS.subPropertyOf},
		{DAML_OIL_2000_12._Class,                RDFS.Class},
		{DAML_OIL_2000_12._Literal,              RDFS.Literal},
		{DAML_OIL_2000_12._Property,             BBN.RDF.Vocabulary.RDF.Property},
		{DAML_OIL_2000_12._type,                 BBN.RDF.Vocabulary.RDF.type},
		{DAML_OIL_2000_12._value,                BBN.RDF.Vocabulary.RDF.Value},
		{DAML_OIL_2000_12._subClassOf,           RDFS.subClassOf},
		{DAML_OIL_2000_12._domain,               RDFS.domain},
		{DAML_OIL_2000_12._range,                RDFS.range},
		{DAML_OIL_2000_12._label,                RDFS.label},
		{DAML_OIL_2000_12._comment,              RDFS.comment},
		{DAML_OIL_2000_12._seeAlso,              RDFS.seeAlso},
		{DAML_OIL_2000_12._isDefinedBy,          RDFS.isDefinedBy}
																		  };


		/** A list of well-known properties that are known a priori to be transitive */
		public Property[] TRANSITIVE_PROPERTIES = 
			new Property[] {
							   DAML_OIL._subClassOf,
							   DAML_OIL._subPropertyOf,
							   DAML_OIL._sameClassAs,
							   DAML_OIL._sameIndividualAs,
							   DAML_OIL._samePropertyAs,
							   DAML_OIL._equivalentTo,

							   DAML_OIL_2000_12._subClassOf,
							   DAML_OIL_2000_12._subPropertyOf,
							   DAML_OIL_2000_12._sameClassAs,
							   DAML_OIL_2000_12._sameIndividualAs,
							   DAML_OIL_2000_12._samePropertyAs,
							   DAML_OIL_2000_12._equivalentTo,

							   RDFS.subClassOf,
							   RDFS.subPropertyOf
						   };

		/** Super-class hierarchy hashmap */
		protected Hashtable m_classHierarchyMap = new Hashtable();

		/** Flag to show initialisation of static structures has taken place */
		protected bool m_initialised = false;

		/** Equivalance map */
		protected Hashtable m_equivalenceMap = new Hashtable();

		/** Table of transitive properties */
		protected Hashtable m_transitiveProperties = new Hashtable();



		// Constructors
		//////////////////////////////////

		/**
		 * Constructor is private to enforce singleton design pattern.
		 */
		private DAMLHierarchy() 
		{
		}


		// External signature methods
		//////////////////////////////////

		/**
		 * Answer the singleton instance.
		 *
		 * @return the unique DAMLHierarchy
		 */
		public static DAMLHierarchy GetInstance() 
		{
			return s_instance;
		}

		
		/**
		 * Answer true if the first URI represents a class in the standard DAML ontology
		 * that is a super-class of the class denoted by the second URI.  This is used to
		 * provide background knowledge from the standard DAML ontology without having
		 * to explicitly import it each time.
		 *
		 * @param uri0 A class URI
		 * @param uri1 A class URI
		 * @return true if the class denoted by uri0 is a sub-class of the class denoted by uri1
		 */
		public bool IsDAMLSubClassOf( string uri0, string uri1 ) 
		{
			// only works for named classes, so if the uri's are null or empty (indicating
			// anonymous resources) we skip
			if ((uri0 == null)  ||  (uri1 == null)  ||
				(uri0.Length == 0)  ||  (uri1.Length == 0)) 
			{
				return false;
			}

			// ensure that the structure has been initialised
			InitialiseMaps();

			// first, is uri0 directly a sub-class of uri1?
			string superClass = (string) m_classHierarchyMap[ uri0 ];

			if (uri1.Equals( superClass )) 
			{
				// found a match
				return true;
			}
			else 
			{
				// NOTE: assumes there is at most one super-class for each class, which is
				// true in current releases of DAML but may not remain true forever.
				return superClass == null ? false : IsDAMLSubClassOf( superClass, uri1 );
			}
		}
		

		/**
		 * Answer true if the given property is well-known to be transitive. That is,
		 * it is transitive even though it is not marked as being in the class of
		 * transitive properties defined by the TransitiveProperty class.  Note: this
		 * test does not check for <code>rdf:type TransitiveProperty</code>, it <i>only</i>
		 * tests the table of well-known transitive properties.
		 *
		 * @param res A resource
		 * @return True if the resource is a Property, and it is a member of the set
		 *         of well-known transitive properties.
		 */
		public bool IsTransitiveProperty( Resource res ) 
		{
			return (res is Property  &&  m_transitiveProperties.ContainsKey( res ));
		}
		

		/**
		 * Answer an iteration of those values that are equivalent to the given resource.
		 *
		 * @param res A resource, that may have known equivalent values
		 * @return an iteration over the resource's equivalent values, which may be empty
		 *         but will not be null.
		 */
		public IEnumerator GetEquivalentValues( Resource res ) 
		{
			// ensure that the cache structure has been initialised
			InitialiseMaps();

			ArrayList equivs = (ArrayList) m_equivalenceMap[res];
			return (equivs == null) ? new ArrayList().GetEnumerator() : equivs.GetEnumerator();
		}



		// Internal implementation methods
		//////////////////////////////////

		/**
		 * Initialise the static structures we need to process class hierarchy queries.
		 * Note that we don't do this as a static to avoid race conditions with setting
		 * up the vocabularies.
		 */
		private void InitialiseMaps() 
		{
			if (!m_initialised) 
			{
				// initialise the class hierachy map for known standard DAML classes
				for (int i = 0;  i < DAML_STANDARD_CLASS_HIERARCHY.GetLength(0);  i++) 
				{
					string subClass   = DAML_STANDARD_CLASS_HIERARCHY[i,0].GetURI();
					string superClass = DAML_STANDARD_CLASS_HIERARCHY[i,1].GetURI();

					m_classHierarchyMap[subClass]= superClass ;
				}

				// initialise the equivalence map - note symmetric relation, so we have to do both directions
				for (int i = 0;  i < DAML_STANDARD_EQUIVALENCES.GetLength(0);  i++) 
				{
					// first do the forwards direction
					ArrayList l = (ArrayList) m_equivalenceMap[ DAML_STANDARD_EQUIVALENCES[i,0] ];

					if (l == null) 
					{
						// if no list of equivs yet, create one
						l = new ArrayList();
						m_equivalenceMap[ DAML_STANDARD_EQUIVALENCES[i,0] ] = l ;
					}

					l.Add( DAML_STANDARD_EQUIVALENCES[i,1] );

					// now do the reverse direction
					l = (ArrayList) m_equivalenceMap[ DAML_STANDARD_EQUIVALENCES[i,1] ];

					if (l == null) 
					{
						// if no list of equivs yet, create one
						l = new ArrayList();
						m_equivalenceMap[ DAML_STANDARD_EQUIVALENCES[i,1]] = l;
					}

					l.Add( DAML_STANDARD_EQUIVALENCES[i,0] );
				}

				// initialise the transitive properties map, for efficiency
				for (int i = 0;  i < TRANSITIVE_PROPERTIES.GetLength(0);  i++) 
				{
					m_transitiveProperties[ TRANSITIVE_PROPERTIES[i]] = true ;
				}

				m_initialised = true;
			}
		}
		


	}
}
