using System;
using System.Collections;
using BBN.DAML.Vocabulary;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class DAMLListImpl : DAMLCommonImpl, DAMLList
	{
		// Constructors
		//////////////////////////////////


		/// <summary>
		/// Constructor, takes the URI for this list, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="uri">The URI of the list</param>
		/// <param name="store">The RDF store that contains the RDF statements defining the properties of the list</param>
		/// <param name="vocabulary">Reference to the DAML vocabulary used by this list</param>
		public DAMLListImpl( string uri, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( uri, store, vocabulary  )	
		{
			SetRDFType( GetVocabulary().List );
		}

		/// <summary>
		/// Constructor, takes the name and namespace for this list, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="ns">The namespace the list inhabits, or null</param>
		/// <param name="name">The name of the list</param>
		/// <param name="store">The RDF store that contains the RDF statements defining the properties of the list</param>
		/// <param name="vocabulary">Reference to the DAML vocabulary used by this list.</param>
		public DAMLListImpl( string ns, string name, DAMLModel store, DAMLVocabulary vocabulary  ) 
			: base( ns, name, store, vocabulary  )
		{
			SetRDFType( GetVocabulary().List );
		}


		// External signature methods
		//////////////////////////////////


		/**
		 * Add an element to the list.  The element will in fact be added to the end
		 * of the list, but DAML collections are, strictly speaking, unordered, so the
		 * position of the element in the list should not be relied upon.  This is a
		 * side-effectful operation.  See {@link #cons}
		 * for an operation that always returns a new list (i.e. no side-effects on
		 * this list).
		 *
		 * @param value A DAML value to add to the list
		 */
		public void Add( DAMLCommon value ) 
		{
			if (IsEmpty()) 
			{
				SetFirst( value );
				SetRestNil();
			}
			else 
			{
				// find the list element that contains the empty list
				DAMLList last = FindLast();

				DAMLList newList = (DAMLList) GetDAMLModel().CreateDAMLValue( null, GetVocabulary().List, GetVocabulary() );

				newList.SetFirst( value );
				newList.SetRestNil();

				// extend the list with this new element
				last.SetRest( newList );
			}
		}


		/**
		 * Remove the given value from the list.  If the value is not in the list, has
		 * no effect.  If the value is in the list multiple times, only one of them
		 * will be removed.
		 *
		 * @param value A DAML value to be removed from the list.
		 */
		public void Remove( DAMLCommon value ) 
		{
			DAMLList parent = null;
			DAMLList child = this;

			// traverse the list to find the value
			while (!child.IsEmpty()  &&  !child.GetFirst().Equals( value )) 
			{
				parent = child;
				child = child.GetRest();
			}

			// check to see if we found the value (if not, ignore it)
			if (!child.IsEmpty()) 
			{
				// we did find a value
				if (parent == null) 
				{
					// the value we are removing is at the head of the list
					// removing first and last properties make the parent an empty list
					// note: ordering of additions/removals is significant
					DAMLList cell = GetRest();

					if (cell.IsEmpty()) 
					{
						// if the tail is null, we make this cell null
						RemoveAll( GetVocabulary().first );
						RemoveAll( GetVocabulary().rest );
					}
					else 
					{
						// otherwise, link to the tail
						ReplaceProperty( GetVocabulary().first, cell.GetFirst() );
						ReplaceProperty( GetVocabulary().rest, cell.GetRest() );
						cell.Remove();
					}
				}
				else 
				{
					// change the parent to omit the child
					parent.SetRest( child.GetRest() );
					child.Remove();
				}
			}
		}


		/**
		 * Answer an iteration over the values in the list.
		 *
		 * @return An iterator over the DAML values in the list
		 */
		public IEnumerator GetAll() 
		{
			return new DAMLListEnumerator( this );
		}


		/**
		 * Answer the first value from the list. Since, strictly speaking, DAML collections
		 * are unordered, the position items in the list should not be relied on in client
		 * code, as the definition of 'first' in the list may change in future releases.
		 * However, the identity
		 * <code><pre>
		 *      List L = L.getFirst() + L.getRest()
		 * </pre></code>
		 * is guaranteed, providing that the contents of L do not change.
		 *
		 * Note that this method assumes that a DAMLList contains only DAML values, since lists
		 * were only introduced in the DAML specification.  However, it is possible and sometimes
		 * useful to construct lists containing non-DAML values.  In such cases, this method
		 * will throw a class cast exception. However, <code>getAll().next()</code> has a weaker
		 * expectation: that the list will contain RDF resources.  This alternative means of
		 * accessing the list may be useful in such cases.
		 *
		 * @return The first value in the list, or, strictly, an unspecified value from the list.
		 * @see #getAll
		 */
		public DAMLCommon GetFirst() 
		{
			return (DAMLCommon) GetFirstResource();
		}


		/**
		 * Answer a new list that consists of all values of the list save the first. Since, strictly
		 * speaking, DAML collections are unordered, this corresponds to returning the collection
		 * minus an unspecified one of its values.  However, the identity
		 * <code><pre>
		 *      List L = L.getFirst() + L.getRest()
		 * </pre></code>
		 * is guaranteed, providing that the contents of L do not change.
		 *
		 * @return a list that contains all the elements of the current list save one.
		 */
		public DAMLList GetRest() 
		{
			Resource rest = GetRestResource();

			if (IsNil( rest )) 
			{
				// rest of the list is nil, so ensure we return the right value
				return GetVocabulary().Nil;
			}
			else 
			{
				if (rest is DAMLList) 
				{
					// this should always be the case for well-formed lists
					return (DAMLList) rest;
				}
				else 
				{
					Console.Error.WriteLine( "Badly formed list: rest of list should be type List, found " + rest );
					return null;
				}
			}
		}


		/**
		 * Answer a count of the items in the list.  Does not check for duplications, so this
		 * is the count of entries in the list, not the count of distinct items in the list.
		 *
		 * @return The number of entries in the list
		 */
		public int Count 
		{
			get 
			{
				int count = 0;

				for (IEnumerator i = GetAll();  i.MoveNext(); ) 
				{
					count++;
				}

				return count;
			}
		}

		/**
		 * Answer true if the list has no values. The nil list is empty,
		 * as is the list with no first and no rest. Note that we retain both
		 * conditions as we define the primary resource of a DAML value to be
		 * invariant (for indexing purposes). Therefore, a newly created list
		 * must be empty, but must not be identical to the nil list since we
		 * could never then change it to be a list of values.
		 *
		 * @return true for an empty list
		 */
		public bool IsEmpty() 
		{
			try 
			{
				return Equals( GetNil() )  ||
					(!HasProperty( GetVocabulary().first )  &&
					!HasProperty( GetVocabulary().rest  ));
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception " + e);
				throw new System.Exception( "RDF Exception " + e );
			}
		}


		/**
		 * Set the property <code>daml:first</code> for the given list element. This is a single
		 * value that denotes the value at this position of the list.
		 *
		 * @param value The value to be assigned to the 'first' property of a list cell
		 */
		public void SetFirst( DAMLCommon value ) 
		{
			SetPropertyValue( GetVocabulary().first, value );
		}


		/**
		 * Set the property <code>daml:rest</code> for the given list element. This is a single
		 * value that denotes the tail of the list.
		 *
		 * @param tail The value to be assigned to the tail of the list.
		 */
		public void SetRest( DAMLList tail ) 
		{
			SetPropertyValue( GetVocabulary().rest, tail );
		}


		/**
		 * Set the property <code>daml:rest</code> for the given list element to be the
		 * nil list. This correctly terminates the list at this point.
		 */
		public void SetRestNil() 
		{
			SetRest( GetVocabulary().Nil );
		}


		/**
		 * Answer a new list formed by creating a new DAMLList element whose first is the
		 * given value and whose rest is the current list.  This is the 'cons' operator
		 * familiar from other list processing languages.
		 *
		 * @param value The new value to be added to the head of the list
		 * @return a new list whose <code>daml:first</code> is the value, and whose
		 *         <code>daml:rest</code> is this list.
		 */
		public DAMLList Cons( DAMLCommon value ) 
		{
			// create a new list cell
			DAMLList l = new DAMLListImpl( null, GetDAMLModel(), GetVocabulary() );

			// set head and tail
			l.SetFirst( value );
			l.SetRest( this );

			return l;
		}


		/**
		 * Answer the well-known constant denototing the nil list.
		 *
		 * @return The resource denoting nil
		 */
		public DAMLList GetNil() 
		{
			return GetVocabulary().Nil;
		}


		/**
		 * Answer true if the given resource is the nil list.  Will check the
		 * namespace of the current vocabulary only.
		 *
		 * @param resource A resource, that may be the nil list
		 * @return true if the resource has the same URI as the nil list.
		 */
		public bool IsNil( Resource resource ) 
		{
			return resource.Equals( GetVocabulary().Nil );
		}


		/**
		 * Find the last list element, i.e.  the one whose rest is nil.
		 *
		 * @return A list element
		 */
		public DAMLList FindLast() 
		{
			if (GetRestResource().Equals( GetNil() )) 
			{
				// end of the line ... all change please
				return this;
			}
			else 
			{
				return GetRest().FindLast();
			}
		}


		/**
		 * Answer the i'th element of the list, if it exists.  If i is
		 * less than 1, or is larger than the length of the list, throw
		 * an illegal argument exception.
		 *
		 * @param i The position of the list to return
		 * @return The DAML value at the i'th position in the list
		 * @exception java.lang.IllegalArgumentException if i is less than one, or
		 *            larger than the length of the list.
		 */

		public DAMLCommon GetItem( int i ) 
		{
			if (i < 1) 
			{
				// there is no zero'th element
				throw new System.Exception( "Element " + i + " of a list is not defined" );
			}
			else 
			{
				// try to find the right list cell
				int count = i;
				DAMLList l = this;

				while (!l.IsEmpty()) 
				{
					if (count == 1) 
					{
						// this is the one
						return l.GetFirst();
					}
					else 
					{
						// go to the next list cell
						l = l.GetRest();
						count--;
					}
				}

				// if we get here, i > length( this )
				throw new System.Exception( "Tried to getItem( " + i + " ) of a list without that many elements" );
			}
		}

		public void SetItem( int i, DAMLCommon value)
		{
			if (i < 1) 
			{
				// there is no zero'th element
				throw new System.Exception( "Element " + i + " of a list is not defined" );
			}
			else 
			{
				// try to find the right list cell
				int count = i;
				DAMLList l = this;

				while (!l.IsEmpty()) 
				{
					if (count == 1) 
					{
						// this is the one
						l.SetFirst(value);
					}
					else 
					{
						// go to the next list cell
						l = l.GetRest();
						count--;
					}
				}

				// if we get here, i > length( this )
				throw new System.Exception( "Tried to getItem( " + i + " ) of a list without that many elements" );
			}

		}

		public DAMLCommon this[int index]
		{
			get 
			{
				return GetItem(index);
			}
			set
			{
				SetItem(index, value);
			}
		}


		// Internal implementation methods
		//////////////////////////////////



		/**
		 * Answer the resource that is the 'rest' property of the list
		 *
		 * @return RDF resource
		 */
		protected Resource GetRestResource() 
		{
			try 
			{
				return (Resource) GetPropertyValue( GetVocabulary().rest );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "rdf error: " + e );
				throw e;
			}
		}


		/**
		 * Answer the RDF resource that is the value of the first property
		 * on this list cell.
		 *
		 * @return The value of the head of the list, as an RDF resource.
		 */
		protected Resource GetFirstResource() 
		{
			return (Resource) GetPropertyValue( GetVocabulary().first );
		}


		/**
		 * Answer a key that can be used to index collections of this DAML list for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public override Object GetKey() 
		{
			return DAML_OIL._List.GetURI();
		}



		//==============================================================================
		// Inner class definitions
		//==============================================================================


		/**
		 * An iterator over the elements of a DAML list
		 */
		protected class DAMLListEnumerator : IEnumerator
		{
			// Instance variables
			/////////////////////

			/** The list cell we are currently pointing to */
			private DAMLList m_list = null;
			private DAMLList m_start = null;


			// Constructors
			/////////////////////

			/**
			 * Construct an iteration over a DAML list
				 *
				 * @param list The list to iterate over
			 */
			public DAMLListEnumerator( DAMLList list ) 
			{
				m_list = list;
				m_start = list;
			}


			// External signature methods
			//////////////////////////////

			/**
			 * Answer true if the iteration has more elements.
			 *
			 * @return true if there are more elements in the iteration
			 */
			public bool MoveNext() 
			{
				return (m_list != null)  &&  !(m_list.IsEmpty());
			}


			/**
			 * Answer the next DAML value from the iteration over the
				 * values in the list.  To comply with a principle of least
				 * committment, we only assume that the values from the
				 * list are resources (and not force them to be DAMLCommon objects).
			 *
				 * @return The next resource in the list
			 */
			public Object Current 
			{
				get 
				{
					Resource val = ((DAMLListImpl)m_list).GetFirstResource();
					m_list = m_list.GetRest();

					return val;
				}
			}


			/**
			 * Remove an element from the list. Not supported.
			 */
			public void Reset() 
			{
				m_list = m_start;
			}
		} // Inner class DAMLListIterator



	}
}
