using System;
using System.Collections;
using System.Net;
using System.IO;
using BBN.DAML.Vocabulary;	
using BBN.RDF.Common;
using BBN.RDF;
using BBN.RDF.Mem;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class DAMLLoader
	{
		// Constants
		//////////////////////////////////

		/* Misc constants */

		/** Prefix for a URI to a file */
		//    public static final string FILE_URI_PREFIX = "file://";

		/* Status flags */

		/** {@link #getStatus Status} flag: OK */
		public const long STATUS_OK = 0;

		/** {@link #getStatus Status} flag: input source is not available (e.g. a file cannot be opened) */
		public const long STATUS_INPUT_UNAVAILABLE = 1;

		/** {@link #getStatus Status} flag: an I/O error occurred - details will be in log file. */
		public const long STATUS_IO_ERROR = 2;

		/** {@link #getStatus Status} flag: a syntax error occurred in the DAML file- details will be in log file. */
		public const long STATUS_SYNTAX_ERROR = 4;

		/** {@link #getStatus Status} flag: a misc error occurred - report a bug to the maintainer! */
		public const long STATUS_MISC_ERROR = 8;


		// Static variables
		//////////////////////////////////

		/** The signature for the constructor we want to invoke on the DAML values */
		static Type[] s_constructSig = new Type[] {typeof(string),
													  typeof(DAMLModel),
													  typeof(DAMLVocabulary)};


		// Instance variables
		//////////////////////////////////

		/** A status flag we can use to inform callers what just happened */
		private long m_status = STATUS_OK;

		/** A reference to the storage service this class is using */
		private DAMLModel m_damlModel = null;

		/** For efficiency, keep a list of the ontologies that were loaded into this store */
		private ArrayList m_sources = new ArrayList();

		/** Flag to control whether imported technologies are loaded automatically */
		private bool m_loadImportedOntologies = true;

		/** Vector of ontologies that are not automatically fetched when loading containing ontologies */
		private ArrayList m_importBlockList = new ArrayList();

		/** Flag used to control the import blocking behaviour */
		private bool m_useImportBlocking = true;

		/** Dictionary that holds the DAML resources that correspond to loaded RDF resources.
			 *  Key = original rdf Resource,  value = corresponding DAMLCommon instance
			 */
		private Hashtable m_rdfDamlMap = new Hashtable();

		/** Queue of resources to check for being instances */
		private Hashtable m_postCheckResources = new Hashtable();

		/** Dispatch table to handle different kinds of resource in the RDF model once parsed */
		private ResourceDispatcher[] m_dispatchTable = {
														   new ResourceDispatcher( DAML_OIL._Class,                 typeof(DAMLClassImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._Class,         typeof(DAMLClassImpl) ),
														   new ResourceDispatcher( BBN.RDF.Vocabulary.RDFS.Class,   typeof(DAMLClassImpl) ),

														   new ResourceDispatcher( DAML_OIL_2000_12._Disjoint,      typeof(DAMLDisjointImpl) ),

														   new ResourceDispatcher( DAML_OIL._Restriction,           typeof(DAMLRestrictionImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._Restriction,   typeof(DAMLRestrictionImpl) ),

														   new ResourceDispatcher( DAML_OIL._List,                  typeof(DAMLListImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._List,          typeof(DAMLListImpl) ),

														   new ResourceDispatcher( DAML_OIL._Ontology,              typeof(DAMLOntologyImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._Ontology,      typeof(DAMLOntologyImpl) ),

														   new ResourceDispatcher( DAML_OIL._Property,              typeof(DAMLPropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._Property,      typeof(DAMLPropertyImpl) ),
														   new ResourceDispatcher( BBN.RDF.Vocabulary.RDF.Property, typeof(DAMLPropertyImpl) ),

														   new ResourceDispatcher( DAML_OIL._DatatypeProperty,      typeof(DAMLDatatypePropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL._ObjectProperty,        typeof(DAMLObjectPropertyImpl) ),

														   new ResourceDispatcher( DAML_OIL_2000_12._UniqueProperty, typeof(DAMLPropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._TransitiveProperty, typeof(DAMLPropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._UnambiguousProperty, typeof(DAMLPropertyImpl) ),

														   new ResourceDispatcher( DAML_OIL._UniqueProperty,        typeof(DAMLPropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL._TransitiveProperty,    typeof(DAMLObjectPropertyImpl) ),
														   new ResourceDispatcher( DAML_OIL._UnambiguousProperty,   typeof(DAMLObjectPropertyImpl) ),

														   new ResourceDispatcher( DAML_OIL._Thing,                 typeof(DAMLInstanceImpl) ),
														   new ResourceDispatcher( DAML_OIL_2000_12._Thing,         typeof(DAMLInstanceImpl) )
													   };


		// Constructors
		//////////////////////////////////

		/**
			 * Construct a new DAMLLoader that will load definitions into the given DAML model.
			 *
			 * @param m The DAML model that will contain the loaded definitions.
			 */

		public DAMLLoader( DAMLModel m ) 
		{
			m_damlModel = m;

			// initialise the list of blocked imports (for efficiency, nothing more)
			InitialiseImportBlockList();
		}


		// External signature methods
		//////////////////////////////////


		/**
			 * Answer the DAML model that this loader is using to store the loaded definitions.
			 *
			 * @return A reference to a knowledge store.
			 */

		public DAMLModel GetDAMLModel() 
		{
			return m_damlModel;
		}


		/**
			 * Read the DAML descriptions from the document at the given URL.
			 *
			 * @param source The URL of the DAML source document to be read
			 * @param base The base URI to use for relative names
			 * @param lang The encoding language for the source document, or null for the default.
			 */
		public void Read( Uri source, string b, string lang ) 
		{
			Read(source.AbsolutePath, b, lang);
		}



		/**
			 * Read the DAML descriptions from the document at the given URI.
			 *
			 * @param uri The source URI to be loaded, as a string
			 * @param base The base URI to use for relative names
			 * @param lang The encoding language for the source document, or null for the default.
			 */

		public void Read( string uri, string b, string lang ) 
		{
			try 
			{
				// TBD - need to add this as a source.  Is this baseURI, or a new arg?

				Model sourceModel = new ModelMem();
				sourceModel._cleanLoad = this.GetDAMLModel()._cleanLoad;

				// collect the RDF sources, including those from imported ontologies
				CollectRDFSources( sourceModel, uri, b, lang );

				// finally add the statements to the main model, and build the DAML wrapper objects
				//DateTime start = DateTime.Now;
				ProcessRDFStatements( sourceModel );
				//DateTime end = DateTime.Now;
				//Console.WriteLine(end.Subtract(start).ToString());
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "Error occurred in reading RDF source: " + e );
				SetStatus( STATUS_SYNTAX_ERROR );
			}
		}





		/**
			 * Read the descriptions from the given input reader, whose
			 * document base URI is as given.
			 *
			 * @param reader An input reader for the definitions document
			 * @param baseURI The definition document's base URI (necessary to resolve
			 *            relative names)
			 * @param lang The encoding language for the source document, or null for the default.
			 */
		public void Read( System.IO.TextReader reader, string source, string baseURI, string lang ) 
		{
			// either this source document is new, or we are reloading anyway
			// first pull the descriptions in to a temporary model
			try 
			{
				// TBD - need to add this as a source.  Is this baseURI, or a new arg?
				AddSource(source);
				Model sourceModel = new ModelMem();

				// collect the RDF sources, including those from imported ontologies
				CollectRDFSources( sourceModel, reader, baseURI, lang );

				// finally add the statements to the main model, and build the DAML wrapper objects
				//DateTime start = DateTime.Now;
				ProcessRDFStatements( sourceModel );
				//DateTime end = DateTime.Now;
				//Console.WriteLine(end.Subtract(start).ToString());
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "Error occurred in reading RDF source: " + e );
				SetStatus( STATUS_SYNTAX_ERROR );
			}
		}

		public void Read( System.IO.TextReader reader, string baseURI, string lang)
		{
			Read(reader, baseURI, baseURI, lang);
		}

		public void Read( System.IO.Stream stream, string source, string baseURI, string lang ) 
		{
			// either this source document is new, or we are reloading anyway
			// first pull the descriptions in to a temporary model
			try 
			{
				// TBD - need to add this as a source.  Is this baseURI, or a new arg?
				AddSource(source);
				Model sourceModel = new ModelMem();

				// collect the RDF sources, including those from imported ontologies
				CollectRDFSources( sourceModel, stream, baseURI, lang );

				// finally add the statements to the main model, and build the DAML wrapper objects
				//DateTime start = DateTime.Now;
				ProcessRDFStatements( sourceModel );
				//DateTime end = DateTime.Now;
				//Console.WriteLine(end.Subtract(start).ToString());
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "Error occurred in reading RDF source: " + e );
				SetStatus( STATUS_SYNTAX_ERROR );
			}
		}

		public void Read( System.IO.Stream stream, string baseURI, string lang)
		{
			Read(stream, baseURI, baseURI, lang);
		}

		public void ReadFromString( string xml, string b, string lang ) 
		{
			try 
			{
				// TBD - need to add this as a source.  Is this baseURI, or a new arg?
				AddSource(b);
				Model sourceModel = new ModelMem();

				// collect the RDF sources, including those from imported ontologies
				CollectRDFSourcesFromString( sourceModel, xml, b, lang );

				// finally add the statements to the main model, and build the DAML wrapper objects
				ProcessRDFStatements( sourceModel );
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "Error occurred in reading RDF source: " + e );
				SetStatus( STATUS_SYNTAX_ERROR );
			}
		}



		/**
			 * Add the statements from the given external model into the model
			 * this loader is attached to.  NB Does not process daml:imports statements.
			 *
			 * @param model A source model to load statements from
			 */

		public void Add( Model model ) 
		{
			ProcessRDFStatements( model );
		}



		/**
			 * <p>
			 * Initialise the list of well-known ontologies that we don't bother
			 * to load in we detect an import statement.  Currently, the default value
			 * for this list is:
			 * <ul>
			 *     <li> The DAML 2000/12 release </li>
			 *     <li> The DAML 2001/3 release </li>
			 *     <li> The RDF Schema 2000/01 release </li>
			 * </ul>
			 * </p><p>
			 * This list can be modified in several ways: use {@link #GetImportBlockList}
			 * to get reference to the list and add/remove elements as required, override
			 * this method in a sub-class, or turn off all automatic importing with
			 * {@link #SetLoadImportedOntologies}.
			 */

		protected void InitialiseImportBlockList() 
		{
			m_importBlockList.Add( GetURIRoot( DAML_OIL.NAMESPACE_DAML_2000_12_URI ) );
			m_importBlockList.Add( GetURIRoot( DAML_OIL.NAMESPACE_DAML_2001_03_URI ) );
			m_importBlockList.Add( GetURIRoot( BBN.RDF.Vocabulary.RDFS.GetURI() ) );
		}



		/**
			 * Answer an iterator over the set of URI strings that <b>will not</b>
			 * be loaded if encountered in an import statement, even if autoloading of
			 * ontologies (see {@link #GetLoadImportedOntologies()}) is on.
			 *
			 * @return iterator over the set of blocked imports, as URI strings.
			 */

		public IEnumerator GetImportBlockList() 
		{
			return m_importBlockList.GetEnumerator();
		}



		/**
			 * Answer true if a given URI is blocked from being imported: that is, it is
			 * assumed to be well-known and will not be automatically imported.  Will
			 * answer false if import blocking is switched off altogether
			 * (see {@link #setUseImportBlocking}).
			 *
			 * @param uri The uri to be tested
			 * @return true if the URI should not be loaded.
			 */

		public bool IsBlockedImport( string uri ) 
		{
			return m_useImportBlocking  &&   m_importBlockList.Contains( uri );
		}



		/**
			 * Add the given URI to the list of ontology urls that will not be loaded
			 * if encountered in an imports statement in the loaded ontology.
			 *
			 * @param uri The URI of the ontology to block from autoloading, as a String.
			 */

		public void AddImportBlock( string uri ) 
		{
			// maintain the invariant that there is only copy of each entry
			if (!m_importBlockList.Contains( uri )) 
			{
				m_importBlockList.Add( uri );
			}
		}



		/**
			 * Remove the given URI from the list of ontology urls that will not be loaded
			 * if encountered in an imports statement in the loaded ontology.  Has no effect
			 * if the URI is not in the list.
			 *
			 * @param uri The URI of the ontology to no longer block from autoloading, as a String.
			 */

		public void RemoveImportBlock( string uri ) 
		{
			m_importBlockList.Remove( uri );
		}



		/**
			 * Answer true if the loader is to process imported ontologies (except for the
			 * ones on the don't load list). Default true.
			 *
			 * @return True if imported ontologies should be loaded as they are encountered.
			 */

		public bool GetLoadImportedOntologies() 
		{
			return m_loadImportedOntologies;
		}



		/**
			 * Set the flag to control whether imported ontologies are to be loaded. Default
			 * true.
			 *
			 * @param loadImports If true, ontologies that are included in this one, via
			 *                     the &lt;imports&gt; element are loaded as they are discovered.
			 */

		public void SetLoadImportedOntologies( bool loadImports ) 
		{
			m_loadImportedOntologies = loadImports;
		}



		/**
			 * Set the flag to control whether certain well-known imports are blocked
			 * from being automatically loaded if they encountered in an <code>imports</code>
			 * statement. Note that automatic loading of all imports can be switched of
			 * with {@link #SetLoadImportedOntologies}.
			 *
			 * @param useBlocking If true, well-known URI's will be blocked from being
			 *                    autoloaded, even if autoloading is on
			 */

		public void SetUseImportBlocking( bool useBlocking ) 
		{
			m_useImportBlocking = useBlocking;
		}



		/**
			 * Answer true if well-known URI's will be blocked from being autoloaded.
			 *
			 * @return True if imports are blocked.
			 * @see #SetUseImportBlocking
			 */

		public bool GetUseImportBlocking() 
		{
			return m_useImportBlocking;
		}



		/**
			 * Answer true if the ontology identified by the given URI has been loaded.
			 *
			 * @param uri The URI of the ontology
			 * @return true if the ontology has already been loaded by the knowledge store.
			 */

		public bool IsLoadedOntology( string uri ) 
		{
			return (GetSourcePosition( uri ) != -1);
		}




		/**
			 * Answer the status of the last operation.  Returns a set of status flags
			 * (or'ed together) from all operations since the last {@link #resetStatus}.
			 *
			 * @return The current status, as a set of flags or'ed together into a long.
			 */

		public long GetStatus() 
		{
			return m_status;
		}



		/**
			 * Clear the status flags. Resets the status to {@link #STATUS_OK}. Note that
			 * status is cleared automatically at the start of a {@link com.hp.hpl.jena.daml.DAMLModel#read( java.lang.string ) read}.
			 */

		public void ResetStatus() 
		{
			m_status = STATUS_OK;
			// SetStatus( STATUS_OK );
		}




		// Internal implementation methods
		//////////////////////////////////


		/**
			 * Set the status.  The given flag is or'ed to the current status.
			 *
			 * @param status A status flag, which should be one of the STATUS_xxx flags.
			 */

		protected void SetStatus( long status ) 
		{
			m_status |= status;
		}



		/**
		 * Collect all of the RDF statements for the model to be processed, including
		 * the transitive closure of the imported models if appropriate.  Each import
		 * should only be processed once.
		 */
		protected void CollectRDFSources( Model m, string uri, string baseURI, string lang )
		{
			// JENA 1.5.0 now uses ARP to read the data.

			// load the input as an RDF document
			//DER - handle case where baseURI is null
			if (baseURI == null)
			{
				AddSource(uri);
				m.Read(uri);
			} 
			else 
			{
				AddSource(baseURI);
				m.Read( uri, baseURI );
			}

			// we load any included ontologies (so that they are available when processing the
			// statements in this document
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL._Ontology );
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL_2000_12._Ontology );
		}


		/**
		 * Collect all of the RDF statements for the model to be processed, including
		 * the transitive closure of the imported models if appropriate.  Each import
		 * should only be processed once.
		 */

		protected void CollectRDFSources( Model m, System.IO.TextReader reader, string baseURI, string lang )
		{
			AddSource(baseURI);
			// load the input as an RDF document
			if (lang == null) 
			{
				m.Read( reader, baseURI );
			}
			else 
			{
				m.Read( reader, baseURI, lang );
			}

			// we load any included ontologies (so that they are available when processing the
			// statements in this document
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL._Ontology );
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL_2000_12._Ontology );
		}

		protected void CollectRDFSources( Model m, System.IO.Stream stream, string baseURI, string lang )
		{
			AddSource(baseURI);
			// load the input as an RDF document
			if (lang == null) 
			{
				m.Read( stream, baseURI );
			}
			else 
			{
				m.Read( stream, baseURI, lang );
			}

			// we load any included ontologies (so that they are available when processing the
			// statements in this document
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL._Ontology );
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL_2000_12._Ontology );
		}

		protected void CollectRDFSourcesFromString( Model m, string xml, string baseURI, string lang )
		{
			AddSource(baseURI);

			// load the input as an RDF document
			m.ReadFromString( xml, baseURI );

			// we load any included ontologies (so that they are available when processing the
			// statements in this document
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL._Ontology );
			LoadImportedOntologies( m, BBN.DAML.Vocabulary.DAML_OIL_2000_12._Ontology );
		}

		
		/**
			 * Load the imported ontologies mentioned by Ontology instances in the given
			 * model.
			 *
			 * @param model An RDF model of the ontology we're loading
			 * @param ontologyInstanceType Resource denoting the rdf:type of the Ontology element
			 *                             that will contain the import statement
			 */

		protected void LoadImportedOntologies( Model model, DAMLClass ontologyInstanceType )
		{
			// first check that we are loading included ontologies
			if (GetLoadImportedOntologies()) 
			{
				// first find if there is an Ontology resource
				ResEnumerator i0 = model.Query( new SelectorImpl( null, BBN.RDF.Vocabulary.RDF.type, ontologyInstanceType ) )
					.ListSubjects();
				ResEnumerator i1 = model.Query( new SelectorImpl( null, BBN.DAML.Vocabulary.DAML_OIL._type, ontologyInstanceType ) )
					.ListSubjects();

				// we collect the imported URI's first, then load them ... otherwise we're
				// iterating over a model that is being updated, which is not guaranteed to be safe
				ArrayList imports = new ArrayList();
				for (IEnumerator j = new ConcatenatedEnumerator( i0, i1 );  j.MoveNext();  ) 
				{
					Resource ontologyRes = (Resource) j.Current;

					for (NodeEnumerator i = model.ListObjectsOfProperty( ontologyRes, ontologyInstanceType.GetVocabulary().imports ); i.MoveNext(); ) 
					{
						// add the value of each import, as a string, to the collection
						imports.Add( i.Current.ToString() );
					}
				}

				// now we can import the ontologies mentioned, subject to some checks
				for (IEnumerator i = imports.GetEnumerator();  i.MoveNext();  ) 
				{
					string importedURI = (string) i.Current;

					// we don't bother loading some well-known URI's for efficiency, nor
					// do we load the same URI twice
					if ((GetSourcePosition( importedURI ) < 0)  &&  (!IsBlockedImport( importedURI ))) 
					{
						//DER - Don't pass base URI if it is the current uri.
						CollectRDFSources( model, importedURI, null, null );
						//CollectRDFSources( model, importedURI, importedURI, null );
					}
				}
			}
		}


		/**
			 * Iterate through all of the top-level resources in the input model and
			 * interpret them as DAML objects, and hence make the appropriate assertions
			 * in the knowledge store.
			 *
			 * @param sourceModel The RDF model that contains all of the source RDF statements
			 *                    we have loaded from the source document.
			 */

		protected void ProcessRDFStatements( Model sourceModel ) 
		{
			try 
			{
				// Get the DAML values that correspond to the vanilla RDF resources
				MapRDFValuesToDAML( sourceModel );

				// now replace the actual statements in the main model
				for (StmtEnumerator i = sourceModel.ListStatements();  i.MoveNext();  ) 
				{
					// Get the statement in vanilla RDF form
					Statement sRDF = i.Current;

					// recreate it with DAML objects
					GetDAMLModel().Add( (Resource) MapDAMLNode( sRDF.GetSubject() ),
						(Property) MapDAMLNode( sRDF.GetPredicate() ),
						MapDAMLNode( sRDF.GetObject() ) );
				}
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "RDF exception while processing DAML statements: " + e );
				SetStatus( STATUS_MISC_ERROR );
			}
		}



		/**
			 * For each resource that is the subject of a statement in the loaded model,
			 * map it to the corresponding DAML resource and save the mapping in a global
			 * dictionary.
			 *
			 * @param sourceModel The model that contains the RDF statements we are loading.
			 */

		protected void MapRDFValuesToDAML( Model sourceModel ) 
		{
			try 
			{
				// iterate through all the known statements so that we can create the DAML shadow classes
				for (StmtEnumerator statements = sourceModel.ListStatements();  statements.MoveNext(); ) 
				{
					// Get the statement, and extract its components
					Statement s = statements.Current;

					Resource subj = s.GetSubject();
					Property prop = s.GetPredicate();
					RDFNode obj = s.GetObject();

					// now map them to DAML values
					MapRDFResourceToDAML( sourceModel, subj );
					MapRDFResourceToDAML( sourceModel, prop );
					if (obj is BBN.RDF.Resource) 
					{
						MapRDFResourceToDAML( sourceModel, (Resource) obj );
					}
				}

				// now check the postponed resources, to see if they correspond to instances
				for (IEnumerator i = m_postCheckResources.Keys.GetEnumerator();  i.MoveNext();  ) 
				{
					// Get the resource and its type
					Resource r = (Resource) i.Current;

					// Get the type statement for the resource, if it has one
					Resource type = FindType( sourceModel, r );

					if (type != null) 
					{
						// can we match the type to a DAML class?
						if (IsDAMLClass( type )) 
						{
							// yes - so map the original resource to an instance object
							m_rdfDamlMap[ r ] = ((DAMLModelImpl) GetDAMLModel()).Index( new DAMLInstanceImpl( r.GetURI(), GetDAMLModel(), VocabularyManager.GetVocabulary( type.GetURI() ) ) );
						}
						else if (IsDAMLDatatype( type )) 
						{
							// the type matches a known datatype, so this is a value from concrete domain
							m_rdfDamlMap[ r ] = ((DAMLModelImpl) GetDAMLModel()).Index( new DAMLDataInstanceImpl( r.GetURI(), GetDAMLModel(), VocabularyManager.GetVocabulary( r.GetURI() ) ) );

							// we also create the Datatype object itself, in place of the type
							if (!m_rdfDamlMap.ContainsKey( type )) 
							{
								m_rdfDamlMap[ type ] = new DAMLDatatypeImpl( type.GetURI(), GetDAMLModel(),
									VocabularyManager.GetVocabulary( type.GetURI() ));
							}
						}
						else 
						{
							/* TBD Should This Be Turned Off?
							Console.Error.WriteLine( "Failed to match resource <" + r.GetURI() +
								"> of type <" +
								(type == null ? "unknown" : type.GetURI()) +
								"> in dispatch table" );
								*/
						}
					}
				}
			}
			catch (Exception e) 
			{
				Console.Error.WriteLine( "RDF exception while processing DAML statements: " + e );
				SetStatus( STATUS_MISC_ERROR );
			}
		}



		/**
			 * For a given resource, attempt to map it to a specific instance of a DAML value
			 * if it's typed as a DAML object.  Only do this once for each RDF resource.
			 *
			 * @param model The model that is the source of the RDF we are processing
			 * @param r The RDF resource from the source model.
			 */

		protected void MapRDFResourceToDAML( Model sourceModel, Resource r ) 
		{
			// first check if we have seen this resource already
			if (m_rdfDamlMap.ContainsKey( r )) 
			{
				// been there. done that. got the t-shirt
				return;
			}
			else 
			{
				// now, is this a value we've seen before?
				DAMLCommon damlRes = GetDAMLModel().GetDAMLValue( r.GetURI() );

				if (damlRes != null) 
				{
					// we know the correspondance for this URI, but it's not yet in the map
					// for the DAML document we're processing at present
					m_rdfDamlMap[ r ] = damlRes;
				}
				else 
				{
					// otherwise, we try to match it based on its type
					// we assume that all the resources we are interested in have an RDF type
					Resource type = FindType( sourceModel, r );

					// only proceed if we have a type
					if (type != null) 
					{
						bool dispatched = false;

						// no existing DAML object, so attempt to match each one in the dispatch table
						for (int i = 0;  (!dispatched)  &&  (i < m_dispatchTable.Length);  i++) 
						{
							dispatched = m_dispatchTable[i].Dispatch( r, type, GetDAMLModel(), m_rdfDamlMap );
						}

						// check for failure to handle a resource, and save for later to check for an Instance instance
						if (!dispatched) 
						{
							m_postCheckResources[ r ] = true;
						}
					}
				}
			}	
		}



		/**
			 * Answer the DAML value that corresponds to the given RDF node, if we
			 * have one, otherwise return the original node.
			 *
			 * @param node An RDF node
			 * @return A DAML value corresponding to the node, or otherwise the node itself.
			 */

		protected RDFNode MapDAMLNode( RDFNode node ) 
		{
			Resource rDAML = (Resource) m_rdfDamlMap[ node ];
			return (rDAML == null) ? node : rDAML;
		}



		/**
			 * Answer true if a resource corresponds to a DAML class, either in the known model or
			 * the freshly loaded classes.
			 *
			 * @param resource The resource to match against
			 * @return true if the resource denotes a class
			 */

		protected bool IsDAMLClass( Resource resource ) 
		{
			// first try the main model
			DAMLCommon dClass = GetDAMLModel().GetDAMLValue( resource.GetURI() );
			if (dClass != null  &&  dClass is BBN.DAML.DAMLClass) 
			{
				return true;
			}

			// otherwise, try the daml classes that have been loaded from the current model
			dClass = (DAMLCommon) m_rdfDamlMap[ resource ];
			return dClass != null  &&  dClass is BBN.DAML.DAMLClass;
		}



		/**
			 * Answer true if a resource corresponds to a DAML datatype, which we can tell
			 * if its type URI is in the type registry for mapping concrete domains.
			 *
			 * @param resource The resource that may represent a concrete datatype
			 */

		protected bool IsDAMLDatatype( Resource resource ) 
		{
			return GetDAMLModel().GetDatatypeRegistry().IsRegisteredType( resource.GetURI() );
		}



		/**
			 * Answer the position of the given ontology in the list of ontology entries,
			 * or -1 if it is not present.
			 *
			 * @param uri The URI of the ontology we are looking for.
			 * @return The position of the ontology in the list, or -1
			 */

		protected int GetSourcePosition( string uri ) 
		{
			return (m_sources.IndexOf( uri ));
		}



		/**
			 * Add a source URI to the list of sources we're maintaining.
			 *
			 * @param uri The URI we have just loaded.
			 */

		protected void AddSource( string uri ) 
		{
			//Console.WriteLine("added "+uri);
			m_sources.Add( uri );
		}




		/**
			 * Utility method to remove get the root name of a URI (e.g. everything up to the anchor).
			 *
			 * @param uri A uri to process
			 * @return the root of the uri, up to but not including the anchor char.
			 */

		private string GetURIRoot( string uri ) 
		{
			// RDF has peculiar rules about separating names from namespaces
			int splitPoint = Util.SplitNamespace( uri );
			return (splitPoint < 0) ? uri : uri.Substring( 0, splitPoint - 1 );
		}



		/**
			 * Answer one of the type statements for this object. Only one type is returned,
			 * so if a resource has more than one type the choice will be non-deterministic -
			 * subject to the heuristic that we eliminate less specific classes in favour of
			 * more specific ones.
			 *
			 * @param model The RDF model to do the lookup in
			 * @param res The resource whose type is required
			 * @return a statement with res as subject and rdf:type or one of its aliases
			 *         as predicate
			 */

		private Resource FindType( Model model, Resource res ) 
		{
			ArrayList types = new ArrayList();

			// first try rdf:type (i.e. the most common occurrence)
			try 
			{
				for (StmtEnumerator t = res.ListProperties( BBN.RDF.Vocabulary.RDF.type );  t.MoveNext(); ) 
				{
					RDFNode typeNode = ((Statement) t.Current).GetObject();
					if (typeNode is Resource) 
					{
						types.Add( typeNode );
					}
				}
			}
			catch (Exception) {}

			// then try aliases to rdf:type, such as daml:type
			for (IEnumerator i = DAMLHierarchy.GetInstance().GetEquivalentValues( BBN.RDF.Vocabulary.RDF.type );  i.MoveNext();  ) 
			{
				Property typeAlias = (Property) i.Current;

				try 
				{
					for (StmtEnumerator t  = res.ListProperties( typeAlias );  t.MoveNext(); ) 
					{
						RDFNode typeNode = ((Statement) t.Current).GetObject();
						if (typeNode is Resource) 
						{
							types.Add( typeNode );
						}
					}
				}
				catch (Exception) {}
			}

			// heuristically, we discard any types that are super-types of others in the set
			ArrayList heurTypes = new ArrayList();
			for (IEnumerator i = types.GetEnumerator();  i.MoveNext();  ) 
			{
				DAMLHierarchy damlHier = DAMLHierarchy.GetInstance();

				// get the type object
				Resource type = (Resource) i.Current;
				string typeURI = type.GetURI();
				bool discard = false;

				for (IEnumerator j = types.GetEnumerator();  (!discard && j.MoveNext()); ) 
				{
					// set a flag if the value we're testing is a known super-class of another type in the set
					discard = damlHier.IsDAMLSubClassOf( ((Resource) j.Current).GetURI(), typeURI );
				}

				// <ickyHack>
				// there's one nasty case which is worth treating specially, which is a unique
				// datatype property. In this case, we want to keep DatatypeProperty, rather than
				// UniqueProperty, as the type. This is because later on UniqueProperty will get mapped to
				// DAMLProperty, which is a super-class of DatatypeProperty.
				discard = discard ||
					(type.Equals( DAML_OIL._UniqueProperty ) &&
					types.Contains( DAML_OIL._DatatypeProperty ));
				// </ickyHack>

				// only keep more specific classes, so don't bother if it's a super-class
				if (!discard) 
				{
					heurTypes.Add( type );
				}
			}

			// now, heurTypes contains the set of types of this resource, with some
			// additional hueristic filtering applied.  We pick a random member of
			// this set to return.
			IEnumerator k = heurTypes.GetEnumerator();
			return (Resource) (k.MoveNext() ? k.Current : null);
		}





		//==============================================================================
		// Inner class definitions
		//==============================================================================

		/**
		 * A dispatcher class for different types of resource that we can handle.
		 */
		internal class ResourceDispatcher
		{
			// Instance variables
			/////////////////////

			/** Holds the URI of the type of the node for a default match */
			internal Resource m_typeRes = null;

			/** The class that this dispatcher will construct by default */
			internal Type m_defaultClass = null;


			// Constructors
			/////////////////////

			/**
			 * Constructor - accepts the rdfs:type object of the resource to match on for default
			 * matches.
			 *
			 * @param typeRes The Resource denoting the rdfs:type to match.
			 * @param cls The class for the default handler to create instances of, or null
			 */
			internal ResourceDispatcher( Resource typeRes, Type cls ) 
			{
				m_typeRes = typeRes;
				m_defaultClass = cls;
			}


			/**
			 * Constructor - does not initialise a default match.  Method matches() should
			 * be overridden to provide specialised match conditions.
			 */
			internal ResourceDispatcher() 
			{
			}


			// External signature methods
			/////////////////////////////

			/**
			 * Test for a match with the locally stored parameters, and, if a match occurs
			 * dispatch to the appropriate handler.
			 *
			 * @param r The resource to be dispatched
			 * @param typeRes The value of the resource's rdfs:type if it has one
			 * @return True if this handler was able to dispatch the resource.
			 */
			internal bool Dispatch( Resource r, Resource typeRes, DAMLModel model, Hashtable rdfDamlMap  ) 
			{
				if (Matches( r, typeRes )) 
				{
					// get the DAML object that is equivalent to the given resource
					Resource damlShadow = HandleResource( r, model, typeRes );

					// save it in the global dictionary for later processing
					rdfDamlMap[r] = damlShadow;

					return true;
				}
				else 
				{
					return false;
				}
			}


			/**
			 * Default matcher: answer true if the node's type URI is the same as the one
			 * cached in this dispatcher object.
			 *
			 * @param res The resource itself (to allow more sophisticated matches in sub-classes)
			 * @param typeRes The resource denoting the node's rdfs:type, or null.
			 */
			internal bool Matches( Resource res, Resource typeRes ) 
			{
				if (m_typeRes == null) 
				{
					Console.Error.WriteLine( "Attempting default match in resource dispatcher, but with null type URI!" );
					return false;
				}
				else 
				{
					return m_typeRes.Equals( typeRes );
				}
			}


			/**
			 * Handler method to consume the resource once it has been matched. Default
			 * action is to construct a new instance of the class that was supplied
			 * to the constructor of this dispatcher.
			 *
			 * @param r The resource that is being handled
			 * @param m The DAML model that will store the resulting resource
			 * @param typeRes The rdf:type of this resource
			 * @return The DAML model resource corresponding to the given resource, or null if error.
			 */
			internal Resource HandleResource( Resource r, DAMLModel m, Resource typeRes ) 
			{
				if (m_defaultClass != null) 
				{
					// default action is to create a DAML object for r and index it
					return ((DAMLModelImpl) m).Index( MakeInstance( m_defaultClass, r.GetURI(), m, typeRes.GetURI() ) );
				}
				else 
				{
					Console.Error.WriteLine( "Default dispatcher attempted to create a DAML value with no class available." );
					return null;
				}
			}


			// Internal implementation methods
			///////////////////////////////////

			/**
			 * Generate an instance of the given class, with the given resource
			 * as its primary.
			 *
			 * @param cls The class object for which we want a new instance
			 * @param uri The URI of the new object
			 * @param model The DAML model that the object is attached to
			 * @param typeURI The URI of the rdf:type of this resource
			 * @return A new DAMLCommon object
			 */
			internal DAMLCommon MakeInstance( Type cls, string uri, DAMLModel model, string typeURI ) 
			{
				try 
				{
					// get the constructor we want to invoke and make a new instance
					/* return (DAMLCommon) cls.GetDeclaredConstructor( s_constructSig )
						.NewInstance( new Object[] {uri, model, VocabularyManager.GetVocabulary( typeURI )} );
					 */
					return (DAMLCommon)Activator.CreateInstance(cls, new object[] {uri, model, VocabularyManager.GetVocabulary( typeURI )});

				}
				catch (Exception e) 
				{
					Console.Error.WriteLine( "Failed to construct DAML value " + cls.FullName+ ": " + e);
					throw new Exception( "Unexpected error while constructing DAML value from RDF model: " + e );
				}
			}

		} // end inner class ResourceDispatcher


	}

}

