using System;
using BBN.DAML.Vocabulary;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class DAMLObjectPropertyImpl : DAMLPropertyImpl, DAMLObjectProperty
	{

		// Instance variables
		//////////////////////////////////

		/** Property accessor for inverseOf */
		private PropertyAccessor m_propInverseOf = null;



		// Constructors
		//////////////////////////////////

		/**
		 * Constructor, takes the name and namespace for this property, and the underlying
		 * model it will be attached to.
		 *
		 * @param namespace The namespace the property inhabits, or null
		 * @param name The name of the property
		 * @param store The RDF store that contains the RDF statements defining the properties of the property
		 * @param vocabulary Reference to the DAML vocabulary used by this property.
		 */
		public DAMLObjectPropertyImpl( string ns, string name, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( ns, name, store, vocabulary )
		{
			SetRDFType( GetVocabulary().ObjectProperty );
		}


		/**
		 * Constructor, takes the URI for this property, and the underlying
		 * model it will be attached to.
		 *
		 * @param uri The URI of the property
		 * @param store The RDF store that contains the RDF statements defining the properties of the property
		 * @param vocabulary Reference to the DAML vocabulary used by this property.
		 */
		public DAMLObjectPropertyImpl( string uri, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( uri, store, vocabulary )
		{
			SetRDFType( GetVocabulary().ObjectProperty );
		}



		// External signature methods
		//////////////////////////////////


		/**
			 * Return a readable representation of the DAML value
			 *
			 * @return a string denoting this value
			 */
		public override string ToString() 
		{
			// get the public name for this value type (e.g. we change DAMLClassImpl -> DAMLClass)
			string cName = GetType().Name;
			int i = cName.IndexOf("Impl");
			if (i > 0) cName = cName.Substring(0, i);
	
			// property attributes
			string attribs;
			attribs = Unique ? "unique " : "";
			attribs = attribs + (Transitive ? "transitive " : "");
			attribs = attribs + (Unambiguous ? "unambiguous " : "");

			// now format the return string
			return (GetURI() == null) ?
				("<" + attribs + "Anonymous " + cName + "@" + GetHashCode().ToString("X") + ">") :
				("<" + attribs + cName + " " + GetURI() + ">");
		}


		/**
		 * Set the flag to indicate that this property is to be considered
		 * transitive - that is, it is defined by the DAML class TransitiveProperty.
		 *
		 * @param transitive True for a transitive property
		 */



		/**
		 * Answer true if this property is to be considered transitive, that is
		 * it is characterised by the DAML class TransitiveProperty
		 *
		 * @return True if this property is transitive
		 */

		public bool Transitive
		{
			get
			{
				return HasRDFType( GetVocabulary().TransitiveProperty )  ||
					DAMLHierarchy.GetInstance().IsTransitiveProperty( this );
			}
			set
			{
				if (value) 
				{
					// add the transitive type to this property
					SetRDFType( GetVocabulary().TransitiveProperty, false );
				}
				else 
				{
					// remove the transitive type from this property
					RemoveProperty( BBN.RDF.Vocabulary.RDF.type, GetVocabulary().TransitiveProperty );
				}
			}

		}


		/**
		 * Answer true if this property is to be considered unabiguous, that is
		 * it is characterised by the DAML class UnambiguousProperty
		 *
		 * @return True if this property is unambiguous
		 */
		public bool Unambiguous 
		{
			get
			{
				return HasRDFType( GetVocabulary().UnambiguousProperty );
			}
			set
			{
				if (value) 
				{
					// add the transitive type to this property
					SetRDFType( GetVocabulary().UnambiguousProperty, false );
				}
				else 
				{
					// remove the transitive type from this property
					RemoveProperty( BBN.RDF.Vocabulary.RDF.type, GetVocabulary().UnambiguousProperty );
				}
			}

		}


		/**
		 * Property accessor for the 'inverseOf' property of a DAML Property. This denotes
		 * that the named property (say, P) is an inverse of this property (say, Q). Formally,
		 * if (x, y) is an instance of P, then (y, x) is an instance of Q.
		 *
		 * @return Property accessor for 'inverseOf'
		 */
		public PropertyAccessor Prop_InverseOf
		{
			get 
			{
				if (m_propInverseOf == null) 
				{
					m_propInverseOf = new PropertyAccessorImpl( GetVocabulary().inverseOf, this );
				}

				return m_propInverseOf;
			}
		}


		/**
		 * Answer a key that can be used to index collections of this DAML property for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public override Object GetKey() 
		{
			return DAML_OIL._Property.GetURI();
		}



		/**
		 * This is a Jena internal method.  I need to override the one in ResourceImpl
		 * to prevent a ClassCastException in some circumstances.
		 *
		 * @param m The model to port to
		 * @return The ported resource
		 * @throws RDFException if an RDF processing error ocurrs.
		 */
		public override Resource Port( Model m )
		{
			if ( GetModel() == m ) 
			{
				return this;
			}
			else 
			{
				if (m is DAMLModel) 
				{
					// the model is a DAML model, so we create a new DAML property
					DAMLObjectPropertyImpl p = new DAMLObjectPropertyImpl( GetNameSpace(), GetLocalName(), ((DAMLModel) m), GetVocabulary() );
					p.id = id;
					return p;
				}
				else 
				{
					// the model is not a DAML model, so return a generic property
					return m.GetProperty( GetURI() );
				}
			}
		}



	}
}
