using System;
using System.Collections;
using BBN.RDF;
using BBN.DAML.Vocabulary;

namespace BBN.DAML
{
	
	public interface PropertyAccessor
	{
		Property GetProperty();
		int Count{get;}
		IEnumerator GetAll(bool closed);
		RDFNode Get();
		DAMLCommon GetDAMLValue();  
		void Add(Resource v);
		void Remove(DAMLCommon v);
		bool HasValue(DAMLCommon v);
	}// end PropertyAccessor
	

	
	public interface LiteralAccessor : PropertyAccessor
	{
		NodeEnumerator GetValues();
		Literal GetValue();
		void AddValue(string v);
		void AddValue(Literal v);
		void RemoveValue(string v);
		void RemoveValue(Literal v);
		bool HasValue(string v);
		bool HasValue(Literal v);
	}// end LiteralAccessor
		

	
	/// <summary>
	/// Encapsulates the standard methods of modifying a property on a DAML object, where
	/// the value of the property i an RDF literal as opposed to another DAML value.
	/// </summary>
	public interface IntLiteralAccessor : LiteralAccessor
	{
		int GetInt();
		void AddInt(int v);
		void RemoveInt(int v);
		bool HasIntValue(int v);
	}// end IntLiteralAccessor
	

	/// <summary>
	/// Abstract superclass for all DAML resources (including properties).  Defines shared
	/// implementations and common services(ie. property manipulation, vocab management, and
	/// <code>rdf:type</code> management).  Also provides accessors to common properties(ie.
	/// comment, label, and equivalentTo).
	/// </summary>

	public interface DAMLCommon : Resource
	{
		DAMLModel GetDAMLModel();
		void SetRDFType(Resource rdfClass);
		void SetRDFType(Resource rdfClass, bool replace);
		bool HasRDFType(string classURI);
		bool HasRDFType(Resource damlClass);
		IEnumerator GetRDFTypes(bool closed);
		RDFNode GetPropertyValue(Property p);
		NodeEnumerator GetPropertyValues(Property p);
		void SetPropertyValue(Property p, RDFNode v);
		void RemoveProperty(Property p, RDFNode v);
		void RemoveAll(Property p);
		void ReplaceProperty(Property p, RDFNode v);
		int GetNumPropertyValues(Property p);
		IEnumerator GetAll(Property p, bool closed);
		DAMLVocabulary GetVocabulary();
		IEnumerator GetEquivalentValues();
		IEnumerator GetEquivalenceSet();
		void Remove();
		LiteralAccessor Prop_Label{get;}
		LiteralAccessor Prop_Comment{get;}
		PropertyAccessor Prop_EquivalentTo{get;}
		PropertyAccessor Prop_Type{get;}
	}// end DAMLCommon



	
	/// <summary>
	/// Represents a DAML Ontology
	/// </summary>
	public interface DAMLOntology : DAMLCommon
	{
		LiteralAccessor Prop_VersionInfo{get;}
		NodeEnumerator GetImportedOntologies();
		void AddImportedOntology(string uri);
	}// end DAMLOntology
	

	/// <summary>
	/// DAMLDatatype represents a set of values from a concrete domain, as opposed to a class
	/// which is a set of values from the instance domain.
	/// </summary>
	public interface DAMLDatatype : DAMLCommon
	{
		DatatypeTranslator GetTranslator();
	}// end DAMLDatatype
	

	
	/// <summary>
	/// Represents a DAML Instance
	/// </summary>
	public interface DAMLInstance : DAMLCommon
	{
		PropertyAccessor Prop_SameIndividualAs{get;}
		IEnumerator GetSameInstances();
		// inherited IEnumerator GetEquivalentValues();
		PropertyAccessor AccessProperty(Property p);
	}// end DAMLInstance
	

	
	public interface Datatype : DAMLCommon
	{
		//DatatypeTranslator GetTranslator();
	}// end Datatype


	
	/// <summary>
	/// Represents a daml:Property.
	/// </summary>
	public interface DAMLProperty : DAMLCommon, Property
	{
		bool Unique{get; set;}
		PropertyAccessor Prop_Domain{get;}
		PropertyAccessor Prop_SubPropertyOf{get;}
		PropertyAccessor Prop_SamePropertyAs{get;}
		PropertyAccessor Prop_Range{get;}
		IEnumerator GetSameProperties();
		// inherited - IEnumerator GetEquivalentValues();
		IEnumerator GetDomainClasses();
		IEnumerator GetRangeClasses();
		IEnumerator GetSuperProperties();
		IEnumerator GetSuperProperties(bool closed);
		IEnumerator GetSubProperties();
		IEnumerator GetSubProperties(bool closed);
	}// end DAMLProperty
	

	
	/// <summary>
	/// Datatype property of a DAML ontology
	/// </summary>
	public interface DAMLDatatypeProperty : DAMLProperty
	{
		// Removed in Jena 1.4.0 LiteralAccessor Prop_DatatypeRange{get;}
	}// end DAMLDatatypeProperty
	

	
	/// <summary>
	/// Encapsulation of an object property in a DAML ontology.
	/// </summary>
	public interface DAMLObjectProperty : DAMLProperty
	{
		bool Transitive{get; set;}
		bool Unambiguous{get; set;}
		PropertyAccessor Prop_InverseOf{get;}
	}// end DAMLObjectProperty
	


	
	/// <summary>
	/// Encapsulates the capability of storing and retrieving DAML ontology information
	/// from the underlying storage or persistence service.
	/// </summary>
	public interface DAMLModel : Model
	{
		DAMLOntology CreateDAMLOntology(string uri);
		DAMLInstance CreateDAMLInstance(DAMLClass damlClass, string uri);
		DAMLClass CreateDAMLClass(string uri);
		DAMLProperty CreateDAMLProperty(string uri);
		DAMLObjectProperty CreateDAMLObjectProperty(string uri);
		DAMLDatatypeProperty CreateDAMLDatatypeProperty(string uri);
		DAMLDatatype CreateDAMLDatatype(string uri);
		DAMLList CreateDAMLList(string uri);
		DAMLRestriction CreateDAMLRestriction(string uri);
		DAMLCommon CreateDAMLValue(string uri, Resource damlClass, DAMLVocabulary vocab);
		DAMLCommon GetDAMLValue(string uri);
		DAMLCommon GetDAMLValue(string uri, DAMLClass damlClass, DAMLVocabulary vocab);
		IEnumerator ListDAMLClasses();
		IEnumerator ListDAMLProperties();
		IEnumerator ListDAMLInstances();
		DAMLLoader GetLoader();
		bool GetLoadSuccessful();
		XMLDatatypeRegistry GetDatatypeRegistry();
		Model Read(string uri, string b, string lang);
		// The following Read methods are inherited
		//Model Read(System.IO.TextReader reader, string b);
		//Model Read(System.IO.TextReader reader, string b, string lang);
		//Model Read(System.IO.Stream stream, string b);
		//Model Read(System.IO.Stream stream, string b, string lang);
		//Model Read(string url);
		//Model Read(string file, string b);
		//Model Read(WebRequest request);
		//Model ReadFromString(string xml, string b);
		//bool _cleanLoad{get; set;}
		//ProtocolErrorHandler ProtocolErrorHandler{get; set;}
		bool UseEquivalence{get; set;}
	}// end DAMLModel
	

	
	/// <summary>
	/// Unifies different forms of class expression as one type.
	/// </summary>
	public interface DAMLClassExpression
	{
		bool IsEnumeration{get;}
		bool IsNamedClass{get;}
		bool IsRestriction{get;}
		bool IsIntersection{get;}
		bool IsUnion{get;}
		bool IsDisjointUnion{get;}
		bool IsComplement{get;}
	}// end DAMLClassExpression
	


	
	/// <summary>
	/// Represents a daml:Class object.
	/// </summary>
	
	public interface DAMLClass : DAMLCommon, DAMLClassExpression
	{
		PropertyAccessor Prop_SubClassOf{get;}
		PropertyAccessor Prop_DisjointWith{get;}
		PropertyAccessor Prop_DisjointUnionOf{get;}
		PropertyAccessor Prop_SameClassAs{get;}
		PropertyAccessor Prop_UnionOf{get;}
		PropertyAccessor Prop_IntersectionOf{get;}
		PropertyAccessor Prop_ComplementOf{get;}
		PropertyAccessor Prop_OneOf{get;}
		IEnumerator GetSubClasses();
		IEnumerator GetSubClasses(bool closed);
		IEnumerator GetSuperClasses();
		IEnumerator GetSuperClasses(bool closed);
		IEnumerator GetSameClasses();
		// inherited IEnumerator GetEquivalentValues();
		bool HasSubClass(Resource cls);
		bool HasSuperClass(Resource cls);
		IEnumerator GetInstances();
		IEnumerator GetDefinedProperties();
		IEnumerator GetDefinedProperties(bool closed);
	}// end DAMLClass
	


		
	/// <summary>
	/// Reprents a daml:Restriction.
	/// </summary>
	public interface DAMLRestriction : DAMLClass, DAMLClassExpression
	{
		PropertyAccessor Prop_OnProperty{get;}
		PropertyAccessor Prop_ToClass{get;}
		PropertyAccessor Prop_HasValue{get;}
		PropertyAccessor Prop_HasClass{get;}
		PropertyAccessor Prop_HasClassQ{get;}
		IntLiteralAccessor Prop_Cardinality{get;}
		IntLiteralAccessor Prop_MinCardinality{get;}
		IntLiteralAccessor Prop_MaxCardinality{get;}
		IntLiteralAccessor Prop_CardinalityQ{get;}
		IntLiteralAccessor Prop_MinCardinalityQ{get;}
		IntLiteralAccessor Prop_MaxCardinalityQ{get;}
	}// end DAMLRestriction
	


	/// <summary>
	/// Represents an instantiation of a DAML datatype
	/// </summary>
	
	public interface DAMLDataInstance : DAMLCommon
	{
		PropertyAccessor Prop_SameIndividualAs{get;}
		IEnumerator GetSameInstances();
		DatatypeTranslator GetTranslator();
		object Value{get; set;}
	}// end DAMLDataInstance
	


	/// <summary>
	/// Representation of a DAML List (ordered version of the daml:Collection)
	/// </summary>
	
	public interface DAMLList : DAMLCommon
	{
		void Add(DAMLCommon v);
		void Remove(DAMLCommon v);
		IEnumerator GetAll();
		DAMLCommon GetFirst();
		DAMLList Cons(DAMLCommon v);
		DAMLList GetRest();
		int Count{get;}
		bool IsEmpty();
		void SetFirst(DAMLCommon v);
		void SetRest(DAMLList tail);
		void SetRestNil();
		DAMLList GetNil();
		bool IsNil(Resource r);
		DAMLList FindLast();
		// this sets up the indexer.  The client can do damlList[4] = ...
		DAMLCommon this[int index]{get; set;}
	}// end DAMLList


	/// <summary>
	/// Representation of a DAML Disjoint instance. Note that Disjoint has been removed from the DAML spec, so this is really obsolete.
	/// </summary>
	public interface DAMLDisjoint : DAMLCommon
	{ 
	}

	public interface DatatypeTranslator
	{
		Object Deserialize(RDFNode node);
		RDFNode Serialize(Object val, DAMLModel model);
	}
	

}
