using System;
using BBN.DAML.Vocabulary;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class DAMLOntologyImpl : DAMLCommonImpl, DAMLOntology
	{


		// Instance variables
		//////////////////////////////////

		/** Literal value accessor for the version info property */
		private LiteralAccessor m_propVersionInfo = null;



		// Constructors
		//////////////////////////////////

		/**
		 * Constructor, takes the URI for this ontology properties, and the underlying
		 * model it will be attached to.
		 *
		 * @param uri The URI of the ontology
		 * @param store The RDF store that contains the RDF statements defining the properties of the ontology
		 * @param vocabulary Reference to the DAML vocabulary used by this ontology resource.
		 */
		public DAMLOntologyImpl( string uri, DAMLModel store, DAMLVocabulary vocabulary )
			:base( uri, store, vocabulary )
		{
			SetRDFType( GetVocabulary().Ontology );
		}


		/**
		 * Constructor, takes the name and namespace for this ontology properties, and the underlying
		 * model it will be attached to.
		 *
		 * @param ns The namespace the ontology inhabits, or null
		 * @param name The name of the ontology
		 * @param store The RDF store that contains the RDF statements defining the properties of the ontology
		 * @param vocabulary Reference to the DAML vocabulary used by this ontology resource.
		 */
		public DAMLOntologyImpl( string ns, string name, DAMLModel store, DAMLVocabulary vocabulary )
			:base( ns, name, store, vocabulary )
		{
			SetRDFType( GetVocabulary().Ontology );
		}



		// External signature methods
		//////////////////////////////////


		/**
		 * Property value accessor for the version info property of the ontology.
		 *
		 * @return A literal accessor that gives access to the version info of the ontology.
		 */
		public LiteralAccessor Prop_VersionInfo
		{ 
			get
			{
				if (m_propVersionInfo == null) 
				{
					m_propVersionInfo = new LiteralAccessorImpl( GetVocabulary().versionInfo, this );
				}

				return m_propVersionInfo;
			}
		}


		/**
		 * Answer an iteration of resources that represent the URI's of the
		 * ontologies that this ontology imports.
		 *
		 * @return An iterator over the resources representing imported ontologies
		 */

		public NodeEnumerator GetImportedOntologies() 
		{
			return new ConcatenatedNodeEnumerator( GetPropertyValues( DAML_OIL._imports ),
				GetPropertyValues( DAML_OIL_2000_12._imports ) );
		}



		/**
		 * Add the given ontology to the list of ontologies managed by the
		 * knowledge store, and add it as an imoport property to this ontology object.
		 *
		 * @param uri The URI of the model.
		 */

		public void AddImportedOntology( string uri ) 
		{
			try 
			{
				// assert the model we're importing
				AddProperty( GetVocabulary().imports, GetModel().CreateLiteral( uri ) );

				// notify the knowledge store about the included model
				if (GetDAMLModel().GetLoader().GetLoadImportedOntologies()) 
				{
					GetModel().Read( uri );
				}
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception while adding imports property: " + e);
			}
		}



		/**
		 * Answer a key that can be used to index collections of this DAML instance for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public override Object GetKey() 
		{
			return DAML_OIL._Ontology.GetURI();
		}


	}
}
