using System;
using System.Collections;
using BBN.RDF;
using BBN.DAML.Vocabulary;

namespace BBN.DAML
{
	public class DAMLPropertyImpl : DAMLCommonImpl, DAMLProperty
	{

		// Instance variables
		//////////////////////////////////

		/** Property accessor for domain */
		private PropertyAccessor m_propDomain = null;

		/** Property accessor for range */
		private PropertyAccessor m_propRange = null;

		/** Property accessor for subPropertyOf */
		private PropertyAccessor m_propSubPropertyOf = null;

		/** Property accessor for samePropertyAs */
		private PropertyAccessor m_propSamePropertyAs = null;



		// Constructors
		//////////////////////////////////

		/// <summary>
		/// Constructor, takes the name and namespace for this property, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="ns">The namespace the property inhabits, or null</param>
		/// <param name="name">The name of the property</param>
		/// <param name="store">The RDF store that contains the RDF statements defining the properties of the property</param>
		/// <param name="vocabulary">Reference to the DAML vocabulary used by this property.</param>
		public DAMLPropertyImpl( string ns, string name, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( ns, name, store, vocabulary )
		{
			SetRDFType( GetVocabulary().Property );
		}


		/// <summary>
		/// Constructor, takes the URI for this property, and the underlying
		/// model it will be attached to.
		/// </summary>
		/// <param name="uri">The URI of the property</param>
		/// <param name="store">The RDF store that contains the RDF statements defining the properties of the property</param>
		/// <param name="vocabulary">Reference to the DAML vocabulary used by this property.</param>
		public DAMLPropertyImpl( string uri, DAMLModel store, DAMLVocabulary vocabulary ) 
			: base( uri, store, vocabulary )
		{
			SetRDFType( GetVocabulary().Property );
		}


		// External signature methods
		//////////////////////////////////


		/**
		 * Return a readable representation of the DAML value
		 *
		 * @return a string denoting this value
		 */
		public override string ToString() 
		{
			// get the public name for this value type (e.g. we change DAMLClassImpl -> DAMLClass)
			string cName = GetType().Name;
			int i = cName.IndexOf( "Impl" );
			if (i > 0) cName = cName.Substring( 0, i );

			// property attributes
			string attribs;
			attribs = Unique ? "unique " : "";

			// now format the return string
			return (GetURI() == null) ?
				("<" + attribs + "Anonymous " + cName + "@" + GetHashCode().ToString("X")  + ">") :
				("<" + attribs + cName + " " + GetURI() + ">");
		}


		/**
		 * Set the flag to indicate that this property is to be considered
		 * unique - that is, it is defined by the DAML class UniqueProperty.
		 *
		 * @param unique True for a unique property
		 */
		public bool Unique
		{
			set
			{
				if (value) 
				{
					// add the transitive type to this property
					SetRDFType( GetVocabulary().UniqueProperty, false );
				}
				else 
				{
					// remove the transitive type from this property
					RemoveProperty( BBN.RDF.Vocabulary.RDF.type, GetVocabulary().UniqueProperty );
				}
			}
			get
			{
				return HasRDFType( GetVocabulary().UniqueProperty );
			}
		}


		/**
		 * Property accessor for the 'domain' property of a property. This
		 * denotes the class that is the domain of the relation denoted by
		 * the property.
		 *
		 * @return Property accessor for 'domain'.
		 */
		public PropertyAccessor Prop_Domain 
		{
			get
			{
				if (m_propDomain == null) 
				{
					m_propDomain = new PropertyAccessorImpl( GetVocabulary().domain, this );
				}

				return m_propDomain;
			}
		}


		/**
		 * Property accessor for the 'subPropertyOf' property of a property. This
		 * denotes the property that is the super-property of this property
		 *
		 * @return Property accessor for 'subPropertyOf'.
		 */
		public PropertyAccessor Prop_SubPropertyOf
		{
			get
			{
				if (m_propSubPropertyOf == null) 
				{
					m_propSubPropertyOf = new PropertyAccessorImpl( GetVocabulary().subPropertyOf, this );
				}

				return m_propSubPropertyOf;
			}
		}


		/**
		 * Property accessor for the 'samePropertyAs' property of a property. This
		 * denotes that two properties should be considered equivalent.
		 *
		 * @return Property accessor for 'samePropertyAs'.
		 */
		public PropertyAccessor Prop_SamePropertyAs
		{
			get
			{
				if (m_propSamePropertyAs == null) 
				{
					m_propSamePropertyAs = new PropertyAccessorImpl( GetVocabulary().samePropertyAs, this );
				}

				return m_propSamePropertyAs;
			}
		}


		/**
		 * Property accessor for the 'range' property of a property. This
		 * denotes the class that is the range of the relation denoted by
		 * the property.
		 *
		 * @return Property accessor for 'range'.
		 */
		public PropertyAccessor Prop_Range
		{
			get
			{
				if (m_propRange == null) 
				{
					m_propRange = new PropertyAccessorImpl( GetVocabulary().range, this );
				}

				return m_propRange;
			}
		}


		/**
		 * Answer an iterator over all of the DAML properties that are equivalent to this
		 * value under the <code>daml:samePropertyAs</code> relation.  Note: only considers
		 * <code>daml:samePropertyAs</code>, for general equivalence, see
		 * {@link #getEquivalentValues}.  Note that the first member of the iteration is
		 * always the DAMLProperty on which the method is invoked: trivially, a property is
		 * a member of the set of properties equivalent to itself.  If the caller wants
		 * the set of properties equivalent to this one, not including itself, simply ignore
		 * the first element of the iteration.
		 *
		 * @return an iterator ranging over every equivalent DAML property - each value of
		 *         the iteration will be a DAMLProperty object.
		 */
		public IEnumerator GetSameProperties() 
		{
			return new PropertyEnumerator( this, GetVocabulary().samePropertyAs, GetVocabulary().samePropertyAs, true, true, true );
		}


		/**
		 * Answer an iterator over all of the DAML objects that are equivalent to this
		 * property, which will be the union of <code>daml:equivalentTo</code> and
		 * <code>daml:samePropertyAs</code>.
		 *
		 * @return an iterator ranging over every equivalent DAML class - each value of
		 *         the iteration should be a DAMLProperty object.
		 */
		public override IEnumerator GetEquivalentValues() 
		{
			ConcatenatedEnumerator i = new ConcatenatedEnumerator(
				// first the iterator over the equivalentTo values
				base.GetEquivalentValues(),
				// followed by the samePropertyAs values
				new PropertyEnumerator( this, GetVocabulary().samePropertyAs, GetVocabulary().samePropertyAs, true, false, false ) );

			// ensure that the iteration includes self
			i.SetDefaultValue( this );

			return i;
		}


		/**
		 * Answer an iterator over all of the super-properties of this property, using the
		 * <code>rdfs:subPropertyOf</code> relation (or one of its aliases).   The set of super-properties
		 * is transitively closed over the subPropertyOf relation.
		 *
		 * @return An iterator over the super-properties of this property,
		 *         whose values will be DAMLProperties.
		 */
		public IEnumerator GetSuperProperties() 
		{
			return GetSuperProperties( true );
		}


		/**
		 * Answer an iterator over all of the super-properties of this property, using the
		 * <code>rdfs:subPropertyOf</code> relation (or one of its aliases).   The set of super-properties
		 * is optionally transitively closed over the subPropertyOf relation.
		 *
		 * @param closed If true, iterate over the super-properties of my super-properties, etc.
		 * @return An iterator over the super-properties of this property,
		 *         whose values will be DAMLProperties.
		 */
		public IEnumerator GetSuperProperties( bool closed ) 
		{
			return new PropertyEnumerator( this, GetVocabulary().subPropertyOf, null, closed, false, true );
		}


		/**
		 * Answer an iterator over all of the sub-properties of this property, using the
		 * <code>rdfs:subPropertyOf</code> relation (or one of its aliases).   The set of sub-properties
		 * is transitively closed over the subPropertyOf relation.
		 *
		 * @return An iterator over the sub-properties of this property,
		 *         whose values will be DAMLProperties.
		 */
		public IEnumerator GetSubProperties() 
		{
			return GetSubProperties( true );
		}


		/**
		 * Answer an iterator over all of the sub-properties of this property, using the
		 * <code>rdfs:subPropertyOf</code> relation (or one of its aliases).   The set of sub-properties
		 * is optionally transitively closed over the subPropertyOf relation.
		 *
		 * @param closed If true, iterate over the sub-properties of my sub-properties, etc.
		 * @return An iterator over the sub-properties of this property,
		 *         whose values will be DAMLProperties.
		 */
		public IEnumerator GetSubProperties( bool closed ) 
		{
			return new PropertyEnumerator( this, null, GetVocabulary().subPropertyOf, closed, false, true );
		}


		/**
		 * Answer an iterator over all of the DAML classes that form the domain of this
		 * property.  The actual domain of the relation denoted by this property is the
		 * conjunction of all of the classes mention by the RDFS:domain property of this
		 * DAML property and all of its super-properties.
		 *
		 * @return an iterator whose values will be the DAML classes that define the domain
		 *         of the relation
		 */
		public IEnumerator GetDomainClasses() 
		{
			// first we want an iteration of this property and all of its super-properties
			IEnumerator i = new ConcatenatedEnumerator(
				GetSelfEnumerator(),
				GetSuperProperties() );

			// from this starting point, get all domain values
			return new PropertyEnumerator( i, GetVocabulary().domain, null, false, false, true );
		}


		/**
		 * Answer an iterator over all of the DAML classes that form the range of this
		 * property.  The actual range of the relation denoted by this property is the
		 * conjunction of all of the classes mention by the RDFS:range property of this
		 * DAML property and all of its super-properties.
		 *
		 * @return an iterator whose values will be the DAML classes that define the range
		 *         of the relation
		 */
		public IEnumerator GetRangeClasses() 
		{
			// first we want an iteration of this property and all of its super-properties
			IEnumerator i = new ConcatenatedEnumerator(
				GetSelfEnumerator(),
				GetSuperProperties() );

			// from this starting point, Get all domain values
			return new PropertyEnumerator( i, GetVocabulary().range, null, false, false, true );
		}



		/// <summary>
		/// Answers the ordinal value of a containment property.  Since DAML properties
		/// are not used as containment properties, this method always returns 0. See
		/// com.hp.hpl.mesa.rdf.jena.model.Property#getOrdinal for
		/// more details.
		/// </summary>
		/// <returns>zero</returns>
		public int GetOrdinal() 
		{
			return 0;
		}

		
		/**
		 * Answer a key that can be used to index collections of this DAML property for
		 * easy access by iterators.  Package access only.
		 *
		 * @return a key object.
		 */
		public override Object GetKey() 
		{
			return DAML_OIL._Property.GetURI();
		}



		/**
		 * This is a Jena internal method.  I need to override the one in ResourceImpl
		 * to prevent a ClassCastException in some circumstances.
		 *
		 * @param m The model to port to
		 * @return The ported resource
		 */
		public override Resource Port( Model m )
		{
			if ( GetModel() == m ) 
			{
				return this;
			}
			else 
			{
				if (m is DAMLModel) 
				{
					// the model is a DAML model, so we create a new DAML property
					DAMLPropertyImpl p = new DAMLPropertyImpl( GetNameSpace(), GetLocalName(), ((DAMLModel) m), GetVocabulary() );
					p.id = id;
					return p;
				}
				else 
				{
					// the model is not a DAML model, so return a generic property
					return m.GetProperty( GetURI() );
				}
			}
		}



	}

}
