using System;
using System.Collections;
using BBN.RDF.Common;

namespace BBN.RDF.Mem
{
	/// <summary>
	/// Implementation of Dynamic
	/// </summary>
#if DYNAMIC
	public class DynamicModelMem : ModelMem, DynamicModel, SourceModel
	{
		///
		/// URLs that have been read into this model
		///
		protected ArrayList urls = new ArrayList();

		///
		/// should we dynamically load URIs (not thread safe)
		///
		public static bool dynamic = true;


		public DynamicModelMem() : base()
		{
			store = new DynamicStoreMem(this);
		}

		public DynamicModelMem(Source source) : base()
		{
			store = new DynamicStoreMem(source, this);
		}

		///
		///
		///
		/**
	 * Enable or disable dynamic loading.
	 * The dynamicStore must have a model in order to enable dynamic loading.
	 * @returns the resulting value of dynamicLoad.
	 */
		public void DynamicLoad(bool dynamicp)
		{
			dynamic = dynamicp;
		}

		/**
		 * Returns true if dynamic loading is enabled.
		 */
		public bool DynamicLoad()
		{
			return dynamic;
		}


		/**
		 * List which URLs have been loaded.
		 * This is a cloned copy of the actual set of URLs.
		 */

		public ArrayList GetLoadedURLs()
		{
			return (ArrayList)urls.Clone();
		}

		/**
		 * Returns true if the given URL has been loaded into the model.
		 */
		public bool IsLoaded(String url)
		{ 
			return urls.Contains(url);
		}


		public override Model Read(String url) 
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource;
			Source newSource = sStore.GetSource(url);
			try 
			{
				sStore.DefaultSource = newSource;
				base.Read(url);
				urls.Add(url);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}

		
		public Model Read(string url, Source source) 
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource;
			try 
			{
				sStore.DefaultSource = source;
				base.Read(url);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}


		public Model Read(System.IO.TextReader reader, String b, Source source) 
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource;
			try 
			{
				sStore.DefaultSource = source;
				base.Read(reader, b);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}

		public Model Read(System.IO.Stream stream, String b, Source source) 
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource;
			try 
			{
				sStore.DefaultSource = source;
				base.Read(stream, b);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}

		public Model Read(System.IO.TextReader reader, string b, string lang, Source source)
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource ;
			try 
			{
				sStore.DefaultSource = source;
				base.Read(reader, b, lang);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}

		public Model Read(System.IO.Stream stream, string b, string lang, Source source)
		{
			SourceStore sStore = (SourceStore)store;
			Source originalSource = sStore.DefaultSource;
			try 
			{
				sStore.DefaultSource = source;
				base.Read(stream, b, lang);
				return this;
			} 
			finally 
			{
				sStore.DefaultSource = originalSource;
			}
		}

		///
		/// Methods for SourceModel
		/// 

		/**
		 * Add all the statements in another model to this model and associate with a source.
		 */
		public Model Add(Model m, Source source)
		{
			return Add(m.ListStatements(), source);
		}

		/**
		 * Add a statement to this model and associate with a source.
		 */
		public Model Add(Statement s, Source source)
		{
			if (! (s is StatementImpl) ||
				(((StatementImpl)s).GetModel() != this) ) 
			{
				s = new StatementImpl(s.GetSubject(),
					s.GetPredicate(),
					s.GetObject(),
					this);
			} 
			((SourceStore)store).Add(s, source);
			return this;
		}

		/**
		 * Add all the statements in an interator to this model and associate with a source.
		 */
		public Model Add(StmtEnumerator iter, Source source)
		{
			try 
			{
				while (iter.MoveNext()) 
				{
					Add(iter.Current, source);
				}
				return this;
			} 
			finally 
			{
				iter.Close();
			}
		}

		/**
		 * List all sources with statements in the model.
		 */
		public IEnumerator ListSources()
		{
			return ((SourceStore)store).ListSources(); 
		}

		public Source GetSource(String uri)
		{
			return ((SourceStore)store).GetSource(uri);
		}
    
		/**
		 * Return an iterator over all the statements associated with a source.
		 */
		public StmtEnumerator ListStatementsOfSource(Source source)
		{
			return new StmtEnumeratorImpl(((SourceStore)store).ListBySource(source));
		}

		/**
		 * Return a model that contains all the statements associated with a source.
		 * TBD - Allow user to specify what type of model to return.
		 */
		public Model GetModelForSource(Source source)
		{
			DynamicModelMem model = new DynamicModelMem();
			model.Add(ListStatementsOfSource(source), source);
			return (Model)model;
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual IEnumerator ListSources() 
		{
			return sourceToStatements.KeySet().GetEnumerator();
		}
		/**
		 * Change the source associated with a statement.
		 */
		public Model SetStatementSource(Statement s, Source source)
		{
			((SourceStore)store).SetStatementSource(s, source);
			return this;
		}

		/**
		 * Change the source associated with the statements in the iterator.
		 */
		public Model SetStatementSource(StmtEnumerator iter, Source source)
		{
			try 
			{
				while (iter.MoveNext()) 
				{
					((SourceStore)store).SetStatementSource(iter.Current, source);
				}
				return this;
			} 
			finally 
			{
				iter.Close();
			}
		}

		/**
		 * Write the model of statements associated with a source to an XML file.
		 */
		/*
		public Model Write(java.io.Writer writer, Source source) 
		{
			Model model = GetModelForSource(source);
			model.Write(writer);
			return this;
		}
		public Model Write(java.io.Writer writer, string lang, Source source)
		{
			Model model = GetModelForSource(source);
			model.Write(writer, lang);
			return this;
		}

		public Model Write(java.io.Writer writer, string lang, string b, Source source)
		{
			Model model = GetModelForSource(source);
			model.Write(writer, lang, b);
			return this;
		}
		*/

		// ================================================================
		// Override methods from ModelMem
		// ================================================================

		public override Resource CreateResource(string uri)
		{
			return base.CreateResource(uri);
		}


	}
#endif

}
