using System;
using BBN.RDF;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class LiteralAccessorImpl : PropertyAccessorImpl, LiteralAccessor

	{

		// Constructors
		//////////////////////////////////

		/// <summary>
		/// Construct a new accessor for literal values of the given property
		/// </summary>
		/// <param name="property">The property that this accessor works on</param>
		/// <param name="val">The DAML value that has this property</param>
		public LiteralAccessorImpl( Property property, DAMLCommon val ) : base(property, val)
		{
		}


		// External signature methods
		//////////////////////////////////

		/**
		 * Answer an iteration over the literal values that this property has in the
		 * RDF model.
		 *
		 * @return An iteration over the values of the encapsulated property. Each
		 *         member of the iteration will be an RDF literal.
		 */
		public NodeEnumerator GetValues() 
		{
			return m_val.GetPropertyValues( GetProperty() );
		}


		/**
		 * Answer the a value of the encapsulated property. If it has no values, answer
		 * null. If it has one value, answer that value. Otherwise, answer an undetermined
		 * member of the set of values.
		 *
		 * @return A value for the encapsulated property, or null
		 *         if the property has no value.
		 */
		public Literal GetValue() 
		{
			try 
			{
				NodeEnumerator i = GetValues();
				return (i == null  ||  !i.MoveNext()) ? null : ((Literal) i.Current);
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception when getting literal values: " + e );
				throw new System.Exception( "RDF exception when getting literal values: " + e );
			}
		}


		/**
		 * Add a value to the encapsulated property.
		 *
		 * @param value The value to be added, as a string.
		 */
		public void AddValue( String value ) 
		{
			try 
			{
				AddValue( m_val.GetModel().CreateLiteral( value ) );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "Saw RDF Exception while creating literal: " + e );
				throw new System.Exception( "Saw RDF Exception while creating literal: " + e );
			}
		}


		/**
		 * Add a value to the encapsulated property.
		 *
		 * @param value The value to be added, as an RDF literal.
		 */
		public void AddValue( Literal value ) 
		{
			try 
			{
				m_val.AddProperty( GetProperty(), value );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception " + e );
			}
		}


		/**
		 * Remove a value from the encapsulated property.
		 *
		 * @param value The value to be removed, as a string.
		 */
		public void RemoveValue( String value ) 
		{
			try 
			{
				RemoveValue( m_val.GetModel().CreateLiteral( value ) );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "Saw RDF Exception while creating literal: " + e );
				throw new System.Exception( "Saw RDF Exception while creating literal: " + e );
			}
		}


		/**
		 * Remove a value from the encapsulated property.
		 *
		 * @param value The value to be removed, as a literal.
		 */
		public void RemoveValue( Literal value ) 
		{
			m_val.RemoveProperty( GetProperty(), value );
		}


		/**
		 * Answer true if the encapsulated property has the given value as one of its
		 * values.
		 *
		 * @param value A DAML value to test for
		 * @return True if the RDF model contains a statement giving a value for
		 *         the encapsulated property matching the given value.
		 */
		public bool HasValue( String value ) 
		{
			try 
			{
				return HasValue( m_val.GetModel().CreateLiteral( value ) );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "Saw RDF Exception while creating literal: " + e );
				throw new System.Exception( "Saw RDF Exception while creating literal: " + e );
			}
		}


		/**
		 * Answer true if the encapsulated property has the given value as one of its
		 * values.
		 *
		 * @param value A DAML value to test for
		 * @return True if the RDF model contains a statement giving a value for
		 *         the encapsulated property matching the given value.
		 */
		public bool HasValue( Literal value ) 
		{
			try 
			{
				return m_val.HasProperty( GetProperty(), value );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF Exception " + e );
				throw new System.Exception( "RDF Exception " + e );
			}
		}

	}
}
