using System;
using System.Collections;
using BBN.RDF;
using BBN.RDF.Common;
using System.Runtime.CompilerServices; // for synchronizing methods
using System.Xml;
using System.Diagnostics;

namespace BBN.RDF.Mem
{
	/// <summary>
	/// Implementation for the Store class.
	/// </summary>
	public class StoreMem : StoreCom, Store
	{
		// bouncing back and forth on whether the statements collection should be a
		// hashtable or an arraylist.  Doing arraylist for now since I don't know what
		// I would key the objects with
		internal ArrayList statements = new ArrayList();
		internal NodeMap subjects = new NodeMap();
		internal NodeMap predicates = new NodeMap();
		internal NodeMap objects = new NodeMap();
		/// <summary>
		/// Indexes the statements by their source.
		/// </summary>
		internal SourceMap sources = new SourceMap();
		/// <summary>
		/// Stores all the sources in the Store.
		/// </summary>
		internal Hashtable sourceList = new Hashtable();

		public StoreMem()
		{}
		public override Resource GetResource(string uri, Model m)
		{
			Resource r = new ResourceImpl(uri, m);
			Statement s = subjects.GetFirst(r);
			if(s != null) return s.GetSubject();
			s = objects.GetFirst(r);
			if(s != null) return (Resource)s.GetObject();
			s = predicates.GetFirst(r);
			if(s != null) return s.GetPredicate();
			return r;
		}
		public override Property GetProperty(Property p)
		{
			Statement s = predicates.GetFirst(p);
			if(s != null) return s.GetPredicate();
			return p;
		}
		/// <summary>
		/// Get the source object for a uri.  Creates a new source in necessary.
		/// </summary>
		/// <param name="uri">The uri string of interest</param>
		/// <returns>The source object for the uri.</returns>
		public override Source GetSource(string uri)
		{
			if (sourceList.ContainsKey(uri))
				return sourceList[uri] as Source;
			else 
			{
				Source s = CreateSource(uri);
				sourceList[uri] = s;
				return s;
			}
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override void Add(Statement s)
		{
			if(statements.Contains(s)) return;
			statements.Add(s);
			subjects.Add(s.GetSubject(), s);
			predicates.Add(s.GetPredicate(), s);
			objects.Add(s.GetObject(), s);
			sources.Add(s.GetSource(), s);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override void Remove(Statement s)
		{
			statements.Remove(s);
			subjects.Remove(s.GetSubject(), s);
			predicates.Remove(s.GetPredicate(), s);
			objects.Remove(s.GetObject(), s);
			sources.Remove(s.GetSource(), s);
		}
		public override int Count
		{
			get
			{
				lock(this){return statements.Count;}
			}
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override bool Contains(Statement s)
		{
			return statements.Contains(s);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override bool Contains(Resource s, Property p)
		{
			System.Collections.IEnumerator iter = null;
			iter = ListBySubject(s);
			while(iter.MoveNext())
				if(((Statement)iter.Current).GetPredicate().Equals(p))
					return true;
			return false;
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator List()
		{
			return statements.GetEnumerator();
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator GetEnumerator()
		{
			return statements.GetEnumerator();
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator List(Resource s, Property p, RDFNode o)
		{
			if(s != null)
				return ListBySubject(s);
			if(o != null)
				return ListByObject(o);
			if(p != null)
				return ListByPredicate(p);
			return List();
		}
		//TROYTROYTROY
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator ListBySubject(Resource s)
		{
			if (s is Statement && ((Statement)s).GetReificationId() != null) 
			{
				return subjects.GetEnumerator(((Statement)s).GetReificationId());
			}
			else
				return subjects.GetEnumerator(s);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator ListByPredicate(Property p)
		{
			return predicates.GetEnumerator(p);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator ListByObject(RDFNode o)
		{
			if (o is Statement && ((Statement)o).GetReificationId() != null) 
				return objects.GetEnumerator(((Statement)o).GetReificationId());
			else
				return objects.GetEnumerator(o);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator ListBySource(Source s)
		{
			return sources.GetEnumerator(s);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override ResEnumerator ListSubjects()
		{
			// Hashtable h = new Hashtable();
			ArrayList h = new ArrayList();
			IEnumerator iter = subjects.ListFirstStatements();
			while(iter.MoveNext())
				h.Add(((Statement)iter.Current).GetSubject());
			return new ResEnumeratorImpl(h);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IEnumerator ListSources() 
		{
			return sources.KeySet().GetEnumerator();
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IList GetStatementsBySubject(Resource s)
		{
			if (s is Statement && ((Statement)s).GetReificationId() != null) 
				return subjects.GetStatements(((Statement)s).GetReificationId());
			else
				return subjects.GetStatements(s);	
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IList GetStatementsByPredicate(Property p)
		{
			return predicates.GetStatements(p);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override IList GetStatementsByObject(RDFNode o)
		{
			if (o is Statement && ((Statement)o).GetReificationId() != null) 
				return objects.GetStatements(((Statement)o).GetReificationId());
			else
				return objects.GetStatements(o);
		}
		[MethodImpl(MethodImplOptions.Synchronized)]
		public override NsEnumerator ListNameSpaces()
		{
			//Hashtable nameSpaces = new Hashtable();
			ArrayList nameSpaces = new ArrayList();
			IEnumerator iter = predicates.ListFirstStatements();
			while(iter.MoveNext())
				nameSpaces.Add(((Statement)iter.Current).GetPredicate().GetNameSpace());
			return new NsEnumeratorImpl(nameSpaces);
		}

	}// end StoreMem

	/// <summary>
	/// Considered having this extend the RdfNodeMap created for the RDF.NET package
	/// but then decided against it.  It would only inherit one method and is only used
	/// in the Store class so why bother with the overhead.
	/// </summary>
	public class NodeMap : IEnumerable
	{
		Hashtable map = new Hashtable();

		internal void Add(RDFNode o, Statement s)
		{
			ArrayList l = (ArrayList)map[o];
			if(l == null)
			{
				l = new ArrayList();
				map[o] = l;
			}
			l.Add(s);
		}
		internal void Remove(RDFNode o, Statement s)
		{
			ArrayList l = (ArrayList)map[o];
			if(l != null)
			{
				l.Remove(s);
				if(l.Count == 0)
					map.Remove(o);			}
		}
		public IEnumerator GetEnumerator()
		{
			// simply going to grab each ArrayList from the Hashtable, combine them into
			// one and return the Enumerator.  I may find out later that this is a bad idea.
			IDictionaryEnumerator e = map.GetEnumerator();
			ArrayList l = new ArrayList();
			while(e.MoveNext())
			{
				l.AddRange((ArrayList)e.Value);
			}
			return l.GetEnumerator();
		}
										  
		internal IEnumerator GetEnumerator(RDFNode o)
		{
			ArrayList l = (ArrayList)map[o];
			if(l == null)
				return (new ArrayList()).GetEnumerator();
			return l.GetEnumerator();
		}
		/*
			internal IEnumerator GetEnumerator(RDFNode o1, RDFNode o2)
			{
				ArrayList l = new ArrayList();
				ArrayList l1 = (ArrayList)map[o1];
				if (l1 != null) l.AddRange(l1);
				ArrayList l2 = (ArrayList)map[o2];	
				if (l2 != null) l.AddRange(l2);
				return l.GetEnumerator();				
			}
			*/
		internal Statement GetFirst(RDFNode o)
		{
			ArrayList l = (ArrayList)map[o];
			if(l == null)
				return null;
			return (Statement)l[0];
		}
		/// <summary>
		/// Get an iterator over the first statement of each value in the hashtable.
		/// </summary>
		/// <returns>An iterator over the first statement of each value in the hashtable.</returns>
		internal IEnumerator ListFirstStatements()
		{
			ArrayList stmts = new ArrayList();
			IEnumerator iter = map.Values.GetEnumerator();
			while(iter.MoveNext())
				stmts.Add(((ArrayList)iter.Current)[0]);
			return stmts.GetEnumerator();
		}
		internal IList GetStatements(RDFNode o)
		{
			return map[o] as IList;
		}	
		public ICollection GetKeys()
		{
			return map.Keys;
		}

	}
#if DYNAMIC
	/// <summary>
	///  Extends StoreMem by implementing dynamic loading and keeping
	///  track of the source of each statement.
	/// </summary>
	public class DynamicStoreMem : StoreMem, DynamicStore, SourceStore
	{
		private DynamicModel m_model;
		private Source m_source;

		// Mapping from a source to the statements associated with it.
		private SourceMap sourceToStatements = new SourceMap();

		// Mapping from a statement to it's source.
		// Do we want to store source in the statement object?
		private Hashtable statementToSource = new Hashtable();

		private Source defaultSource = null;

		private Hashtable mySources = new Hashtable();


		public DynamicStoreMem() : base()
		{
			m_model = null;
			m_source = null;
			//m_dynamicLoad = false;
		}
		public DynamicStoreMem(DynamicModel model) : base()
		{
			m_source = null;
			if (model == null) 
			{
				m_model = null;
				//m_dynamicLoad = false;
			} 
			else 
			{
				m_model = model;
				//m_dynamicLoad = true;
			}
		}
		public DynamicStoreMem(Source source) : base()
		{
			m_model = null;
			m_source = source;
			//m_dynamicLoad = false;			
		}
		public DynamicStoreMem(Source source, DynamicModel model) : base()
		{
			m_source = source;
			if (model == null) 
			{
				m_model = null;
				//m_dynamicLoad = false;
			} 
			else 
			{
				m_model = model;
				//m_dynamicLoad = true;
			}
		}

		public DynamicModel GetModel()
		{
			return m_model;
		}

		public void SetModel(DynamicModel model)
		{
			m_model = model;
			
		}

		// ================================================================
		// SourceStore Methods
		// ================================================================
		
		/**
		 * Add a statement associated with a source.
		 * TBD: What to do if statement already belongs to store, but with a different source?
		 */
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual void Add(Statement s, Source source) 
		{
			if (statements.Contains(s)) return;
			statementToSource[s] = source;
			sourceToStatements.Add(source, s);
			base.Add(s);
		}

	/**
	 * Return an iterator over all the statements with a given source.
	 */
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual IEnumerator ListBySource(Source source) 
		{
			return sourceToStatements.GetEnumerator(source);
		}

	/**
	 * Return the store of a statement.
	 */
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual Source GetStatementSource(Statement s) 
		{
			return (Source)statementToSource[s];
		}

	/**
	 * Return an iterator over all the sources in the store.
	 */
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual IEnumerator ListSources() 
		{
			return sourceToStatements.KeySet().GetEnumerator();
		}

	/**
	 * Set the default source that is used when adding a statement.
	 */
		public virtual Source DefaultSource
		{
			set 
			{
				defaultSource = value;
			}
			get
			{
				return defaultSource;
			}
		}

	/**
	 * Change the source of a statement.
	 */
		[MethodImpl(MethodImplOptions.Synchronized)]
		public virtual void SetStatementSource(Statement s, Source source)
		{
			if (statements.Contains(s)) 
			{
				//originalSource = GetStatementSource(s);
				statementToSource[s] = source;
			}
		}

	/**
	 * Create a new source.
	 */
		protected virtual Source CreateSource(String uri)
		{
			Source source = new SourceImpl(uri);
			mySources[uri] = source;
			return source;
		}

		/**
		 * Get a Source if it exists, otherwise create a new Source
		 */
		public virtual Source GetSource(String uri)
		{
			if (mySources[uri] == null) 
			{
				//Console.Error.WriteLine("Source does not exist");
				return CreateSource(uri);
			} 
			else 
			{
				//Console.Error.WriteLine("Source exists");	
				return (Source)mySources[uri];
			}
		}

		// ================================================================
		// Overridden methods
		// ================================================================

		
		// ================================================================
		// Implementation of methods specific to Source
		// ================================================================

		[MethodImpl(MethodImplOptions.Synchronized)]
		public override void Add(Statement s)
		{	
			Add(s, defaultSource);
		}

		[MethodImpl(MethodImplOptions.Synchronized)]
		public override void Remove(Statement s)
		{
			Source source = GetStatementSource(s);
			sourceToStatements.Remove(source, s);
			base.Remove(s);
		}


		public override Resource CreateResource(Model m)
		{
			return new DynamicResourceImpl(m);
		}
    
		public override Resource CreateResource(string uri, Model m) 
		{
			return new DynamicResourceImpl(uri, m);
		}
    
		public override Property CreateProperty(string uri, Model m)
		{
			return new DynamicPropertyImpl(uri, m);
		}
    
		public override Property CreateProperty(string ns, string localPart, Model m)
		{
			return new DynamicPropertyImpl(ns, localPart, m);
		}
    
		public override Resource GetResource(string uri, Model m)
		{
			Resource r = base.GetResource(uri, m);
			if (r is DynamicResourceImpl) 
			{
				return r;
			}
			else 
			{
				return new DynamicResourceImpl(uri, m);
			}
		}

		public override Property GetProperty(string uri, Model m)
		{
			return GetProperty(CreateProperty(uri, m));
		} 
    
		public override Property GetProperty(string ns, string localPart, Model m)
		{
			return GetProperty(CreateProperty(ns, localPart, m));
		}
    
		public override Property GetProperty(Property p) 
		{
			// check and see if we already have an object for this resource
			Statement s = predicates.GetFirst(p);
			if (s != null) return s.GetPredicate();
        
			// we don't so return the one we were passed
			return p;
		}
	}
#endif
	public class ModelMem : ModelCom, Model, ModelI
	{
		public ModelMem()
		{
			store = new StoreMem();
		}
		public override bool SupportsTransactions()
		{
			return false;
		}
		public override bool SupportsSetOperations()
		{
			return true;
		}

		public override Model Query(Selector selector)
		{
			ModelMem model = new ModelMem();
			StmtEnumerator iter = null;
			try 
			{
				iter = ListStatements(selector);
				while (iter.MoveNext()) 
				{
					model.Add(iter.Current);
				}
				return model;
			} 
			finally 
			{
				if (iter != null) 
				{
					iter.Close();
				}
			}
		}
		public override Model Union(Model model)
		{
			return ((new ModelMem()).Add(this)).Add(model);
		}
		public override Model Intersection(Model model)
		{
			Model largerModel = this;
			Model smallerModel = model;
			ModelMem resultModel = new ModelMem();
			StmtEnumerator e = null;
			Statement stmt;
			
			if(model.Count > this.Count)
			{
				largerModel = model;
				smallerModel = this;
			}
			try
			{
				e = smallerModel.ListStatements();
				while(e.MoveNext())
				{
					stmt = e.Current;
					if(largerModel.Contains(stmt))
						resultModel.Add(stmt);
				}
				return resultModel;
			}
			finally
			{
				e.Close();
			}
		}
		public override Model Difference(Model model)
		{
			ModelMem resultModel = new ModelMem();
			StmtEnumerator e = null;
			Statement stmt;
			try
			{
				e = ListStatements();
				while(e.MoveNext())
				{
					stmt = e.Current;
					if(!model.Contains(stmt))
						resultModel.Add(stmt);
				}
				return resultModel;
			}
			finally
			{
				e.Close();
			}
		}
		public override Model GetModelForSource(Source s)
		{
			ModelMem resultModel = new ModelMem();
			StmtEnumerator stmts = this.ListStatementsOfSource(s);
			resultModel.Add(stmts);
			return resultModel;
		}
	}// end ModelMem

	/// <summary>
	/// Index a set of statements based on their source.
	/// </summary>
	public class SourceMap 
	{
		/// <summary>
		/// Stores the statements based on their source.
		/// </summary>
		Hashtable map = new Hashtable();

		// Right now, doesn't keep track of statement when source is null
	
		/// <summary>
		/// Add a new source/statement pair.  If the source is null, the statment will not be stored.
		/// </summary>
		/// <param name="source">The source of the statement</param>
		/// <param name="s">The statement to add</param>
		public void Add(Source source, Statement s) 
		{
			if (source == null) 
			{
				return;
			} 
			ArrayList l = (ArrayList) map[source];
			if (l == null) 
			{
				l = new ArrayList();
				map[source] = l;
			}
			l.Add(s);
		}

		/// <summary>
		/// Remove a source/statement pair.
		/// </summary>
		/// <param name="source">The source of the statement</param>
		/// <param name="s">The statement to remove</param>
		public void Remove(Source source, Statement s) 
		{
			if (source == null) { return; }
			ArrayList l = (ArrayList) map[source];
			if (l != null) 
			{
				l.Remove(s);
				if (l.Count == 0) 
				{
					map.Remove(source);
				}
			}
		}
    
		/// <summary>
		/// Enumerator all the statements for a source.
		/// </summary>
		/// <param name="source">The source of interest</param>
		/// <returns>An IEnumerator over all the statements associated with the source.</returns>
		public IEnumerator GetEnumerator(Source source)
		{
			if (source == null) 
			{
				Trace.WriteLine("Null passed to GetEnumerator.");
				return (new ArrayList()).GetEnumerator();
			}
			ArrayList l = (ArrayList) map[source];
			if (l == null) 
			{
				Trace.WriteLine("No Statements for source");
				return (new ArrayList()).GetEnumerator();
			} 
			else 
			{
				Trace.WriteLine(l.Count + " statements for source");
				return l.GetEnumerator();
			}
		}

		/// <summary>
		/// Get the collection of stored sources.
		/// </summary>
		/// <returns>A ICollection of all Source objects in the SourceMap</returns>
		public ICollection KeySet()
		{
			return map.Keys;
		}

	}

}
