using System;
using System.Collections;
using BBN.RDF.Common;

namespace BBN.RDF
{
	public class AnonId
	{
		string id;
		
		public AnonId()
		{
			id = System.Guid.NewGuid().ToString("N");
		}
		public AnonId(string id)
		{
			this.id = id;
		}
		public override bool Equals(object o)
		{
			return (o is AnonId && id.Equals(((AnonId)o).id));
		}
		public override string ToString()
		{
			return id;
		}
		public override int GetHashCode()
		{
			return id.GetHashCode();
		}
	}// end AnonID
	/// <summary>
	/// Stores the common functions used by resource and literal nodes
	/// </summary>
	/// <author>Troy Self</author>
	public interface RDFNode
	{
		/// <summary>
		/// Returns the <code>string</code> representation of this node
		/// </summary>
		/// <returns>string representation of this node</returns>
		string ToString();
	}// end RDFNode

	/// <summary>
	///  RDF Resource node
	/// </summary>
	public interface Resource : RDFNode
	{
		/// <summary>
		/// Returns the URI of the resource or an empty string if it is anonymous
		/// </summary>
		/// <returns>URI of resource or empty string if anonymous</returns>
		string GetURI();

		/// <summary>
		/// Returns the namespace associated with this resource
		/// </summary>
		/// <returns>Namespace of this resource</returns>
		string GetNameSpace();

		/// <summary>
		/// Gets the name of this property within its namespace
		/// </summary>
		/// <returns>Local name of this property within its namespace</returns>
		string GetLocalName();

		Model GetModel();

		/// <summary>
		/// Returns the URI of the resource unless it is anonymous
		/// </summary>
		/// <returns>URI of the resource unless it is anonymous</returns>
		new string ToString();

		AnonId GetId();

		bool Equals(object o);

		bool IsAnon();

		Statement GetProperty(Property p);

		StmtEnumerator ListProperties(Property p);

		StmtEnumerator ListProperties();

		StmtEnumerator ListIncomingProperties(Property p);

		StmtEnumerator ListIncomingProperties();

		Resource AddProperty(Property p, bool o);

		Resource AddProperty(Property p, long o);

		Resource AddProperty(Property p, char o);

		Resource AddProperty(Property p, float o);

		Resource AddProperty(Property p, double o);

		Resource AddProperty(Property p, string o);

		Resource AddProperty(Property p, string o, string l);

		Resource AddProperty(Property p, object o);

		Resource AddProperty(Property p, RDFNode o);

		bool HasProperty(Property p);

		bool HasProperty(Property p, bool o);

		bool HasProperty(Property p, long o);

		bool HasProperty(Property p, char o);

		bool HasProperty(Property p, float o);

		bool HasProperty(Property p, double o);

		bool HasProperty(Property p, string o);

		bool HasProperty(Property p, string o, string l);

		bool HasProperty(Property p, object o);

		bool HasProperty(Property p, RDFNode o);

		Resource RemoveProperties();

		Resource Begin();

		Resource Abort();

		Resource Commit();
	}// end Resource

	/// <summary>
	/// A Statement is itself a Resource and can be both the subject or object of other
	/// Statement objects.
	/// </summary>
	public interface Statement : Resource
	{
		new bool Equals(object o);

		Resource GetSubject();

		Property GetPredicate();

		RDFNode GetObject();
		
		Source GetSource();

		new Statement GetProperty(Property p);

		Resource GetReificationId();

		Statement GetStatementProperty(Property p);

		Resource GetResource();

		Literal GetLiteral();

		bool GetBoolean();

		byte GetByte();

		short GetShort();

		int GetInteger();

		long GetLong();

		float GetFloat();

		char GetChar();

		double GetDouble();

		string GetString();

		Resource GetResource(ResourceF f);

		object GetObject(ObjectF f);

		//Bag GetBag();

		//Alt GetAlt();

		//Seq GetSeq();

		string GetLanguage();

		bool IsWellFormed();

		Statement Set(bool o);

		Statement Set(long o);

		Statement Set(char o);

		Statement Set(float o);

		Statement Set(double o);

		Statement Set(string o);

		Statement Set(string o, bool wellFormed);

		Statement Set(string o, string l);

		Statement Set(string o, string l, bool wellFormed);

		Statement Set(RDFNode o);

		Statement Set(object o);

		Statement Remove();

		bool IsReified();
	}// end Statement

	public interface Property : Resource
	{
		new string GetLocalName();

		new string GetNameSpace();

		int GetOrdinal();
	}// end Property

	public interface Literal : RDFNode
	{
		bool GetBoolean();

		byte GetByte();

		short GetShort();

		int GetInteger();

		long GetLong();

		char GetChar();

		float GetFloat();

		double GetDouble();

		string GetString();

		object GetObject(ObjectF f);

		string GetLanguage();

		bool IsWellFormed();

		bool Equals(object o);
	} // end Literal

	public interface ResourceF
	{
		Resource CreateResource(Resource r);
	}// end ResourceF

	public interface ObjectF
	{
		object CreateObject(string s);
	}// end ObjectF

	public interface Model : IEnumerable
	{
		Model Abort();
		Model Add(Model m);
		Model Add(Statement s);
		Model Add(StmtEnumerator s);
		Model Add(Resource s, Property p, RDFNode o);
		Model Add(Resource s, Property p, bool o);
		Model Add(Resource s, Property p, long o);
		Model Add(Resource s, Property p, char o);
		Model Add(Resource s, Property p, float o);
		Model Add(Resource s, Property p, double o);
		Model Add(Resource s, Property p, string o);
		Model Add(Resource s, Property p, string o, bool wellFormed);
		Model Add(Resource s, Property p, string o, string l);
		Model Add(Resource s, Property p, string o, string l, bool wellFormed);
		Model Add(Resource s, Property p, object o);
		
		Model Begin();

		void Close();

		Model Commit();

		bool Contains(Resource s, Property p);

		bool Contains(Resource s, Property p, RDFNode o);

		bool Contains(Resource s, Property p, bool o);

		bool Contains(Resource s, Property p, long o);

		bool Contains(Resource s, Property p, char o);

		bool Contains(Resource s, Property p, float o);

		bool Contains(Resource s, Property p, double o);

		bool Contains(Resource s, Property p, string o);

		bool Contains(Resource s, Property p, string o, string l);

		bool Contains(Resource s, Property p, object o);

		bool Contains(Statement s);

		bool ContainsAll(Model model);

		bool ContainsAll(StmtEnumerator iter);

		bool ContainsAny(StmtEnumerator iter);

		bool ContainsAny(Model model);

		Literal CreateLiteral(bool o);

		Literal CreateLiteral(long o);
		
		Literal CreateLiteral(char o);
		
		Literal CreateLiteral(float o);
		
		Literal CreateLiteral(double o);
		
		Literal CreateLiteral(object o);

		Literal CreateLiteral(string v);

		Literal CreateLiteral(string v, string language);

		Literal CreateLiteral(string v, string language, bool wellFormed);

		Property CreateProperty(string nameSpace, string localName);

		Resource CreateResource();

		Resource CreateResource(string uri);

		Statement CreateStatement(Resource s, Property p, RDFNode o);

		Statement CreateStatement(Resource s, Property p, RDFNode o, string uri);

		/// <summary>
		/// Create a new statement.
		/// </summary>
		Statement CreateStatement(Resource s, Property p, RDFNode o, string uri, string source);

		Model Difference(Model model);

		bool Equals(object model);

		Statement GetProperty(Resource s, Property p);

		Property GetProperty(string nameSpace, string localName);

		Property GetProperty(string uri);

		Resource GetResource(string uri);

		/// <summary>
		/// Get the source object for a uri.
		/// </summary>
		Source GetSource(string uri);

		/// <summary>
		/// Get a model containing all the statements associated with a source.
		/// </summary>
		Model GetModelForSource(Source s);

		bool Independent();

		Model Intersection(Model model);

		bool IsReified(Statement s);

		NsEnumerator ListNameSpaces();

		NodeEnumerator ListObjects();

		NodeEnumerator ListObjectsOfProperty(Property p);

		NodeEnumerator ListObjectsOfProperty(Resource s, Property p);

		StmtEnumerator ListReifiedStatements();

		StmtEnumerator ListStatements();

		StmtEnumerator ListStatements(Selector s);

		ResEnumerator ListSubjects();

		/// <summary>
		/// List all the sources in the model.
		/// </summary>
		IEnumerator ListSources();		
		
		/// <summary>
		/// List all the statements associated with a source.
		/// </summary>
		StmtEnumerator ListStatementsOfSource(Source s);

		ResEnumerator ListSubjectsWithProperty(Property p);

		ResEnumerator ListSubjectsWithProperty(Property p, RDFNode o);

		Model Query(Selector s);

		Model Read(System.IO.TextReader reader, string b);

		Model Read(System.IO.TextReader reader, string b, string lang);

		Model Read(System.IO.Stream stream, string b);

		Model Read(System.IO.Stream stream, string b, string lang);

		Model Read(System.Net.WebRequest request);

		Model Read(string url);

		Model Read(string file, string b);

		Model ReadFromString(string xml, string b);

		// Added to Model interface
		bool _cleanLoad{get; set;}

		Injektilo.RDF.ProtocolErrorHandler ProtocolErrorHandler{get; set;}

		//Model Read(string url, string lang);

		Model Remove(Statement s);

		long Count{get;}

		bool SupportsSetOperations();

		bool SupportsTransactions();

		Model Union(Model model);

		new IEnumerator GetEnumerator();

		Model Write(System.IO.TextWriter writer);

		Model Write(System.IO.TextWriter writer, StmtEnumerator statements);

		//Model Write(System.IO.StreamWriter writer, string lang);

		//Model Write(System.IO.StreamWriter writer, string lang, string b);

		Model Write(string path);

		Model Write(string path, StmtEnumerator statements);

	}// end Model

	public interface NodeEnumerator : IEnumerator
	{
		new RDFNode Current{get;}

		new bool MoveNext();

		void Close();

		new void Reset();
	}// end NodeEnumerator

	public interface ResEnumerator : IEnumerator
	{
		new Resource Current{get;}

		new bool MoveNext();

		void Close();
		
		new void Reset();
	}// end ResEnumerator

	public interface NsEnumerator : IEnumerator
	{
		new string Current{get;}

		new bool MoveNext();

		void Close();

		new void Reset();
	}// end NsEnumerator

	public interface StmtEnumerator : IEnumerator
	{
		new Statement Current{get;}

		new bool MoveNext();

		void Close();

		new void Reset();
	}// end StmtEnumerator	
	public interface ModelI
	{
		//NodeEnumerator ListContainerMembers(Container cont, NodeEnumeratorFactory f);
		//int ContainerSize(Container cont);
		//int ContainerIndexOf(Container cont, RDFNode n);
		//bool ContainerContains(Container cont, RDFNode n);
		Resource Convert(Resource r);
		Property Convert(Property p);
		RDFNode Convert(RDFNode n);
	}// end ModelI

	public interface Selector
	{
		bool Test(Statement s);
	}

	public interface ClosableEnumerator : IEnumerator
	{
		void Close();
	}
	#region non-implemented classes


/*
	public interface Alt : Container
	{
		RDFNode GetDefault();

		Alt GetDefaultAlt();

		Bag GetDefaultBag();

		bool GetDefaultBoolean();

		byte GetDefaultByte();

		char GetDefaultChar();

		double GetDefaultDouble();

		float GetDefaultFloat();

		int GetDefaultInteger();

		string GetDefaultLanguage();

		long GetDefaultLong();

		Literal GetDefaultLiteral();

		object GetDefaultObject(ObjectF f);

		Resource GetDefaultResource();

		Resource GetDefaultResource(ResourceF f);

		Seq GetDefaultSeq();

		short GetDefaultShort();

		string GetDefaultString();

		new Container Remove(Statement s);

		Alt SetDefault(bool o);

		Alt SetDefault(char o);

		Alt SetDefault(double o);

		Alt SetDefault(float o);

		Alt SetDefault(long o);

		Alt SetDefault(object o);

		Alt SetDefault(RDFNode o);

		Alt SetDefault(string o);

		Alt SetDefault(string o, string l);
	}// end Alt

	public interface Seq : Container
	{
		Seq Add(int index, bool o);

		Seq Add(int index, char o);

		Seq Add(int index, double o);

		Seq Add(int index, float o);

		Seq Add(int index, long o);

		Seq Add(int index, object o);

		Seq Add(int index, RDFNode o);

		Seq Add(int index, string o);

		Seq Add(int index, string o, string l);

		Alt GetAlt(int index);

		Bag GetBag(int index);

		bool GetBoolean(int index);

		byte GetByte(int index);

		char GetChar(int index);

		double GetDouble(int index);

		float GetFloat(int index);

		int GetInteger(int index);

		string GetLanguage(int index);

		Literal GetLiteral(int index);

		long GetLong(int index);

		object GetObject(int index);

		object GetObject(int index, ObjectF f);

		Resource GetResource(int index);

		Resource GetResource(int index, ResourceF f);

		Seq GetSeq(int index);
		
		short GetShort(int index);

		string GetString(int index);

		int IndexOf(bool o);

		int IndexOf(char o);

		int IndexOf(double o);

		int IndexOf(float o);

		int IndexOf(long o);

		int IndexOf(object o);

		int IndexOf(RDFNode o);

		int IndexOf(string o);

		int IndexOf(string o, string l);

		Seq Remove(int index);

		Seq Set(int index, bool o);

		Seq Set(int index, char o);

		Seq Set(int index, double o);

		Seq Set(int index, float o);

		Seq Set(int index, long o);

		Seq Set(int index, object o);

		Seq Set(int index, RDFNode o);

		Seq Set(int index, string o);

		Seq Set(int index, string o, string l);
	}// end Seq

	public interface Bag : Container
	{
		new Container Remove(Statement s);
	}// end Bag	
	public interface Container : Resource
	{
		Container Add(bool o);

		Container Add(char o);

		Container Add(double o);

		Container Add(float o);

		Container Add(long o);

		Container Add(object o);

		Container Add(RDFNode o);

		Container Add(string o);

		Container Add(string o, string l);

		bool Contains(bool o);

		bool Contains(char o);

		bool Contains(double o);

		bool Contains(float o);

		bool Contains(long o);

		bool Contains(object o);

		bool Contains(RDFNode o);

		bool Contains(string o);

		bool Contains(string o, string l);

		Container Remove(Statement s);

		int Count{get;}

	}// end Container
			
	*/
	#endregion




}// end BBN.DAML.Model namespace
