using System;
using System.Collections;
using System.Diagnostics;
using BBN.RDF;
using Injektilo.RDF;

namespace BBN.RDF.Common
{
	/// <summary>
	/// Implementations of common classes
	/// </summary>
	/// <author>Troy Self</author>
	
	/// <summary>
	/// Implements common methods for Model implementations
	/// </summary>
	public abstract class ModelCom : Model, ModelI
	{
		protected Store store;

		// DER 
		protected bool m_cleanLoad = false;
		protected ProtocolErrorHandler m_protocolErrorHandler;
	
		public bool _cleanLoad
		{
			get
			{
				return m_cleanLoad;
			}
			set
			{
				m_cleanLoad = value;
			}
		}

		public ProtocolErrorHandler ProtocolErrorHandler
		{
			get
			{
				return m_protocolErrorHandler;
			}
			set
			{
				m_protocolErrorHandler = value;
			}
		}

		public ModelCom()
		{}

		public virtual Model Abort()
		{
			throw new SystemException("Unsupported Operation");
		}

		public virtual Model Add(Model m)
		{
			return Add(m.ListStatements());
		}

		public virtual Model Add(Statement s)
		{
			if(!(s is StatementImpl))
				s = new StatementImpl(s.GetSubject(), s.GetPredicate(), s.GetObject(), this);
			else if (s.GetModel() != this)
			{
				Resource r = ((StatementImpl)s).Port(this);
				if (r is Statement) 
					s = (Statement)r;
			}
			store.Add(s);
			return this;
		}


		public virtual Model Add(StmtEnumerator s)
		{
			while(s.MoveNext())
			{
				Add(s.Current);
			}
			return this;
		}
		
		public virtual Model Add(Resource s, Property p, RDFNode o)
		{
			Add(CreateStatement(s, p, o));
			return this;
		}
		public virtual Model Add(Resource s, Property p, bool o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, long o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, char o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, float o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, double o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, string o)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, string o, bool wellFormed)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o, wellFormed)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, string o, string l)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o, l)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, string o, string l, bool wellFormed)
		{
			Add(CreateStatement(s, p, new LiteralImpl(o, l, wellFormed)));
			return this;
		}
		public virtual Model Add(Resource s, Property p, object o)
		{
			if(o is RDFNode)
				Add(CreateStatement(s, p, (RDFNode)o));
			else
				Add(CreateStatement(s, p, new LiteralImpl(o)));
			return this;
		}

		public virtual Model Begin()
		{
			throw new SystemException("Unsupported Operation");
		}

		public virtual Model Commit()
		{
			throw new SystemException("Unsupported Operation");
		}

		public virtual bool Contains(Resource s, Property p)
		{
			return store.Contains(s, p);
		}

		public virtual bool Contains(Resource s, Property p, RDFNode o)
		{
			return Contains(new StatementImpl(s, p, o));
		}

		public virtual bool Contains(Resource s, Property p, bool o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, long o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, char o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, float o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, double o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, string o)
		{
			return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Resource s, Property p, string o, string l)
		{
			return Contains(s, p, new LiteralImpl(o, l));
		}

		public virtual bool Contains(Resource s, Property p, object o)
		{
			if(o is RDFNode)
				return Contains(s, p, (RDFNode)o);
			else
				return Contains(s, p, new LiteralImpl(o));
		}

		public virtual bool Contains(Statement s)
		{
			return store.Contains(s);
		}

		public virtual bool ContainsAll(Model model)
		{
			StmtEnumerator e = model.ListStatements();
			return ContainsAll(e);
		}

		public virtual bool ContainsAll(StmtEnumerator e)
		{
			while(e.MoveNext())
				if(!Contains(e.Current)) return false;
			return true;
		}

		public virtual bool ContainsAny(StmtEnumerator e)
		{
			while(e.MoveNext())
				if(Contains(e.Current)) return true;
			return false;
		}

		public virtual bool ContainsAny(Model model)
		{
			StmtEnumerator e = model.ListStatements();
			return ContainsAny(e);
		}

		public virtual Literal CreateLiteral(bool v)
		{
			return CreateLiteral(v.ToString());
		}
		public virtual Literal CreateLiteral(long v)
		{
			return CreateLiteral(v.ToString());
		}
		public virtual Literal CreateLiteral(char v)
		{
			return CreateLiteral(v.ToString());
		}
		public virtual Literal CreateLiteral(float v)
		{
			return CreateLiteral(v.ToString());
		}
		public virtual Literal CreateLiteral(double v)
		{
			return CreateLiteral(v.ToString());
		}
		public virtual Literal CreateLiteral(string v)
		{
			return CreateLiteral(v, "");
		}
		public virtual Literal CreateLiteral(string v, string l)
		{
			return new LiteralImpl(v, l);
		}
		public virtual Literal CreateLiteral(string v, string l, bool wellFormed)
		{
			return new LiteralImpl(v, l, wellFormed);
		}
		public virtual Literal CreateLiteral(object v)
		{
			return CreateLiteral(v.ToString(), "");
		}
		public virtual Statement CreateStatement(Resource s, Property p, bool o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, long o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, char o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, float o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, double o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, string o)
		{
			return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, object o)
		{
			if(o is RDFNode)
				return CreateStatement(s, p, (RDFNode)o);
			else
				return CreateStatement(s, p, CreateLiteral(o));
		}
		public virtual Statement CreateStatement(Resource s, Property p, string o, bool wellFormed)
		{
			return CreateStatement(s, p, new LiteralImpl(o, wellFormed));
		}
		public virtual Statement CreateStatement(Resource s, Property p, string o, string l)
		{
			return CreateStatement(s, p, CreateLiteral(o, l));
		}
		public virtual Statement CreateStatement(Resource s, Property p, string o, string l,
			bool wellFormed)
		{
			return CreateStatement(s, p, CreateLiteral(o, l, wellFormed));
		}

		public virtual Statement CreateStatement(Resource s, Property p, RDFNode o)
		{
			return new StatementImpl(s, p, o, this);
		}
		public virtual Statement CreateStatement(Resource s, Property p, RDFNode o, string uri)
		{
			return new StatementImpl(s, p, o, uri, null, this);
		}
		public virtual Statement CreateStatement(Resource s, Property p, RDFNode o, string uri, string source)
		{
			return new StatementImpl(s, p, o, uri, source, this);
		}
		public virtual Property CreateProperty(string nameSpace, string localName)
		{
			return store.CreateProperty(nameSpace, localName, this);
		}
		public virtual Property CreateProperty(string uri)
		{
			return store.CreateProperty(uri, this);
		}
		public virtual Resource CreateResource()
		{
			return store.CreateResource(this);
		}
		public virtual Resource CreateResource(string uri)
		{
			return store.CreateResource(uri, this);
		}
		public abstract Model Difference(Model model);

		public virtual Statement GetProperty(Resource s, Property p)
		{
			StmtEnumerator e = null;
			try
			{
				e = ListStatements(new SelectorImpl(s, p, (RDFNode) null));

				if(e.MoveNext())
					return e.Current;
				else
					throw new System.Exception("Property not found");
			}
			finally
			{
				e.Close();
			}
		}

		public virtual Property GetProperty(string nameSpace, string localName)
		{
			return store.GetProperty(nameSpace, localName, this);
		}

		public virtual Property GetProperty(string uri)
		{ 
			return store.GetProperty(uri, this);
		}

		public virtual Resource GetResource(string uri)
		{
			return store.GetResource(uri, this);
		}

		public virtual Source GetSource(string uri)
		{
			return store.GetSource(uri);
		}

		public virtual bool Independent()
		{
			// no idea why this is necessary but this is how Jena implements this
			return true;
		}
		public abstract Model Intersection(Model model);

		public abstract Model GetModelForSource(Source s);

		public virtual bool IsReified(Statement s)
		{
			StmtEnumerator iter = null;
			try 
			{
				iter = ListStatements(new SelectorImpl(s.GetReificationId(), null, (RDFNode) null));
				if (iter.MoveNext())
					return true;
				iter = ListStatements(new SelectorImpl(null, null, s.GetReificationId()));
				if (iter.MoveNext())
					return true;
				/*
				iter = ListStatements(new SelectorImpl(s, null, (RDFNode) null));
				if (iter.MoveNext()) 
					return true;
				iter = ListStatements(new SelectorImpl(null, null, s));
				if (iter.MoveNext())
					return iter.MoveNext();
				*/
				return false;
			} 
			finally 
			{
				iter.Close();
			}
		}

		public virtual NsEnumerator ListNameSpaces()
		{
			return store.ListNameSpaces();
		}

		public virtual NodeEnumerator ListObjects()
		{
			Hashtable objects = new Hashtable();
			IEnumerator e = store.List();
			while(e.MoveNext())
				objects[((Statement)e.Current).GetObject()] = true;
			return new NodeEnumeratorImpl(objects.Keys);
		}

		public virtual NodeEnumerator ListObjectsOfProperty(Property p)
		{
			Hashtable objects = new Hashtable();
			IEnumerator e = store.ListByPredicate(p);
			Statement stmt = null;
			while(e.MoveNext())
			{
				stmt = (Statement)e.Current;
				objects[stmt.GetObject()]=true;
			}
			return new NodeEnumeratorImpl(objects.Keys);
		}

		public virtual NodeEnumerator ListObjectsOfProperty(Resource s, Property p)
		{
			Hashtable objects = new Hashtable();
			IEnumerator e = store.List(s, p, null);
			Statement stmt = null;
			while(e.MoveNext())
			{
				stmt = (Statement)e.Current;
				if(stmt.GetPredicate().Equals(p))
					objects[stmt.GetObject()]=true;
			}
			return new NodeEnumeratorImpl(objects.Keys);
		}

		public virtual StmtEnumerator ListReifiedStatements()
		{
			ArrayList reifiedStatements = new ArrayList();
			IEnumerator e = store.List();
			object subject;
			RDFNode obj;
			while(e.MoveNext())
			{
				subject = ((Statement)e.Current).GetSubject();
				if(subject is Statement)
					reifiedStatements.Add(subject);
				obj = ((Statement)e.Current).GetObject();
				if (obj is Statement)
					reifiedStatements.Add(obj);
			}
			return new StmtEnumeratorImpl(reifiedStatements);
		}

		public virtual StmtEnumerator ListStatements()
		{
			return new StmtEnumeratorImpl(store.List());
		}

		public virtual StmtEnumerator ListStatements(Selector selector)
		{
			IEnumerator iter;

			if (selector is SelectorImpl) 
			{
				SelectorImpl s = (SelectorImpl) selector;
				iter = store.List(s.GetSubject(),
					s.GetPredicate(),
					s.GetObject());
			}
			else 
			{
				iter = store.List();
			}

			if (iter == null) 
			{
				return new StmtEnumeratorImpl();
			}
			return ListStatements(selector, iter);
		}

		protected virtual StmtEnumerator ListStatements(Selector s, IEnumerator iter)
		{
			ArrayList matching = new ArrayList();
			Statement stmt;
			while (iter.MoveNext()) 
			{
				stmt = (Statement) iter.Current;
				try 
				{
					if (s.Test(stmt)) 
					{
						matching.Add(stmt);
					}
				} 
				catch (Exception e) 
				{
					if (iter is ClosableEnumerator) 
					{
						((ClosableEnumerator)iter).Close();
					}
					throw e;
				}
			}
			return new StmtEnumeratorImpl(matching.GetEnumerator());
		}

		public virtual StmtEnumerator ListStatementsOfSource(Source s)
		{

			return new StmtEnumeratorImpl(store.ListBySource(s));
		}

		public virtual ResEnumerator ListSubjects()
		{
			return store.ListSubjects();
		}

		public virtual IEnumerator ListSources()
		{
			return store.ListSources();
		}

		public virtual ResEnumerator ListSubjectsWithProperty(Property p)
		{
			Hashtable subjects = new Hashtable();
			IEnumerator e = store.ListByPredicate(p);
			Statement stmt;
			while(e.MoveNext())
			{
				stmt = (Statement)e.Current;
				if(stmt.GetPredicate().Equals(p))
					subjects[stmt.GetSubject()]=true;
			}
			return new ResEnumeratorImpl(subjects.Keys);
		}

		public virtual ResEnumerator ListSubjectsWithProperty(Property p, RDFNode o)
		{
			Hashtable subjects = new Hashtable();
			IEnumerator e = store.List(null, p, o);
			Statement stmt;
			while(e.MoveNext())
			{
				stmt = (Statement)e.Current;
				if(stmt.GetPredicate().Equals(p) && stmt.GetObject().Equals(o))
					subjects[stmt.GetSubject()]=true;
			}
			return new ResEnumeratorImpl(subjects.Keys);
		}

		public abstract Model Query(Selector selector);

		public virtual Model Read(System.IO.TextReader reader, string b)
		{
			RdfModel rdfModel = new RdfModel();
			rdfModel.Load(reader, b);
			rdfModel.ConvertStatements(this);
			return this;
		}

		public virtual Model Read(System.IO.TextReader reader, string b, string lang)
		{
			RdfModel rdfModel = new RdfModel();
			rdfModel.Load(reader, b, lang);
			rdfModel.ConvertStatements(this);
			return this;
		}

		public virtual Model Read(System.IO.Stream stream, string b)
		{
			RdfModel rdfModel = new RdfModel();
			rdfModel.Load(stream, b);
			rdfModel.ConvertStatements(this);
			return this;
		}


		public virtual Model Read(System.IO.Stream stream, string b, string lang)
		{
			RdfModel rdfModel = new RdfModel();
			rdfModel.Load(stream, b, lang);
			rdfModel.ConvertStatements(this);
			return this;
		}

		public virtual Model Read(System.Net.WebRequest request)
		{
			RdfModel rdfModel = new RdfModel();
			rdfModel.ProtocolErrorHandler = m_protocolErrorHandler;
			rdfModel.Load(request);
			rdfModel.ConvertStatements(this);
			return this;
		}

		/// <summary>
		/// Read a RDF file into the model.
		/// </summary>
		/// <param name="url">The URL of the RDF file to load.</param>
		/// <returns>The model</returns>
		/// <remarks>The url resolver can only handle http:// and file:// protocols.</remarks>
		public virtual Model Read(string url)
		{
			RdfModel rdfModel = new RdfModel();		
			if (_cleanLoad) 
			{
				rdfModel.CleanLoad(url);
			} 
			else 
			{
				rdfModel.Load(url);

			}
			rdfModel.ConvertStatements(this);
			return this;
		}

		public virtual Model Read(string file, string b)
		{
			RdfModel rdfModel = new RdfModel();
			if (_cleanLoad)
			{
				rdfModel.CleanLoad(file, b);
			} 
			else
			{
				rdfModel.Load(file, b);
			}
			rdfModel.ConvertStatements(this);
			return this;
		}

		/// <summary>
		/// Read directly from string data.
		/// </summary>
		/// <param name="xml">The string data to parse.</param>
		/// <param name="b">The base URI for the data.</param>
		/// <returns></returns>
		public virtual Model ReadFromString(string xml, string b)
		{
			RdfModel rdfModel = new RdfModel();
			if (_cleanLoad)
			{
				rdfModel.CleanLoadFromString(xml, b);
			} 
			else 
			{
				rdfModel.LoadFromString(xml, b);
			}
			rdfModel.ConvertStatements(this);
			return this;
		}

		//public Model Read(string url, string lang){return null;}

		public virtual Model Remove(Statement s)
		{
			store.Remove(s);
			return this;
		}
		public virtual long Count
		{
			get{return store.Count;}
		}
		// Make the following methods abstract so the implementing class will be forced to
		// show whether it supports transactions or Set operations
		public abstract bool SupportsSetOperations();

		public abstract bool SupportsTransactions();

		public abstract Model Union(Model model);

		public virtual IEnumerator GetEnumerator()
		{
			return store.GetEnumerator();
		}

		// At this time we can only support lang = RDF/XML and absolute URIs.
		//public Model Write(System.IO.StreamWriter writer, string lang){return null;}

		//public Model Write(System.IO.StreamWriter writer, string lang, string b){return null;}

		public virtual Model Write(string path)
		{
			System.IO.StreamWriter writer = new System.IO.StreamWriter(path);
			return Write(writer);
		}

		public virtual Model Write(string path, StmtEnumerator statements)
		{
			System.IO.StreamWriter writer = new System.IO.StreamWriter(path);
			return Write(writer, statements);
		}

		
		/// <summary>write (serialize) the model to the specified file</summary>
		/// <remarks>This is a simple implementation missing some desirable features,
		/// e.g. shared namespaces, handling of anonymous ndoes, automatic determination of CDATA.  
		/// It's not very efficient to build the whole XML model and then serialize it; 
		/// this could be done incrementally.
		/// Output isn't indented.</remarks>
		/// <author>Mike Dean, mdean@bbn.com</author>
		/// <summary>write (serialize) the model to the specified file</summary>
		/// <remarks>This is a simple implementation missing some desirable features,
		/// e.g. shared namespaces, handling of anonymous ndoes, automatic determination of CDATA.  
		/// It's not very efficient to build the whole XML model and then serialize it; 
		/// this could be done incrementally.
		/// Output isn't indented.</remarks>
		/// <author>Mike Dean, mdean@bbn.com; Modified for new .NET API by Troy Self</author>
		public virtual Model Write(System.IO.TextWriter writer)
		{
			Hashtable namespaces = new Hashtable(); // stores the prefixes indexed by URI
			// start off by putting the rdf namespace in the hashtable
			namespaces[RdfSyntax.URI] = "rdf";
			System.Xml.XmlDocument document = new System.Xml.XmlDocument();
			//System.Xml.XmlDeclaration declaration = document.CreateXmlDeclaration("1.0", "ISO-8859-1", "");
			//document.AppendChild(declaration);
			System.Xml.XmlElement rdf = document.CreateElement("rdf", RdfSyntax.RDF, RdfSyntax.URI);
			document.AppendChild(rdf);
			ResEnumerator resEnum = store.ListSubjects();
			while(resEnum.MoveNext())
			{
				System.Xml.XmlElement description = document.CreateElement("rdf", RdfSyntax.Description, RdfSyntax.URI);
				if (resEnum.Current is Statement)
				{ 
					Console.WriteLine("Didn't expect to find a statement as the subject "+resEnum.Current);
					description.SetAttribute(RdfSyntax.About, RdfSyntax.URI, ((Statement)resEnum.Current).GetReificationId().ToString());
				}
				else
					description.SetAttribute(RdfSyntax.About, RdfSyntax.URI, resEnum.Current.GetURI());
				rdf.AppendChild(description);
					//IEnumerator stmtEnumerator = store.ListBySubject(resEnum.Current);
					//IList statements = subjects.GetStatements(subject);
					IList statements = store.GetStatementsBySubject(resEnum.Current);
					string nodeName, nodePrefix, nodeNS;
				foreach (Statement statement in statements)
				{
					string predicate = statement.GetPredicate().GetURI();
					int hash = predicate.LastIndexOf('#');
					nodeName = predicate.Substring(hash+1);
					nodeNS = predicate.Substring(0, hash+1);
					// if the namespace is new, add it to the table and assign a prefix
					if(namespaces[nodeNS] == null)
					{
						namespaces[nodeNS] = "ns" + namespaces.Count;
						rdf.SetAttribute("xmlns:"+(string)namespaces[nodeNS], nodeNS);
						//Trace.WriteLine("Added namespace: ns" + (namespaces.Count-1) + " = " + nodeNS);
					}
					// now we can get the prefix from the hashtable
					nodePrefix = (string)namespaces[nodeNS];
					System.Xml.XmlElement element = document.CreateElement(nodePrefix, nodeName, nodeNS);
					//if (statement.ObjectType == RdfObjectType.Literal)

					if(statement.GetObject() is Literal)
						element.AppendChild(document.CreateTextNode(statement.GetObject().ToString()));
					else
						element.SetAttribute(RdfSyntax.Resource, RdfSyntax.URI, statement.GetObject().ToString());
					// Add statement ID if one exists
					if (!statement.IsAnon() || statement.IsReified()) 
					{
						element.SetAttribute(RdfSyntax.ID,RdfSyntax.URI, statement.GetReificationId().ToString());
					}
					description.AppendChild(element);
				}
			}
			System.Xml.XmlWriter xmlwriter = new System.Xml.XmlTextWriter(writer);
			document.WriteContentTo(xmlwriter);
			writer.WriteLine();
			//writer.Close();
			return this;
		}

		/// <summary>
		/// Write a set of statements out to a file.
		/// </summary>
		/// <param name="writer">What to write to</param>
		/// <param name="stmts">The statements to write</param>
		/// <returns></returns>
		public virtual Model Write(System.IO.TextWriter writer, StmtEnumerator stmts)
		{
			// Preindex statements by their subjects
			BBN.RDF.Mem.NodeMap stmtMap = new BBN.RDF.Mem.NodeMap();
			while (stmts.MoveNext()) 
			{
				stmtMap.Add(stmts.Current.GetSubject(), stmts.Current);
			}

			Hashtable namespaces = new Hashtable(); // stores the prefixes indexed by URI
			// start off by putting the rdf namespace in the hashtable
			namespaces[RdfSyntax.URI] = "rdf";
			System.Xml.XmlDocument document = new System.Xml.XmlDocument();
			System.Xml.XmlElement rdf = document.CreateElement("rdf", RdfSyntax.RDF, RdfSyntax.URI);
			document.AppendChild(rdf);

			IEnumerator resEnum = stmtMap.GetKeys().GetEnumerator();
			while(resEnum.MoveNext())
			{
				System.Xml.XmlElement description = document.CreateElement("rdf", RdfSyntax.Description, RdfSyntax.URI);
				if (resEnum.Current is Statement)
				{ 
					Console.WriteLine("Didn't expect to find a statement as the subject "+resEnum.Current);
					description.SetAttribute(RdfSyntax.About, RdfSyntax.URI, ((Statement)resEnum.Current).GetReificationId().ToString());
				}
				else if (resEnum.Current is Resource)
					description.SetAttribute(RdfSyntax.About, RdfSyntax.URI, ((Resource)resEnum.Current).GetURI());
				rdf.AppendChild(description);

				if (resEnum.Current is RDFNode)
				{
					IList statements = stmtMap.GetStatements((RDFNode)resEnum.Current);
					string nodeName, nodePrefix, nodeNS;
					foreach (Statement statement in statements)
					{
						string predicate = statement.GetPredicate().GetURI();
						int hash = predicate.LastIndexOf('#');
						nodeName = predicate.Substring(hash+1);
						nodeNS = predicate.Substring(0, hash+1);
						// if the namespace is new, add it to the table and assign a prefix
						if(namespaces[nodeNS] == null)
						{
							namespaces[nodeNS] = "ns" + namespaces.Count;
							rdf.SetAttribute("xmlns:"+(string)namespaces[nodeNS], nodeNS);
							//Trace.WriteLine("Added namespace: ns" + (namespaces.Count-1) + " = " + nodeNS);
						}
						// now we can get the prefix from the hashtable
						nodePrefix = (string)namespaces[nodeNS];
						System.Xml.XmlElement element = document.CreateElement(nodePrefix, nodeName, nodeNS);
						//if (statement.ObjectType == RdfObjectType.Literal)

						if(statement.GetObject() is Literal)
							element.AppendChild(document.CreateTextNode(statement.GetObject().ToString()));
						else
							element.SetAttribute(RdfSyntax.Resource, RdfSyntax.URI, statement.GetObject().ToString());
						// Add statement ID if one exists
						if (!statement.IsAnon() | statement.IsReified()) 
						{
							element.SetAttribute(RdfSyntax.ID,RdfSyntax.URI, statement.GetReificationId().ToString());
						}
						description.AppendChild(element);
					}
				}
			}
			System.Xml.XmlWriter xmlwriter = new System.Xml.XmlTextWriter(writer);
			document.WriteContentTo(xmlwriter);
			writer.WriteLine();
			//writer.Close();
			return this;
		}

		/*public NodeEnumerator ListContainerMembers(Container cont, NodeEnumeratorFactory f)
		{
			IEnumerator iter = store.ListBySubject(cont);
			Statement stmt;
			string rdfURI = Vocabulary.RDF.GetURI();
			ArrayList result = new ArrayList();
			int maxOrdinal = 0;
			int ordinal;
			while(iter.MoveNext())
			{
				stmt = (Statement)iter.Current;
				ordinal = stmt.GetPredicate().GetOrdinal();
				if(stmt.GetSubject().Equals(cont) && ordinal != 0)
				{
					if(ordinal > maxOrdinal)
					{
						maxOrdinal = ordinal;
						result.Capacity = ordinal;
					}
					result[ordinal-1] = stmt;
				}
			}//end while
			/*
			 * stuff here about ClosableIterator.  I'll have to look into this to see what
			 * it does. Java...
			 * if(iter instanceof ClosableIterator)
			 *    ((ClosableIterator)iter).close();
			 *//*
			return f.CreateEnumerator(result.GetEnumerator(), result, cont);
		}*/
		/*public int ContainerSize(Container cont)
		{
			int result = 0;
			IEnumerator iter = store.ListBySubject(cont);
			Property predicate;
			Statement stmt;
			string rdfURI = Vocabulary.RDF.GetURI();
			while(iter.MoveNext())
			{
				stmt = (Statement)iter.Current;
				predicate = stmt.GetPredicate();
				if(stmt.GetSubject().Equals(cont) && predicate.GetOrdinal() != 0)
					result++;
			}
			// more ClosableIterator stuff here.  Gotta find out what closing an iterator does
			return result;
		}*/
		/*public int ContainerIndexOf(Container cont, RDFNode n)
		{
			int result = 0;
			IEnumerator iter = store.ListBySubject(cont);
			Property predicate;
			Statement stmt;
			string rdfURI = Vocabulary.RDF.GetURI();
			while(iter.MoveNext())
			{
				stmt = (Statement)iter.Current;
				predicate = stmt.GetPredicate();
				if(stmt.GetSubject().Equals(cont) && predicate.GetOrdinal() != 0
					&& n.Equals(stmt.GetObject()))
				{
					result = predicate.GetOrdinal();
					break;
				}
			}
			// ClosableIterator stuff
			return result;
		}*/
		/*public bool ContainerContains(Container cont, RDFNode n)
		{
			return (ContainerIndexOf(cont, n) != 0);
		}*/
		public virtual Resource Convert(Resource r)
		{
			if(r.GetModel() == this)
				return r;
			else return ((ResourceI)r).Port(this);
		}
		public virtual Property Convert(Property p)
		{
			if(p.GetModel() == this)
				return p;
			else return (Property)((ResourceI)p).Port(this);
		}
		public virtual RDFNode Convert(RDFNode n)
		{
			if(n is Property)
				return Convert((Property)n);
			else if(n is Resource)
				return Convert((Resource)n);
			else return n;
		}
		/*
		 * For new models are the same only if they are the same object.
		 * 
		public override bool Equals(object model)
		{
			if(model == null || !(model is Model))
				return false;
			// TBD ?? Is this in Jena yet?
			//return ModelMatcher.Equals(this, (Model)model);
			return false; // implement ModelMatcher later
		}
		*/
		public virtual void Close()
		{
			store.Close();
		}
	}// end ModelCom
}
