using System;
using System.Collections;
using BBN.RDF;


namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class PropertyAccessorImpl : PropertyAccessor
	{
		// Instance variables
		//////////////////////////////////

		/** The property that this accessor works on */
		protected Property m_property = null;

		/** The underlying value that this is an accessor to */
		protected DAMLCommon m_val = null;



		// Constructors
		//////////////////////////////////

		/**
		 * Construct a new accessor for the given property, which takes
		 * the given value type as it range.
		 *
		 * @param property The property that this accessor works on
		 * @param val The underlying DAML value that this is an accessor to
		 */
		public PropertyAccessorImpl( Property property, DAMLCommon val ) 
		{
			m_property = property;
			m_val = val;
		}



		// External signature methods
		//////////////////////////////////



		/**
		 * Answer the property that this accessor works on
		 *
		 * @return A property
		 */
		public Property GetProperty() 
		{
			return m_property;
		}


		/**
		 * Answer the number of values that the encapsulated property has in the
		 * RDF model.  Note that <code>count</code> counts all RDF values, so it
		 * is possible for this method to return a non-zero value, but {@link #GetDAMLValue}
		 * to return null, if none of the values are DAML values.
		 *
		 * @return The number RDF statements for this property in the model.
		 */
		public int Count 
		{
			get 
			{
				return m_val.GetNumPropertyValues( GetProperty() );
			}
		}


		/**
		 * Answer an iteration over the values that this property has in the
		 * RDF model.  Note that these values may be any RDF value, depending
		 * on what is in the model. Client code should only assume that the
		 * values return by the Iterator are {@link com.hp.hpl.mesa.rdf.jena.model.RDFNode RDFNode}
		 * objects.
		 *
		 * @param closed If true, and the property is transitive, generate the closure
		 *               of the property starting from the encapsulated resource.
		 * @return An iteration over the RDF values of the encapsulated property.
		 */
		public IEnumerator GetAll( bool closed ) 
		{
			return m_val.GetAll( GetProperty(), closed );
		}


		/**
		 * Answer a general value of the encapsulated property. If it has no values, answer
		 * null. If it has one value, answer that value. Otherwise, answer an undetermined
		 * member of the set of values. This version of the method makes no assumptions
		 * about the property value, other than it is an RDF node. This is the safest, most
		 * conservative, assumption.  If it is known that a value is certain to be a
		 * DAML value, you can use {@link #GetDAMLValue} instead.
		 *
		 * @return A value for the encapsulated property in the RDF model, or null
		 *         if the property has no value.
		 */
		public RDFNode Get() 
		{
			IEnumerator i = GetAll( false );
			return (i == null  ||  !i.MoveNext()) ? null : ((RDFNode) i.Current);
		}


		/**
		 * Answer a value of the encapsulated property, making the assumption that it is
		 * a DAML value. If the property has no DAML value, answer
		 * null. If it has one DAML value, answer that value. Otherwise, answer an undetermined
		 * member of the set of values.  This method is optimised to select only DAML
		 * values for the property - that is, values that extend DAMLCommon.  <b>This
		 * method will therefore answer null if there is no DAML value for the property, even
		 * if there is one or more vanilla-RDF values</b>. For
		 * a more general version of this method, which returns all RDF values, see {@link
		 * #Get}.
		 *
		 * @return A DAML value for the encapsulated property in the RDF model, or null
		 *         if the property has no DAML value.
		 */
		public DAMLCommon GetDAMLValue() 
		{
			IEnumerator i = GetAll( false );

			if (i != null) 
			{
				while (i.MoveNext()) 
				{
					Object next = i.Current;

					// return the first DAML value we find
					if (next is DAMLCommon) 
					{
						return (DAMLCommon) next;
					}
				}
			}

			// otherwise return null
			return null;
		}


		/**
		 * Add a value to the encapsulated property.
		 *
		 * @param value The value to be added.
		 */
		public void Add( Resource value ) 
		{
			try 
			{
				m_val.AddProperty( GetProperty(), value );
			}
			catch (System.Exception e) 
			{
				Console.Error.WriteLine( "RDF exception: " + e );
			}
		}


		/**
		 * Remove a value from the encapsulated property.
		 *
		 * @param value The value to be removed.
		 */
		public void Remove( DAMLCommon value ) 
		{
			m_val.RemoveProperty( GetProperty(), value );
		}


		/**
		 * Answer true if the encapsulated property has the given value as one of its
		 * values.  Note: does not traverse the closure of the property.
		 *
		 * @param value A DAML value to test for
		 * @return True if the RDF model contains a statement giving a value for
		 *         the encapsulated property matching the given value.
		 */
		public bool HasValue( DAMLCommon value ) 
		{
			// test all of my values to see if they contain the given target
			for (IEnumerator i = GetAll( false );  i.MoveNext(); ) 
			{
				if (value.Equals( i.Current )) 
				{
					return true;
				}
			}

			return false;
		}



		// Internal implementation methods
		//////////////////////////////////


	}
}
