using System;
using System.Collections;
using BBN.RDF;
using BBN.DAML.Vocabulary;

namespace BBN.DAML
{
	/// <summary>
	/// 
	/// </summary>
	public class PropertyEnumerator : IEnumerator
	{

		
		// Instance variables
		//////////////////////////////////

		/** The queue of nodes to be evaluated */
		// LinkedList
		protected ArrayList m_nodeQueue = new ArrayList();

		/** The property we're evaluating */
		protected Property m_pred = null;

		/** The inverse of pred, or null */
		protected Property m_inverse = null;

		/** Collection of properties equivalent to m_pred */
		// HashSet
		protected IList m_predEquivs = new ArrayList();

		/** Collection of properties equivalent to m_inverse */
		// HashSet
		protected IList m_inverseEquivs = new ArrayList();

		/** Flag to say if pred is transitive */
		protected bool m_transitive = false;

		/** Flag to say if pred is reflexive */
		protected bool m_isReflexive = false;

		/** Set of nodes we've seen on this trip */
		// WeakHashMap
		protected Hashtable m_seen = new Hashtable();

		/** The resources we started from (if many) */
		// Vector
		protected IList m_roots = new ArrayList();

		/** The default value for the iterator, or null if no default */
		protected Object m_defaultValue = null;

		/** A flag to show that the default value has been returned */
		protected bool m_defaultValueSeen = false;

		/** A flag to control whether we use equivalent values during the iteration */
		protected bool m_useEquivalence = true;

		/** The model we are operating on */
		protected Model m_model = null;

		/** The current object to return */
		protected Object m_current = null;

		/** Is there a current object to return? */
		protected bool m_next = false;


		// Constructors
		//////////////////////////////////


		public PropertyEnumerator( Resource root, Property pred, Property inverse, bool isTransitive, bool isReflexive ) 
		{
			Initialise( root, pred, inverse, isTransitive, isReflexive, true, null);
			Traverse( root );
		}


		/// <summary>
		/// Construct a property iterator for the given property, starting from the
		/// given resource.  The property may be defined to be symmetric by supplying
		/// its inverse (which could be itself), and/or transitive.  The property may also
		/// be defined to be reflexive, in which case root itself will be returned as
		/// a member of the iteration.
		/// </summary>
		/// <param name="root">The root resource from whence to start iterating over the closure of pred</param>
		/// <param name="pred">The property to iterate over</param>
		/// <param name="inverse">The inverse of pred, or null if pred has no inverse. The inverse is used
		/// to include resource y in the iteration if P' = inverse(P) and y P' x.</param>
		/// <param name="isTransitive">If true, the property is transitive</param>
		/// <param name="isReflexive">If true, the property is reflexive (so, the root resource will be included
		/// in the iteration)</param>
		/// <param name="useEquivalence">If true, equivalence between DAML values will be included in the
		/// iteration (unless the model containing the DAML values has equivalence
		///switched off via {@link com.hp.hpl.jena.daml.Model#setUseEquivalence}.</param>
		public PropertyEnumerator( Resource root, Property pred, Property inverse, bool isTransitive,
			bool isReflexive, bool useEquivalence ) 
		{
			Initialise( root, pred, inverse, isTransitive, isReflexive, useEquivalence, null);
			Traverse( root );
		}


		/**
		 * Construct a property iterator for the given property, starting from the
		 * given resource.  The property may be defined to be symmetric by supplying
		 * its inverse (which could be itself), and/or transitive.  The property may also
		 * be defined to be reflexive, in which case root itself will be returned as
		 * a member of the iteration.
		 *
		 * @param root The root resource from whence to start iterating over the closure of pred
		 * @param pred The property to iterate over
		 * @param inverse The inverse of pred, or null if pred has no inverse. The inverse is used
		 *                to include resource y in the iteration if P' = inverse(P) and y P' x.
		 * @param isTransitive If true, the property is transitive
		 * @param isReflexive If true, the property is reflexive (so, the root resource will be included
		 *                    in the iteration).
		 * @param useEquivalence If true, equivalence between DAML values will be included in the
		 *                       iteration (unless the model containing the DAML values has equivalence
		 *                       switched off via {@link com.hp.hpl.jena.daml.DAMLModel#setUseEquivalence}.
		 * @param model The model that is the source of statements to traverse.
		 */
		public PropertyEnumerator( Resource root, Property pred, Property inverse, bool isTransitive,
			bool isReflexive, bool useEquivalence, Model model ) 
		{
			Initialise( root, pred, inverse, isTransitive, isReflexive, useEquivalence, model );
			Traverse( root );
		}

		public PropertyEnumerator( IEnumerator roots, Property pred, Property inverse, bool isTransitive, bool isReflexive ) 
		{
			Initialise( null, pred, inverse, isTransitive, isReflexive, true, null );
			Traverse( roots );
		}

		/// <summary>
		/// Construct a property iterator for the given property, starting from the
		/// given set of resources.  The property may be defined to be symmetric by supplying
		/// its inverse (which could be itself), and/or transitive.  The property may also
		/// be defined to be reflexive, in which case all of the given root resources will
		/// be returned as members of the iteration.
		/// </summary>
		/// <param name="roots">A set of root resources from whence to start iterating over the closure of pred,
		/// represented as an iterator</param>
		/// <param name="pred">The property to iterate over</param>
		/// <param name="inverse">The inverse of pred, or null if pred has no inverse. The inverse is used
		/// to include resource y in the iteration if P' = inverse(P) and y P' x.</param>
		/// <param name="isTransitive">If true, the property is transitive</param>
		/// <param name="isReflexive">If true, the property is reflexive (so, the root resources will be included
		/// in the iteration).</param>
		/// <param name="useEquivalence">If true, equivalence between DAML values will be included in the
		/// iteration (unless the model containing the DAML values has equivalence
		/// switched off via {@link com.hp.hpl.jena.daml.DAMLModel#setUseEquivalence}.</param>
		public PropertyEnumerator( IEnumerator roots, Property pred, Property inverse, bool isTransitive,
			bool isReflexive, bool useEquivalence ) 
		{
			Initialise( null, pred, inverse, isTransitive, isReflexive, useEquivalence, null);
			Traverse( roots );
		}

		/**
		 * Construct a property iterator for the given property, starting from the
		 * given set of resources.  The property may be defined to be symmetric by supplying
		 * its inverse (which could be itself), and/or transitive.  The property may also
		 * be defined to be reflexive, in which case all of the given root resources will
		 * be returned as members of the iteration.
		 *
		 * @param roots A set of root resources from whence to start iterating over the closure of pred,
		 *              represented as an iterator
		 * @param pred The property to iterate over
		 * @param inverse The inverse of pred, or null if pred has no inverse. The inverse is used
		 *                to include resource y in the iteration if P' = inverse(P) and y P' x.
		 * @param isTransitive If true, the property is transitive
		 * @param isReflexive If true, the property is reflexive (so, the root resources will be included
		 *                    in the iteration).
		 * @param useEquivalence If true, equivalence between DAML values will be included in the
		 *                       iteration (unless the model containing the DAML values has equivalence
		 *                       switched off via {@link com.hp.hpl.jena.daml.DAMLModel#setUseEquivalence}.
		 * @param model The model that is the source of statements to traverse.
		 */
		public PropertyEnumerator( IEnumerator roots, Property pred, Property inverse, bool isTransitive,
			bool isReflexive, bool useEquivalence, Model model ) 
		{
			Initialise( null, pred, inverse, isTransitive, isReflexive, useEquivalence, model );
			Traverse( roots );
		}


		// External signature methods
		//////////////////////////////////

		/// <summary>
		/// Answer true if the iteration over the closure of the predicate will answer any values that have
		/// not yet been returned.
		/// </summary>
		/// <returns>True if there is at least one more element in the iteration</returns>
		public bool MoveNext() 
		{
			// is there at least one more node in the queue?
			if (m_nodeQueue.Count != 0  ||  (HasDefaultValue() && !m_defaultValueSeen)) 
			{
				m_current = GetCurrent();
				m_next = true;
				return true;
			} 
			else 
			{
				m_next = false;
				return false;
			}
		}


		/**
		 * Answer the next RDFNode in the iteration over the given predicate.  Note that each node in the
		 * closure of the predicate will be answered only once, to avoid endless loops.
		 *
		 * @return The next RDFNode in the iteration over the closure of the predicate.
		 * @exception java.util.NoSuchElementException if the iterator has no more elements
		 */
		public Object Current 
		{
			get 
			{
				if (m_next) 
				{
					return m_current;
				} 
				else 
				{
					throw new System.InvalidOperationException("Enumeration has not started. Call MoveNext.");
				}
			}
		}

		internal Object GetCurrent()
		{
			if (m_nodeQueue.Count != 0) 
			{
				// get the next node from the queue, this will be the one that we return
				RDFNode next = (RDFNode) m_nodeQueue[0];
				m_nodeQueue.RemoveAt(0);

				// is this the default value?
				if (HasDefaultValue()  &&  m_defaultValue.Equals( next )) 
				{
					m_defaultValueSeen = true;
				}

				// now add the relations of the node to the end of the queue
				if (next is Resource) 
				{
					ExpandQueue( (Resource) next );
				}

				// answer the next node in the interation
				return next;
			}
			else if (HasDefaultValue()  &&  !m_defaultValueSeen) 
			{
				// return the default value for this iterator
				m_defaultValueSeen = true;
				return m_defaultValue;
			}
			else 
			{
				// no more nodes, so this is an error
				throw new System.Exception( "Tried to access next() element from empty property iterator" );
			}
		}


		public void Reset()
		{
		}

		/**
		 * Set the default value for this iteration, which will be a value that
		 * is guaranteed to be returned as a member of the iteration.  To guarantee
		 * that the default value is only returned if it has not already been
		 * returned by the iterator, setting the default value should occur before
		 * the first call to {@link #next}.
		 *
		 * @param defaultValue The default value for the iteration, or null for
		 *                     there to be no default value.  The default default
		 *                     value is null.
		 */
		public void SetDefaultValue( Object defaultValue ) 
		{
			m_defaultValue = defaultValue;
		}


		/**
		 * Answer true if this iteration has a default value.
		 *
		 * @return true if there is a default value
		 */
		public bool HasDefaultValue() 
		{
			return m_defaultValue != null;
		}



		// Internal implementation methods
		//////////////////////////////////


		/**
		 * Queue a node onto the evaluation queue, but only if it has not already been evaluated
		 * (to avoid loops).
		 *
		 * @param node The node to add to the queue
		 */
		private void Enqueue( RDFNode node ) 
		{
			if (!m_seen.ContainsKey( node )) 
			{
				// mark this node as seen
				m_seen[node] = true;

				// add it to the queue
				m_nodeQueue.Add( node );
				//m_nodeQueue.AddLast( node );

				// also queue up the equivalens to the node
				if (GetUseEquivalence()  &&  node is DAMLCommon) 
				{
					for (IEnumerator i = ((DAMLCommon) node).GetEquivalentValues();  i.MoveNext(); ) 
					{
						Enqueue( (RDFNode) i.Current );
					}
				}
			}
		}

		/**
		 * Expand the queue with the arcs to/from the given resource
		 *
		 * @param r The resource we're expanding
		 */
		protected void ExpandQueue( Resource r ) 
		{
			try 
			{
				// add all outgoing arcs if we're at the root or the predicate is transitive
				if (m_pred != null  &&  (m_transitive  ||  IsRoot( r ))) 
				{
					// we want all the related items from this node
					for (IEnumerator i = GetStatementObjects( r );
						i.MoveNext();
						Enqueue( (RDFNode) i.Current ));
				}

				// if we know the inverse, we also add the incoming arcs to this node
				if (m_inverse != null  &&  (m_transitive  ||  IsRoot( r ))) 
				{
					// find statements whose predicate is m_inverse and object is current node
					for (IEnumerator i = GetStatementSubjects( r );
						i.MoveNext();
						Enqueue( (RDFNode) i.Current ));
				}
			}
			catch (System.Exception e) 
			{
				//Log.severe( "RDF exception while traversing graph: " + e, e );
				Console.Error.WriteLine("RDF exception while traversing graph: " + e);
			}
		}


		/**
		 * Answer true if the given resource was one of the starting points.
		 *
		 * @param r A resource
		 * @return True if resource r was one of the roots of this iteration.
		 */
		protected bool IsRoot( Resource r ) 
		{
			//bool isRoot = false;

			// started with a set of roots
			return m_roots.Contains( r );
		}


		/**
		 * Answer an iterator over objects of statements with res
		 * as subject, and m_pred or one of its equivalent properties as property,
		 * including all of the equivalent values to the object.
		 *
		 * @param res A resource
		 * @return An iterator over the object of any statements whose subject is res.
		 * @exception RDFException if an RDF processing error occurs.
		 */
		protected IEnumerator GetStatementObjects( Resource res )
		{
			IEnumerator i = null;

			if (GetUseEquivalence()  &&  res is DAMLCommon) 
			{
				// consider equivalents for res and for m_pred
				for (IEnumerator j = m_predEquivs.GetEnumerator();   j.MoveNext();  ) 
				{
					// get an iterator for the values of the resource with this predicate
					IEnumerator pIter = new PropertyEnumerator( res, (Property) j.Current, null, m_transitive, false, false, m_model );

					// build a joint iterator
					i = (i == null) ? pIter  : new ConcatenatedEnumerator( pIter,  i );
				}
			}
			else 
			{
				// don't use equivalents
				if (m_model != null) 
				{
					// we have a model to query, so use it to get the object of the triple
					//i = new NodeIteratorWrapper( m_model.ListObjectsOfProperty( res, m_pred ) );
					// Our impl should behave properly
					i = m_model.ListObjectsOfProperty( res, m_pred );
				}
				else 
				{
					// no model (can occur when using built-in constants from vocab)
					i = new ArrayList().GetEnumerator();
				}
			}

			return i;
		}


		/**
		 * Answer an iterator over subjects of statements with res (or one of its equivalents)
		 * as object, and m_inverse or one of its equivalent properties as property.
		 *
		 * @param res A resource
		 * @return An iterator over the subject of any statements whose object is res.
		 * @exception RDFException if an RDF processing error occurs.
		 */
		protected IEnumerator GetStatementSubjects( Resource res )
		{
			IEnumerator i = null;

			if (GetUseEquivalence()  &&  res is DAMLCommon) 
			{
				// consider equivalents for res and for m_pred
				for (IEnumerator j = m_inverseEquivs.GetEnumerator();   j.MoveNext();  ) 
				{
					// get an iterator for the values of the resource with this predicate
					IEnumerator pIter = new PropertyEnumerator( res, null, (Property) j.Current, m_transitive, false, false, m_model );

					// build a joint iterator
					i = (i == null) ? pIter  : new ConcatenatedEnumerator( pIter,  i );
				}
			}
			else 
			{
				// don't use equivalents
				if (m_model != null) 
				{
					// we have a model to query, so use it to get the subject of the triple
					//i = new ResIteratorWrapper( m_model.ListSubjectsWithProperty( m_inverse, res ) );
					// No need to wrap this Enumerator because it will return objects when treated
					// like an IEnumerator.
					i = m_model.ListSubjectsWithProperty( m_inverse, res );
				}
				else 
				{
					// no model (can occur when using built-in constants from vocab)
					i = new ArrayList().GetEnumerator();
				}
			}

			return i;
		}


		/**
		 * Cache the equivalent properties to the principal properties and their inverses we
		 * are using
		 */
		protected void CachePropertyEquivs() 
		{
			if (GetUseEquivalence()) 
			{
				// store the equivalent properties to m_pred
				if (m_pred != null) 
				{
					if (m_pred is DAMLProperty) 
					{
						// daml property - we know about equivalence classes for these
						for (IEnumerator i = ((DAMLProperty) m_pred).GetEquivalentValues();  i.MoveNext(); ) 
						{
							CacheProperty( m_predEquivs, (Property) i.Current );
						}
					}
					else 
					{
						// normal rdf property, is its own equiv
						CacheProperty( m_predEquivs, m_pred );
					}
				}

				// store the equivalent properties to m_inverse
				if (m_inverse != null) 
				{
					if (m_inverse is DAMLProperty) 
					{
						// daml property - we know about equivalence classes for these
						for (IEnumerator i = ((DAMLProperty) m_inverse).GetEquivalentValues();  i.MoveNext(); ) 
						{
							CacheProperty( m_inverseEquivs, (Property) i.Current );
						}
					}
					else 
					{
						// normal rdf property, is its own equiv
						CacheProperty( m_inverseEquivs, m_inverse );
					}
				}
			}
		}


		/**
		 * Add the given property, and it sub-properties, to the set of cached properties.
		 *
		 * @param s A set
		 * @param p A property to add to the set s.
		 */
		protected void CacheProperty( IList s, Property p ) 
		{
			if (!s.Contains(p)) s.Add( p );

			// also add the sub-properties of this property: if the sub-prop holds,
			// this property holds also
			// check for p being 'subPropertyOf' to avoid infinite regress!
			if (p is DAMLProperty  &&  !p.GetLocalName().Equals( DAML_OIL._subPropertyOf.GetLocalName() )) 
			{
				for (IEnumerator i = ((DAMLProperty) p).GetSubProperties();  i.MoveNext(); ) 
				{
					if (!s.Contains(i.Current)) s.Add( i.Current );
				}
			}
		}


		/**
		 * Attempt to determine which model we are working on
	 *
	 * @param model A model to use, or null
	 * @param root The root resource, or null
	 * @param pred The predicate we are using, or null
	 * @param inverse The inverse of pred, or null
	 */
		protected void SetModel( Model model, Resource root, Property pred, Property inverse ) 
		{
			if (model != null) 
			{
				m_model = model;
			}
			else if (root != null  &&  root.GetModel() != null) 
			{
				m_model = root.GetModel();
			}
			else if (pred != null  &&  pred.GetModel() != null) 
			{
				m_model = pred.GetModel();
			}
			else if (inverse != null  &&  inverse.GetModel() != null) 
			{
				m_model = inverse.GetModel();
			}
		}



		/**
		 * Answer true if we are computing with equivalence classes at the moment.
		 *
		 * @return True if equivlance is to be computed.
		 */
		protected bool GetUseEquivalence() 
		{
			// if we have a model, it must have equivalence switched on
			// if no model, default to relying on the m_useEquivalence setting
			bool modelEquivFlag = m_model == null  ||
				!(m_model is DAMLModel) ||
				((DAMLModel) m_model).UseEquivalence;

			return m_useEquivalence  &&  modelEquivFlag;
		}

	


		/**
		 * Traverse the graph give the starting conditions that have been stored
		 * in various instance variables, starting from the given root resource.
		 *
		 * @param root A resource
		 */
		internal void Traverse( Resource root ) 
		{
			m_roots.Add( root );

			// use this model if we don't have one already
			if (m_model == null  &&  root.GetModel() != null) 
			{
				// keep a reference to the model if we find one
				m_model = root.GetModel();
			}

			// the root only goes on the queue if the relation is reflexive
			if (m_isReflexive) 
			{
				Enqueue( root );
			}
			else 
			{
				// otherwise, we'll just start with the relations of the root
				ExpandQueue( root );
			}
		}


		/**
		 * Traverse the graph give the starting conditions that have been stored
		 * in various instance variables, starting from the given root resources.
		 *
		 * @param roots An iteration of resources
		 */
		internal void Traverse( IEnumerator roots ) 
		{
			while (roots.MoveNext()) 
			{
				Traverse( (Resource) roots.Current );
			}
		}


		/**
		 * Initialises key member variables.
		 *
		 * @param root The root resource from whence to start iterating over the closure of pred
		 * @param pred The property to iterate over
		 * @param inverse The inverse of pred, or null if pred has no inverse. The inverse is used
		 *                to include resource y in the iteration if P' = inverse(P) and y P' x.
		 * @param isTransitive If true, the property is transitive
		 * @param isReflexive If true, the property is reflexive (so, the root resource will be included
		 *                    in the iteration).
		 * @param useEquivalence If true, equivalence between DAML values will be included in the
		 *                       iteration (unless the model containing the DAML values has equivalence
		 *                       switched off via {@link com.hp.hpl.jena.daml.DAMLModel#setUseEquivalence}.
		 * @param model The model that is the source of statements to traverse.
		 */
		internal void Initialise( Resource root, Property pred, Property inverse, bool isTransitive,
			bool isReflexive, bool useEquivalence, Model model ) 
		{
			m_pred = pred;
			m_inverse = inverse;
			m_transitive = isTransitive;
			m_useEquivalence = useEquivalence;
			m_isReflexive = isReflexive;

			SetModel( model, root, pred, inverse );
			CachePropertyEquivs();
		}


		//==============================================================================
		// Inner class definitions
		//==============================================================================
	}
}
