// Copyright (C) 2000, 2001 Jason Diamond <http://injektilo.org/>

// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:

// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

namespace Injektilo.RDF
{
	using System.Xml;

		/// <summary>
		///    Represents a statement reader that provides forward only access
		///    to a serialized RDF model.
		/// </summary>
	public abstract class RdfReader : XmlReader
	{
			/// <summary>
			///    Gets the base URI for the RDF document, if known.
			/// </summary>
		public abstract string DocumentURI
		{
            get;
		}

			/// <summary>
			///    Indicates that the current "node" is an RDF statement.
			/// </summary>
		public abstract bool IsStatement
		{
			get;	
		}

			/// <summary>
			///    Document me.
			/// </summary>
		public abstract RdfStatement Statement
		{
			get;
		}

			/// <summary>
			///    Gets the type of the current statement's subject.
			/// </summary>
			/// <seealso cref="RdfSubjectType"/>
		public abstract RdfSubjectType SubjectType 
		{
			get;
		}

			/// <summary>
			///    Gets the current statement's subject.
			/// </summary>
		public abstract string Subject
		{
			get;
		}

			/// <summary>
			///    Gets the current statement's predicate.
			/// </summary>
		public abstract string Predicate
		{
			get;
		}

			/// <summary>
			///    Gets the current object's index into the current subject
			///    (which should be an rdfs:Container) if the current predicate is an
			///    rdfs:ContainerMembershipProperty (rdf:_n) or 0 if it's not.
			/// </summary>
		public abstract int Index
		{
			get;
		}

			/// <summary>
			///    Gets the current statement's object type.
			/// </summary>
			/// <seealso cref="RdfObjectType"/>
		public abstract RdfObjectType ObjectType
		{
			get;
		}

			/// <summary>
			///    Gets the current statements's object.
			/// </summary>
		public abstract string Object
		{
			get;
		}
	}

		/// <summary>
		///    Represents the type of subjects a statement may contain.
		/// </summary>
		/// <seealso cref="RdfReader.SubjectType"/>
	public enum RdfSubjectType
	{
			/// <summary>
			///    Indicates the subject was specified using an about or ID attribute.
			/// </summary>
		URI,
			/// <summary>
			///    Indicates the subject was not specified.
			/// </summary>
		Anonymous,
			/// <summary>
			///    Indicates the subject was specified using the aboutEach attribute.
			/// </summary>
		Distributed,
			/// <summary>
			///    Indicates the subject was specified using the aboutEachPrefix attribute.
			/// </summary>
		Prefix
	}

		/// <summary>
		///    Represents the type of objects a statement may contain.
		/// </summary>
		/// <seealso cref="RdfReader.ObjectType"/>
	public enum RdfObjectType
	{
			/// <summary>
			///    Indicates the object is an identifiable resource.
			/// </summary>
		URI,
			/// <summary>
			///    Indicates the object is a simple string literal.
			/// </summary>
		Literal,
			/// <summary>
			///    Indicates the object is well-formed XML.
			/// </summary>
			/// <remarks>
			///    Objects containing well-formed XML (parseType="Literal")
			///    are not yet supported.
			/// </remarks>
		XML
	}

		/// <summary>
		///    RDF Model and Syntax namespace, element, attribute, and value constants.
		/// </summary>
	public class RdfSyntax
	{
		public const string URI               = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
		public const string RDF               = "RDF";
		public const string Description       = "Description";
		public const string ID                = "ID";
		public const string About             = "about";
		public const string AboutEach         = "aboutEach";
		public const string AboutEachPrefix   = "aboutEachPrefix";
		public const string BagID             = "bagID";
		public const string Resource          = "resource";
		public const string Type              = "type";
		public const string Value             = "value";
		public const string ParseType         = "parseType";
		public const string ParseTypeLiteral  = "Literal";
		public const string ParseTypeResource = "Resource";
		public const string ParseTypeDamlCollection = "daml:collection";
		public const string Bag               = "Bag";
		public const string Seq               = "Seq";
		public const string Alt               = "Alt";
		public const string Li                = "li";
		public const string Subject           = "subject";
		public const string Predicate         = "predicate";
		public const string Object            = "object";
	}
	public class RdfsSyntax
	{
		public const string URI = "http://www.w3.org/2000/01/rdf-schema#";
		public const string Label = "label";
	}
}
