//#define DEBUG_MODEL
// Copyright (C) 2000, 2001 Jason Diamond <http://injektilo.org/>

// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:

// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System.Collections;
using System;  // XXX
using System.IO;
using System.Diagnostics;
using System.Xml;
using System.Net;
using MSXML2;
using BBN.RDF;
#if DEBUG_MODEL
using LogSupport;
#endif

namespace Injektilo.RDF
{

	/// <summary>
	/// This delegate can be used to handle protocol errors encountered while trying to read using a WebRequest Object.
	/// If a protocol error occurs and a ProtocolErrorHandler was set for this model, the handler is called with the
	/// original WebRequest, the HttpStatusCode returned (from the WebException), and the WebException that occurred.
	/// The handler should return a new WebRequest (you can not reuse the old one) that will be used to try again.
	/// If the handler returns null, the read function will throw the WebException.  It should return null when
	/// the status code is not one of the errors that the handler can handle.
	/// </summary>
	public delegate System.Net.WebRequest ProtocolErrorHandler(WebRequest request, HttpStatusCode code, WebException exception);

	public class RdfModel
	{
		ArrayList statements = new ArrayList();
		ArrayList distributedStatements = new ArrayList();
		RdfNodeMap subjects = new RdfNodeMap();
		RdfNodeMap predicates = new RdfNodeMap();
		RdfNodeMap objects = new RdfNodeMap();
		FileInfo fileInfo; // stores FileInfo about a source file for this model(ie. mod time)
		ProtocolErrorHandler _protocolErrorHandler;
		
		public RdfModel()
		{
		}

		public RdfModel LoadStatements(string uri)
		{
			RdfModel m = new RdfModel();
			m.Load(uri);
			return m;
		}

		public ProtocolErrorHandler ProtocolErrorHandler
		{
			get 
			{
				return _protocolErrorHandler;
			}
			set
			{
				_protocolErrorHandler = value;
			}
		}

		public void Load(System.IO.TextReader reader, string b)
		{
			XmlValidatingReader xmlReader = new XmlValidatingReader(new XmlTextReader(b, reader));
			xmlReader.ValidationType = ValidationType.None;
			RdfReader rdfReader = new RdfXmlReader(xmlReader, b);
			Load(rdfReader);
		}

		public void Load(System.IO.Stream stream, string b)
		{
			XmlValidatingReader xmlReader = new XmlValidatingReader(new XmlTextReader(b, stream));
			xmlReader.ValidationType = ValidationType.None;
			RdfReader rdfReader = new RdfXmlReader(xmlReader, b);
			Load(rdfReader);
		}

		public void Load(WebRequest req)
		{
			WebRequest r = req;
			bool done = false;
			while (!done) 
			{
				try 
				{
					System.Net.WebResponse response = r.GetResponse();
					try 
					{
						Stream stream = response.GetResponseStream();
						Load(stream, r.RequestUri.AbsoluteUri);
					} 
					finally 
					{
						response.Close();
						done = true;
					}
				}
				catch (System.Net.WebException e) 
				{
					// If this is a ProtocolError try the handler
					if (e.Status == WebExceptionStatus.ProtocolError &&
						e.Response is HttpWebResponse &&
						_protocolErrorHandler != null) 
					{
						r = _protocolErrorHandler(r, ((HttpWebResponse)e.Response).StatusCode, e);	
						if (r == null) 
						{
							throw e;
						} 
						// Pop back out and try with the new request.
					} 
					else 
					{
						throw e;
					}
				}
			}
		}

		public void Load(System.IO.TextReader reader, string b, string lang)
		{
			XmlValidatingReader xmlReader = new XmlValidatingReader(new XmlTextReader(b, reader));
			xmlReader.ValidationType = ValidationType.None;
			RdfReader rdfReader = new RdfXmlReader(xmlReader, b);
			Load(rdfReader);
		}

		public void Load(System.IO.Stream stream, string b, string lang)
		{
			XmlValidatingReader xmlReader = new XmlValidatingReader(new XmlTextReader(b, stream));
			xmlReader.ValidationType = ValidationType.None;
			RdfReader rdfReader = new RdfXmlReader(xmlReader, b);
			Load(rdfReader);
		}

		public void Load(string uri)
		{
			RdfReader rdfReader = new RdfXmlReader(uri);
			Load(rdfReader);
		}

		public void Load(string file, string b)
		{
			TextReader tr = new StreamReader(file);
			Load(tr, b);
		}

		public void LoadFromString(string xml, string b)
		{
			TextReader tr = new StringReader(xml);
			Load(tr, b);
		}

		protected void Load(RdfReader rdfReader) 
		{

			while (rdfReader.Read())
			{
				RdfStatement statement = rdfReader.Statement;

				if (statement.SubjectType == RdfSubjectType.Distributed)
				{
					distributedStatements.Add(statement);
				}
				else
				{
					Add(rdfReader.Statement);
				}
			}

			rdfReader.Close();

			Distribute();
		}

		/// <summary>
		/// Performs the entity substitutions before passing the content to the
		/// .net XML parser.  This gets around the entity/namespace bugs, but
		/// takes a little longer and uses more resources.
		/// </summary>
		/// <param name="uri">The URI of the content to parse and load.</param>
		public void CleanLoad(string uri)
		{
			CleanLoad(uri, uri);
		}

		/// <summary>
		/// Performs the entity substitutions before passing the content to the
		/// .net XML parser.  This gets around the entity/namespace bugs, but
		/// takes a little longer and uses more resources.
		/// </summary>
		/// <param name="file">The file (or URI) of the content to parse and load.</param>
		/// <param name="b">The base URI used to resolve relative URIs.</param>
		public void CleanLoad(string file, string b)
		{
			try
			{
				// Create FreeThreadedDOMDocument
				MSXML2.FreeThreadedDOMDocument oStyleSheet = new FreeThreadedDOMDocumentClass();

				// Load the XSLT style sheet
				oStyleSheet.async = false;
				string identXML = "<?xml version=\"1.0\" ?><xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" xmlns:daml=\"http://www.daml.org/2001/03/daml+oil#\" xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><xsl:output method=\"xml\" encoding=\"UTF-8\" /><xsl:template name=\"process-list\"><xsl:param name=\"list\" /><xsl:if test=\"count($list) &gt; 0\"><daml:List xmlns:daml=\"http://www.daml.org/2001/03/daml+oil#\"><daml:first><xsl:apply-templates select=\"$list[1]\" /></daml:first><xsl:choose><xsl:when test=\"count($list) &gt; 1\"><daml:rest><xsl:call-template name=\"process-list\"><xsl:with-param name=\"list\" select=\"$list[position() &gt; 1]\" /></xsl:call-template></daml:rest></xsl:when><xsl:otherwise><daml:rest rdf:resource=\"http://www.daml.org/2001/03/daml+oil#nil\" /></xsl:otherwise></xsl:choose></daml:List></xsl:if></xsl:template><xsl:template match=\"@rdf:parseType\" mode=\"attr\"/><xsl:template match=\"@*\" mode=\"attr\"><xsl:copy/></xsl:template><xsl:template match=\"*[@rdf:parseType='daml:collection']\" ><xsl:copy><xsl:apply-templates select=\"@*\" mode=\"attr\"/><xsl:call-template name=\"process-list\"><xsl:with-param name=\"list\" select=\"*\" /></xsl:call-template></xsl:copy></xsl:template><xsl:template match=\"node()|@*\"><xsl:copy><xsl:apply-templates select=\"node()|@*\" /></xsl:copy></xsl:template></xsl:stylesheet>";
				//string identXML = "<?xml version=\"1.0\"?><xsl:stylesheet xmlns:xsl='http://www.w3.org/1999/XSL/Transform' version=\"1.0\" > <xsl:template match=\"/\"> <xsl:copy-of select=\"*\" /> </xsl:template> </xsl:stylesheet>";
				oStyleSheet.loadXML(identXML);

				// Create the XSLTemplate object
				MSXML2.XSLTemplate oXSLT = new XSLTemplateClass();

				// Link the XSLT style sheet to the XSLTemplate object
				oXSLT.stylesheet = oStyleSheet.documentElement;

				// Create the XSLProcessor object
				MSXML2.IXSLProcessor oXSLTProc = oXSLT.createProcessor();

				// Load XML source in DOMDocument
				//DOMDocument40 oXMLSource = new DOMDocument40Class();
				//oXMLSource.async = false;
				//oXMLSource.validateOnParse = false;
				//oXMLSource.load("c:/daml/test/bug1.daml");			

				// Assign XML source to XSLProcessor object
				// Can be DOMDocument, URI string, IXMLDOMNode, or anything that implements
				// ISequentialStream, IStream, IPersistStream
				//oXSLTProc.input = oXMLSource;
				oXSLTProc.input = file;

				// Invoke the transformation
				oXSLTProc.transform();

				// Use output of transformation
				string output = (string)oXSLTProc.output;
				//Trace.WriteLine(output.IndexOf("UTF-16"));
				output = output.Replace("UTF-16", "UTF-8");
				//Trace.WriteLine(output);
				TextReader reader = new StringReader(output);
				Load(reader, b);
			} 
			catch (System.Runtime.InteropServices.COMException e)
			{
				//throw new System.Xml.XmlException("A COMException occurred while doing identity xslt.", e);
				throw e;
			}
		}

		/// <summary>
		/// Performs the entity substitutions before passing the content to the
		/// .net XML parser.  This gets around the entity/namespace bugs, but
		/// takes a little longer and uses more resources.
		/// </summary>
		/// <param name="file">The file (or URI) of the content to parse and load.</param>
		/// <param name="b">The base URI used to resolve relative URIs.</param>
		

		public void CleanLoadFromString(string xml, string b)
		{
			try
			{
				// Create FreeThreadedDOMDocument
				MSXML2.FreeThreadedDOMDocument oStyleSheet = new FreeThreadedDOMDocumentClass();

				// Load the XSLT style sheet
				oStyleSheet.async = false;
				//string identXML = "<?xml version=\"1.0\"?><xsl:stylesheet xmlns:xsl='http://www.w3.org/1999/XSL/Transform' version=\"1.0\" > <xsl:template match=\"/\"> <xsl:copy-of select=\"*\" /> </xsl:template> </xsl:stylesheet>";
				string identXML = "<?xml version=\"1.0\" ?><xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" xmlns:daml=\"http://www.daml.org/2001/03/daml+oil#\" xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><xsl:output method=\"xml\" encoding=\"UTF-8\" /><xsl:template name=\"process-list\"><xsl:param name=\"list\" /><xsl:if test=\"count($list) &gt; 0\"><daml:List xmlns:daml=\"http://www.daml.org/2001/03/daml+oil#\"><daml:first><xsl:apply-templates select=\"$list[1]\" /></daml:first><xsl:choose><xsl:when test=\"count($list) &gt; 1\"><daml:rest><xsl:call-template name=\"process-list\"><xsl:with-param name=\"list\" select=\"$list[position() &gt; 1]\" /></xsl:call-template></daml:rest></xsl:when><xsl:otherwise><daml:rest rdf:resource=\"http://www.daml.org/2001/03/daml+oil#nil\" /></xsl:otherwise></xsl:choose></daml:List></xsl:if></xsl:template><xsl:template match=\"@rdf:parseType\" mode=\"attr\"/><xsl:template match=\"@*\" mode=\"attr\"><xsl:copy/></xsl:template><xsl:template match=\"*[@rdf:parseType='daml:collection']\" ><xsl:copy><xsl:apply-templates select=\"@*\" mode=\"attr\"/><xsl:call-template name=\"process-list\"><xsl:with-param name=\"list\" select=\"*\" /></xsl:call-template></xsl:copy></xsl:template><xsl:template match=\"node()|@*\"><xsl:copy><xsl:apply-templates select=\"node()|@*\" /></xsl:copy></xsl:template></xsl:stylesheet>";
				oStyleSheet.loadXML(identXML);

				// Create the XSLTemplate object
				MSXML2.XSLTemplate oXSLT = new XSLTemplateClass();

				// Link the XSLT style sheet to the XSLTemplate object
				oXSLT.stylesheet = oStyleSheet.documentElement;

				// Create the XSLProcessor object
				MSXML2.IXSLProcessor oXSLTProc = oXSLT.createProcessor();

				// Load XML source in DOMDocument
				DOMDocument oXMLSource = new DOMDocumentClass();
				oXMLSource.async = false;
				oXMLSource.validateOnParse = false;
				oXMLSource.loadXML(xml);			

				// Assign XML source to XSLProcessor object
				// Can be DOMDocument, URI string, IXMLDOMNode, or anything that implements
				// ISequentialStream, IStream, IPersistStream
				oXSLTProc.input = oXMLSource;

				// Invoke the transformation
				oXSLTProc.transform();

				// Use output of transformation
				//Trace.WriteLine(oXSLTProc.output);
				string output = (string)oXSLTProc.output;
				output = output.Replace("UTF-16", "UTF-8");
				//Trace.WriteLine(output);
				TextReader reader = new StringReader(output);

				Load(reader, b);
			} 
			catch (System.Runtime.InteropServices.COMException e)
			{
				throw new System.Xml.XmlException("A COMException occurred while doing identity xslt.", e);
			}
		}


		/// <summary>
		/// Add the statements of this RdfModel to the given Jena Model.
		/// </summary>
		/// <param name="m">Model to add the statements to</param>
		/// <returns>The Jena Model</returns>
		public Model ConvertStatements(Model m)
		{
			Statement newStatement;
			Resource sub;
			Property pred;
			RDFNode obj;
			foreach(RdfStatement s in statements)
			{
				sub = m.GetResource(s.Subject);
				pred = m.GetProperty(s.Predicate);
				if (s.ObjectType.Equals(RdfObjectType.URI))
					obj = m.GetResource(s.Object);
				else
					obj = m.CreateLiteral(s.Object);
				if (s.Id == null)
					newStatement = m.CreateStatement(sub,pred,obj,s.Id,s.Source);
				else 
				{
					newStatement = m.CreateStatement(sub,pred,obj,s.Id,s.Source);
				}
				m.Add(newStatement);
			}
			return m;
		}


		/// <summary>write (serialize) the model to the specified file</summary>
		/// <remarks>This is a simple implementation missing some desirable features,
		/// e.g. shared namespaces, handling of anonymous ndoes, automatic determination of CDATA.  
		/// It's not very efficient to build the whole XML model and then serialize it; 
		/// this could be done incrementally.
		/// Output isn't indented.</remarks>
		/// <author>Mike Dean, mdean@bbn.com</author>
		public void Write(string path)
		{
			Hashtable namespaces = new Hashtable(); // stores the prefixes indexed by URI
			// start off by putting the rdf namespace in the hashtable
			namespaces[RdfSyntax.URI] = "rdf";
			System.Xml.XmlDocument document = new System.Xml.XmlDocument();
			System.Xml.XmlDeclaration declaration = document.CreateXmlDeclaration("1.0", "ISO-8859-1", "");
			document.AppendChild(declaration);
			System.Xml.XmlElement rdf = document.CreateElement("rdf", RdfSyntax.RDF, RdfSyntax.URI);
			document.AppendChild(rdf);
			foreach (string subject in subjects.GetKeys())
			{
				System.Xml.XmlElement description = document.CreateElement("rdf", RdfSyntax.Description, RdfSyntax.URI);
				// Make sure that the statement object does not appear as the subject of a statement
				// Change to the statement ID instead.
				// TBD
				description.SetAttribute(RdfSyntax.About, RdfSyntax.URI, subject);
				rdf.AppendChild(description);
				IList statements = subjects.GetStatements(subject);
				string nodeName, nodePrefix, nodeNS;
				foreach (RdfStatement statement in statements)
				{
					string predicate = statement.Predicate;
					int hash = predicate.LastIndexOf('#');
					nodeName = predicate.Substring(hash+1);
					nodeNS = predicate.Substring(0, hash+1);
					// if the namespace is new, add it to the table and assign a prefix
					if(namespaces[nodeNS] == null)
					{
						namespaces[nodeNS] = "ns" + namespaces.Count;
						rdf.SetAttribute("xmlns:"+(string)namespaces[nodeNS], nodeNS);
						//Trace.WriteLine("Added namespace: ns" + (namespaces.Count-1) + " = " + nodeNS);
					}
					// now we can get the prefix from the hashtable
					nodePrefix = (string)namespaces[nodeNS];
					System.Xml.XmlElement element = document.CreateElement(nodePrefix, nodeName, nodeNS);
					if (statement.ObjectType == RdfObjectType.Literal)
						element.AppendChild(document.CreateTextNode(statement.Object));
					else
						element.SetAttribute(RdfSyntax.Resource, RdfSyntax.URI, statement.Object);
					// Add statement ID if one exists
					if (statement.Id != null)
					{
						int ihash = statement.Id.LastIndexOf('#');
						String idTag;
						if (ihash >= 0) 
						{
							idTag = statement.Id.Substring(ihash+1);
						} 
						else 
						{
							idTag = statement.Id;
						}
						element.SetAttribute(RdfSyntax.ID,RdfSyntax.URI, idTag);
					}
					description.AppendChild(element);
				}
			}
			System.IO.StreamWriter stream = new System.IO.StreamWriter(path);
			System.Xml.XmlWriter writer = new System.Xml.XmlTextWriter(stream);
			document.WriteContentTo(writer);
			stream.WriteLine();
			stream.Close();
		}
		private void Distribute()
		{
			ArrayList impliedStatements = new ArrayList();

			foreach (RdfStatement statement in distributedStatements)
			{
				IList members = subjects.GetStatements(statement.Subject);

				if (members != null)
				{
					foreach (RdfStatement member in members)
					{
						if (member.Index > 0)
						{
							impliedStatements.Add(
								new RdfStatement(
								RdfSubjectType.URI,
								member.Object,
								statement.Predicate,
								statement.Index,
								statement.ObjectType,
								statement.Object,
								statement.XmlLang,
								member.Id,
								member.Source)); // TBD
						}
					}
				}
			}

			foreach (RdfStatement statement in impliedStatements)
			{
				Add(statement);
			}
		}
		// Not actually going to use this method anymore.
		/* public void Add(Statement s)
	   {
		   RdfObjectType oType;
		   String val;
		   RDFNode n = s.GetObject();
		   if(typeof(Literal).Equals(n.GetType()))
		   {
			   oType = RdfObjectType.Literal;
			   val = ((Literal)n).ToString();
		   }
		   else
		   {
			   oType = RdfObjectType.URI;
			   val = ((Resource)n).ToString();
		   }
		   Add(new RdfStatement(RdfSubjectType.URI, s.GetSubject().GetURI(),
			   s.GetPredicate().GetURI(), 0, oType, val, "en-us"));*/

		public void Add(RdfStatement statement)
		{
			// first ensure that the URIs are well-formed
			string ns = statement.Subject.Substring(0, statement.Subject.IndexOf('#')+1);
			string name = statement.Subject.Substring(statement.Subject.LastIndexOf('#')+1);
			statement.Subject = ns+name;
			if(statement.ObjectType.Equals(RdfObjectType.URI))
			{
				ns = statement.Object.Substring(0, statement.Object.IndexOf('#')+1);
				name = statement.Object.Substring(statement.Object.LastIndexOf('#')+1);
				statement.Object = ns+name;
			}
			statements.Add(statement);
			subjects.Add(statement.Subject, statement);
			predicates.Add(statement.Predicate, statement);
			objects.Add(statement.Object, statement);
		}
		/// <summary>
		/// Removes the given statement from the model
		/// </summary>
		/// <param name="statement">RdfStatement to be removed</param>
		public void RemoveStatement(RdfStatement statement)
		{
			if(!ContainsSame(statement)) // statement not in model
				return;
			// remove from Statements
			foreach (RdfStatement s in this.GetStatementsWithSubject(statement.Subject))
			{
				if (statement.Predicate.Equals(s.Predicate) &&
					statement.Object.Equals(s.Object) &&
					statement.ObjectType.Equals(s.ObjectType) &&
					//statement.Subject.Equals(s.Subject) &&
					statement.SubjectType.Equals(s.SubjectType))
					Statements.Remove(s);					
			}
			// now remove from the RdfNodeMaps
			subjects.RemoveLike(statement.Subject, statement);
			predicates.RemoveLike(statement.Predicate, statement);
			objects.RemoveLike(statement.Object, statement);
		}			
		/// <summary>
		/// Searches model for a similar statement.
		/// </summary>
		/// <param name="statement">The statement to look for.</param>
		/// <returns>Returns true if there is a statement in the model with
		/// the same subject, predicate, and object.</returns>
		public bool ContainsSame(RdfStatement statement)
		{
			foreach (RdfStatement s in this.GetStatementsWithSubject(statement.Subject))
			{
				if (statement.Predicate.Equals(s.Predicate) &&
					statement.Object.Equals(s.Object) &&
					statement.ObjectType.Equals(s.ObjectType) &&
					//statement.Subject.Equals(s.Subject) &&
					statement.SubjectType.Equals(s.SubjectType))
					return true;
			}
			return false;
		}
		/// <summary>
		/// Sets the object value of the given RdfStatement
		/// </summary>
		/// <param name="statement">Statement to change</param>
		/// <param name="obj">new object value</param>
		public void SetStatementObject(RdfStatement statement, string obj)
		{
#if DEBUG_MODEL
			Logger.GetLogger().Log("SetStatementObject: " + statement.Object + "-->"+obj);
#endif
			if (this.ContainsSame(statement))
			{
				RdfStatement toRemove = null;
				foreach(RdfStatement s in Statements)
					if(s.SameAs(statement))
						toRemove = s;
#if DEBUG_MODEL
				Logger.GetLogger().Log("Changing...");
#endif
				// Remove Statement from Objects
				objects.Remove(toRemove.Object, toRemove);

				// Change the object of the Statement
				toRemove.Object = obj;

				// Add the statement to the Model
				objects.Add(toRemove.Object, toRemove);
#if DEBUG_MODEL
				foreach(RdfStatement s in Statements)
					Logger.GetLogger().Log(s.ToString() + "\n");
#endif
			}
		}
		/// <summary>
		/// Removes all statements with the given subject
		/// </summary>
		/// <param name="subject">URI of object to remove</param>
		public void RemoveObject(string subject)
		{
			// first need to get all statements where this is the subject
			ArrayList subStatements = (ArrayList)subjects.GetStatements(subject);
			if(subStatements == null || subStatements.Count < 1) return;
			ArrayList statements = (ArrayList)subStatements.Clone();

			foreach(RdfStatement statement in statements)
			{
				// now for each predicate in those statements, remove them from the predicates
				predicates.Remove(statement.Predicate, statement);
				// now for each object in those statements, remove them from the objects
				objects.Remove(statement.Object, statement);
				// now remove those statements from the subjects nodemap
				subjects.Remove(subject, statement);
				// now remove the actual statements
				this.statements.Remove(statement);
			}
		}

		public string GetRdfsLabel(string resourceUri)
		{
			IList stmts = this.GetStatementsWithPredicate(RdfsSyntax.URI+RdfsSyntax.Label);
			foreach(RdfStatement s in stmts)
				if(s.Subject.Equals(resourceUri))
					return s.Object;
			int index = resourceUri.IndexOf('#');
			if(index < 0)
				index = resourceUri.LastIndexOf('/');
			return resourceUri.Substring(index); // no label found, just return the partial uri
		}

		public IList Statements
		{
			get
			{
				return statements;
			}
		}

		public int Count
		{
			get
			{
				return statements.Count;
			}
		}
		/// <summary>
		/// Returns an IList of RdfStatements where subject == s
		/// </summary>
		/// <param name="s">Subject string to filter</param>
		/// <returns>IList of RdfStatements</returns>
		public IList GetStatementsWithSubject(string s)
		{
			if(subjects[s] != null)
				return subjects[s];
			else return new ArrayList() as IList;
		}
		/// <summary>
		/// Returns an IList of RdfStatement objects where predicate == s
		/// </summary>
		/// <param name="s">Predicate string to filter</param>
		/// <returns>IList of RdfStatements</returns>
		public IList GetStatementsWithPredicate(string s)
		{
			if(predicates[s] != null)
				return predicates[s];
			else return new ArrayList() as IList;
		}
		/// <summary>
		/// Returns an IList of RdfStatement objects where the object == s
		/// </summary>
		/// <param name="s">Object value to filter</param>
		/// <returns>IList of RdfStatements</returns>
		public IList GetStatementsWithObject(string s)
		{
			if(objects[s] != null)
				return objects[s];
			else return new ArrayList() as IList;
		}
		public FileInfo SourceFileInfo
		{
			get{return fileInfo;}
			set{fileInfo = value;}
		}

		private class RdfNodeMap
		{
			Hashtable nodes = new Hashtable();

			public IList this[string o]
			{
				get
				{
					return nodes[o] as IList;
				}
			}
			internal void Add(string node, RdfStatement statement)
			{
				ArrayList statements = nodes[node] as ArrayList;

				if (statements == null)
				{
					statements = new ArrayList();
					nodes[node] = statements;
				}

				statements.Add(statement);
			}

			internal IList GetStatements(string node)
			{
				return nodes[node] as IList;
			}

			internal void Remove(string node, RdfStatement statement)
			{
				ArrayList statements = (ArrayList)nodes[node];
				if(statements == null) return;
				statements.Remove(statement);
				// if this is the last statement for this node, remove the node from the hashtable
				if(statements.Count < 1)
					nodes.Remove(node);
			}
			internal void RemoveLike(string node, RdfStatement statement)
			{
				ArrayList statements = (ArrayList)nodes[node];
				if(statements == null) return;
				RdfStatement toRemove = null;
				foreach(RdfStatement s in statements)
					if(s.SameAs(statement))
						toRemove = s;
				statements.Remove(toRemove);
			}

			internal ICollection GetKeys()
			{
				return nodes.Keys;
			}
		}
	}
}
