// Copyright (C) 2000, 2001 Jason Diamond <http://injektilo.org/>

// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:

// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// #define DEBUG2

namespace Injektilo.RDF
{
	using System;
	using System.Collections;
	using System.Diagnostics;
	using System.Text;
	using System.Xml;
	using System.Xml.XPath;
	using System.Xml.Xsl;


	/// <summary>
	///    Represents a statement reader that provides forward only access
	///    to a serialized RDF model.
	/// </summary>
	public class RdfXmlReader : RdfReader
	{
		string documentURI;

		XmlReader xmlReader;

		Stack elements;
		RdfXmlElement top;

		Queue statements;
		RdfStatement current;

		uint lastAnonymousID;

		bool parseTypeLiteral;
		bool isStatement;
		int literalElements;

		/// <summary>
		///    Constructs an RdfXmlReader with the given URI.
		/// </summary>
		/// <param name="uri">
		///    The URI for the file containing the RDF to read.
		/// </param>
		public RdfXmlReader(string uri)
		{
			documentURI = new Uri(uri).ToString();
			//Console.WriteLine("in RdfXmlReader("+uri+")");

			// Hack to get around .net xml parser bug - do identity xslt first
			//XPathDocument xpathdoc = new XPathDocument(documentURI);
			//XslTransform xslt = new XslTransform();
			//xslt.Load("c:/daml/test/ident.xslt");
			// This doesn't return an XmlReader that's compatable with creating a new XmlValidatingReader.
			//XmlReader xmlreader = xslt.Transform(xpathdoc, null);
			//System.IO.Stream stream = new System.IO.MemoryStream();
			//xslt.Transform(xpathdoc, null, stream);

			XmlValidatingReader xmlv = new XmlValidatingReader(new XmlTextReader(documentURI));
			xmlv.ValidationType = ValidationType.None;
			Init(xmlv);

		}

		/// <summary>
		///    Constructs an RdfXmlReader with the given XmlReader.
		/// </summary>
		/// <param name="xmlReader">
		///    The XmlReader containing the RDF data to read.
		/// </param>
		public RdfXmlReader(XmlReader xmlReader)
		{
			// TBD - This is not going to work because "" will produce an error when you create a URI object.
			documentURI = null;
			Init(xmlReader);
		}

		/// <summary>
		///    Constructs an RdfXmlReader with the given XmlReader.
		/// </summary>
		/// <param name="xmlReader">
		///    The XmlReader containing the RDF data to read.
		/// </param>
		/// <param name="documentURI">
		///    The URI to use as the DocumentURI.
		/// </param>
		public RdfXmlReader(XmlReader xmlReader, string documentURI)
		{
			if (documentURI == null) 
			{
				this.documentURI = null;
			} 
			else 
			{
				this.documentURI = new Uri(documentURI).ToString();
			}
			//this.documentURI = documentURI;
			Init(xmlReader);
		}

		private void Init(XmlReader xmlReader)
		{
			this.xmlReader = xmlReader;

			elements = new Stack();
			statements = new Queue();

			lastAnonymousID = 0;

			parseTypeLiteral = false;
			isStatement = false;
		}

		public override bool Read()
		{
			current = null;

			bool moreXml = false;

			// If in the middle of a parseType="Literal" read the next
			// element and throw it away.
			if (parseTypeLiteral)
			{
				isStatement = false;

				if (moreXml = xmlReader.Read())
				{
					if (xmlReader.NodeType == XmlNodeType.Element)
					{
						if (!xmlReader.IsEmptyElement)
						{
							// We found a new element, keep track of how deep we are
							++literalElements;
						}
					}
					else if (xmlReader.NodeType == XmlNodeType.EndElement)
					{
						// closing element, decrease the depth.
						if (--literalElements == 0)
						{
							// Done with the parseType="Literal"
							// remove ???something???
							PopElement();
							PopElement();
							parseTypeLiteral = false;
						}
					}
				}
			}

			// regular parsing...
			if (!parseTypeLiteral)
			{
				isStatement = true;

				while (statements.Count < 1 && (moreXml = xmlReader.Read()))
				{
#if DEBUG2
					MyTrace();
#endif

					RdfXmlState state = top != null ? top.m_state : RdfXmlState.OutsideRDF;

					switch (xmlReader.NodeType)
					{
						case XmlNodeType.Element:
						switch (state)
						{
							case RdfXmlState.OutsideRDF:
								OutsideRDF_Element();
								break;
							case RdfXmlState.InRDF:
								InRDF_Element();
								break;
							case RdfXmlState.InDescription:
								InDescription_Element();
								break;
							case RdfXmlState.InProperty:
								InProperty_Element();
								break;
							case RdfXmlState.InLiteralProperty:
								break;
							case RdfXmlState.InResourceProperty:
								break;
							case RdfXmlState.InParseTypeLiteralProperty:
								break;
							case RdfXmlState.InParseTypeResourceProperty:
								InParseTypeResourceProperty_Element();
								break;
						}
							break;
						case XmlNodeType.Text:
						switch (state)
						{
							case RdfXmlState.OutsideRDF:
								break;
							case RdfXmlState.InRDF:
								break;
							case RdfXmlState.InDescription:
								break;
							case RdfXmlState.InProperty:
								InProperty_Text();
								break;
							case RdfXmlState.InLiteralProperty:
								break;
							case RdfXmlState.InResourceProperty:
								break;
							case RdfXmlState.InParseTypeLiteralProperty:
								break;
							case RdfXmlState.InParseTypeResourceProperty:
								break;
						}
							break;
						case XmlNodeType.EndElement:
						switch (state)
						{
							case RdfXmlState.OutsideRDF:
								OutsideRDF_EndElement();
								break;
							case RdfXmlState.InRDF:
								InRDF_EndElement();
								break;
							case RdfXmlState.InDescription:
								InDescription_EndElement();
								break;
							case RdfXmlState.InProperty:
								InProperty_EndElement();
								break;
							case RdfXmlState.InLiteralProperty:
								break;
							case RdfXmlState.InResourceProperty:
								InResourceProperty_EndElement();
								break;
							case RdfXmlState.InParseTypeLiteralProperty:
								break;
							case RdfXmlState.InParseTypeResourceProperty:
								InParseTypeResourceProperty_EndElement();
								break;
						}
							break;
						default:
							break;
					}
				}

				if (statements.Count > 0)
				{
					current = (RdfStatement)statements.Dequeue();
				}
			}

			return (current != null)
				|| moreXml;
		}

		private void CheckCurrent()
		{
			if (current == null)
			{
				throw new Exception("current is null");
			}
		}

		public override string DocumentURI
		{
			get
			{
				return documentURI;
			}
		}

		public override bool IsStatement
		{
			get
			{
				return isStatement;
			}
		}

		public override RdfStatement Statement
		{
			get
			{
				CheckCurrent();
				return current;
			}
		}

		public override RdfSubjectType SubjectType
		{
			get
			{
				CheckCurrent();
				return current.SubjectType;
			}
		}

		public override string Subject
		{
			get
			{
				CheckCurrent();
				return current.Subject;
			}
		}

		public override string Predicate
		{
			get
			{
				CheckCurrent();
				return current.Predicate;
			}
		}

		public override int Index
		{
			get
			{
				CheckCurrent();
				return current.Index;
			}
		}

		public override RdfObjectType ObjectType
		{
			get
			{
				CheckCurrent();
				return current.ObjectType;
			}
		}

		public override string Object
		{
			get
			{
				CheckCurrent();
				return current.Object;
			}
		}

		private void MyTrace()
		{
			if (top != null)
			{
				Trace.Write("  ");

				switch (top.m_state)
				{
					case RdfXmlState.OutsideRDF:
						Trace.Write("OutsideRDF");
						break;
					case RdfXmlState.InRDF:
						Trace.Write("InRDF");
						break;
					case RdfXmlState.InDescription:
						Trace.Write("InDescription");
						break;
					case RdfXmlState.InProperty:
						Trace.Write("InProperty");
						break;
					case RdfXmlState.InLiteralProperty:
						Trace.Write("InLiteralProperty");
						break;
					case RdfXmlState.InResourceProperty:
						Trace.Write("InResourceProperty");
						break;
					case RdfXmlState.InParseTypeLiteralProperty:
						Trace.Write("InParseTypeLiteralProperty");
						break;
					case RdfXmlState.InParseTypeResourceProperty:
						Trace.Write("InParseTypeResourceProperty");
						break;
				}
			}

			Trace.Write("  ");

			switch (xmlReader.NodeType)
			{
				case XmlNodeType.Element:
					Trace.WriteLine("Element: {0}", xmlReader.Name);
					break;
				case XmlNodeType.EndElement:
					Trace.WriteLine("EndElement: {0}", xmlReader.Name);
					break;
				case XmlNodeType.Text:
					Trace.WriteLine("Text");
					break;
				case XmlNodeType.Whitespace:
					Trace.WriteLine("Whitespace");
					break;
				case XmlNodeType.ProcessingInstruction:
					Trace.WriteLine("ProcessingInstruction");
					break;
				default:
					Trace.WriteLine("");
					break;
			}
		}

		private void PushElement(RdfXmlState state)
		{
#if DEBUG2
			Trace.WriteLine("PushElement");
#endif
			RdfXmlElement element = new RdfXmlElement();
			element.m_parent = top;
			top = element;
			top.m_state = state;
			elements.Push(top);
		}

		private void PopElement()
		{
#if DEBUG2
			Trace.WriteLine("PopElement");
#endif

			if (elements.Count > 0)
			{
				elements.Pop();
				top = elements.Count > 0 ? (RdfXmlElement)elements.Peek() : null;
			}
		}

		private void EnqueueStatement(
			RdfSubjectType subjectType,
			string subject,
			string predicate,
			int index,
			RdfObjectType objectType,
			string value,
			string id)
		{
#if DEBUG2
				if (id == null) 
				{
					Trace.WriteLine(
						"  Enqueueing Statement:\n   "+
						subjectType+":   "+
						subject+"\n   "+
						predicate+"\n   "+
						objectType+":   "+
						value);
				}
				else 
				{
					Trace.WriteLine(
						"  Enqueueing Statement (rdf:ID=\""+id+"\"):\n   "+
						subjectType+":   "+
						subject+"\n   "+
						predicate+"\n   "+
						objectType+":   "+
						value);
				}
#endif

			if (documentURI == null) 
			{
				subject = new Uri(subject).ToString();
			} 
			else 
			{
				subject = new Uri(new Uri(documentURI), subject).ToString();
			}


			if (objectType == RdfObjectType.URI)
			{
				if (documentURI == null) 
				{
					value = new Uri(value).ToString();
				} 
				else 
				{
					value = new Uri(new Uri(documentURI), value).ToString();
				}
			}

			if (predicate.Equals(RdfSyntax.URI + RdfSyntax.Li))
			{
				if (top.m_parent.m_state == RdfXmlState.InResourceProperty)
				{
					index = ++top.m_parent.m_parent.m_members;
				}
				else
				{
					index = ++top.m_parent.m_members;
				}

				predicate = RdfSyntax.URI + "_" + index;
			}
			else if (predicate.StartsWith(RdfSyntax.URI + "_"))
			{
				index = Int32.Parse(
					predicate.Substring(predicate.LastIndexOf('_') + 1));

				if (top.m_parent.m_state == RdfXmlState.InResourceProperty)
				{
					top.m_parent.m_parent.m_members = index;
				}
				else
				{
					top.m_parent.m_members = index;
				}
			}

			if (id != null)
			{
				if (id.IndexOf('#') < 0) id = "#"+id;
				if (documentURI == null)
				{
					id = new Uri(id).ToString();
				}
				else
				{
					id = new Uri(new Uri(documentURI), id).ToString();
				}
			}

			string xmlLang = xmlReader.XmlLang;

			statements.Enqueue(
				new RdfStatement(
				subjectType,
				subject,
				predicate,
				index,
				objectType,
				value,
				xmlLang,
				id,
				documentURI));
		}

		// DER - An Attempt to get around the problem where entities are not resolved in xmlns 
		// declarations that are declared in the same tag that they are used.
		/*
		private void Test_Entity()
		{
			Trace.WriteLine("In Test_Entity(): " + xmlReader.NamespaceURI+" "+xmlReader.Name);		
			for(int i = 0; i<xmlReader.AttributeCount; i++)
			{
				//Trace.WriteLine("Attr["+i+"]="+xmlReader.GetAttribute(i));
				xmlReader.MoveToAttribute(i);
				Trace.WriteLine("Attr["+i+"] "+xmlReader.Name+" = "+xmlReader.Value);
				xmlReader.MoveToElement();

			}
			int i1 = xmlReader.NamespaceURI.IndexOf('&');
			if (i1 >= 0) 
			{
				int i2 = xmlReader.NamespaceURI.IndexOf(';', i1);
				if (i2 > i1) 
				{
					// There's an entity to substitute.  Check the attributes for a namespace defn
					Trace.WriteLine(xmlReader.NamespaceURI.Substring(i1+1, (i2-i1-1)));
					Trace.WriteLine("Attr:" + xmlReader.GetAttribute("xmlns", xmlReader.NamespaceURI.Substring(i1+1, (i2-i1-1))));
					for(int i = 0; i<xmlReader.AttributeCount; i++)
					{
						//Trace.WriteLine("Attr["+i+"]="+xmlReader.GetAttribute(i));
						xmlReader.MoveToAttribute(i);
						Trace.WriteLine(xmlReader.Name+" "+xmlReader.Value);
						xmlReader.MoveToElement();

					}
				}
			}
		}
		*/

		private void OutsideRDF_Element()
		{
			if (xmlReader.NamespaceURI == RdfSyntax.URI &&
				xmlReader.LocalName == RdfSyntax.RDF)
			{
				PushElement(RdfXmlState.InRDF);
			}
		}

		private void OutsideRDF_EndElement()
		{
			PopElement();
		}

		private void InRDF_Element()
		{
			PushElement(RdfXmlState.InDescription);

			bool isEmptyElement = xmlReader.IsEmptyElement;

			DiscoverDescriptionURI();

			HandlePropertyAttributes();

			if (isEmptyElement)
			{
				PopElement();
			}
		}

		private bool IsValidID(string id)
		{
			bool result = false;

			if (id != null)
			{
				char c = id[0];

				if (Char.IsLetterOrDigit(c) ||
					c == '_' ||
					c == ':')
				{
					result = true;

					for (int i = 1, n = id.Length; result && (i < n); ++i)
					{
						c = id[i];

						if (!(Char.IsLetterOrDigit(c) ||
							c == '.' ||
							c == '-' ||
							c == '_' ||
							c == ':'))
						{
							result = false;
						}
					}
				}
			}

			return result;
		}

		private void DiscoverDescriptionURI()
		{
			int identifiers = 0;
			string id = null;
			string about = null;
			string aboutEach = null;
			string aboutEachPrefix = null;

			bool rdfElement = RdfSyntax.URI.Equals(xmlReader.NamespaceURI);

			while (xmlReader.MoveToNextAttribute())
			{
				string attributeNamespaceURI = xmlReader.NamespaceURI;
				string attributeLocalName = xmlReader.LocalName;
				string attributeValue = xmlReader.Value;

				if (rdfElement || RdfSyntax.URI.Equals(attributeNamespaceURI))
				{
					switch (attributeLocalName)
					{
						case RdfSyntax.ID:
							++identifiers;
							id = attributeValue;
							break;
						case RdfSyntax.About:
							++identifiers;
							about = attributeValue;
							break;
						case RdfSyntax.AboutEach:
							++identifiers;
							aboutEach = attributeValue;
							break;
						case RdfSyntax.AboutEachPrefix:
							++identifiers;
							aboutEachPrefix = attributeValue;
							break;
					}
				}
			}

			if (identifiers == 0)
			{
				top.m_subjectType = RdfSubjectType.Anonymous;
				top.m_name = GenerateAnonymousURI();
			}
			else if (identifiers > 1)
			{
				throw new Exception("too many identifiers");
			}
			else if (id != null)
			{
				if (!IsValidID(id))
				{
					throw new Exception(String.Format("bad ID attribute: {0}", id));
				}
		 
				if (documentURI == null)
				{
					throw new Exception("Document URI was not specified.");
				}

				top.m_subjectType = RdfSubjectType.URI;
				top.m_name = documentURI + "#" + id;
			}
			else if (about != null)
			{
				top.m_subjectType = RdfSubjectType.URI;

				if ("".Equals(about))
				{
					if (documentURI == null) 
					{
						throw new Exception("Document URI was not specified.");
					}
					top.m_name = documentURI;
				}
				else
				{
					top.m_name = about;
				}
			}
			else if (aboutEach != null)
			{
				top.m_subjectType = RdfSubjectType.Distributed;
				top.m_name = aboutEach;
			}
			else if (aboutEachPrefix != null)
			{
				top.m_subjectType = RdfSubjectType.Prefix;
				top.m_name = aboutEachPrefix;
			}

			xmlReader.MoveToElement();

			// if this isn't an rdf:Description element then report the type.
			if (!(rdfElement && RdfSyntax.Description.Equals(xmlReader.LocalName)))
			{
				EnqueueStatement(
					top.m_subjectType,
					top.m_name,
					RdfSyntax.URI + RdfSyntax.Type,
					0,
					RdfObjectType.URI,
					xmlReader.NamespaceURI + xmlReader.LocalName,
					null);
			}
		}

		private void HandlePropertyAttributes()
		{
			xmlReader.MoveToElement();

			bool rdfElement = RdfSyntax.URI.Equals(xmlReader.NamespaceURI);

			while (xmlReader.MoveToNextAttribute())
			{
				string attributeNamespaceURI = xmlReader.NamespaceURI;
				string attributeLocalName = xmlReader.LocalName;
				string attributeValue = xmlReader.Value;

				if ((rdfElement && "".Equals(attributeNamespaceURI))
					|| (RdfSyntax.URI.Equals(attributeNamespaceURI)))
				{
					// rdf:type objects are always resources.
					// this probably isn't allowed by the M&S.
					if (RdfSyntax.Type.Equals(attributeLocalName))
					{
						EnqueueStatement(
							top.m_subjectType,
							top.m_name,
							RdfSyntax.URI + RdfSyntax.Type,
							0,
							RdfObjectType.URI,
							attributeValue,
							null);
					}
					else if (RdfSyntax.Value.Equals(attributeLocalName))
					{
						EnqueueStatement(
							top.m_subjectType,
							top.m_name,
							RdfSyntax.URI + RdfSyntax.Value,
							0,
							RdfObjectType.Literal,
							attributeValue,
							null);
					}
					else if (attributeLocalName.StartsWith("_"))
					{
						int index = Int32.Parse(attributeLocalName.Substring(1));

						if (index < 1)
						{
							throw new Exception("bad member attribute");
						}
						else
						{
							EnqueueStatement(
								top.m_subjectType,
								top.m_name,
								RdfSyntax.URI + attributeLocalName,
								index,
								RdfObjectType.Literal,
								attributeValue,
								null);
						}
					}
					else
					{
						//throw new Exception(String.Format("unrecognized or out of context RDF attribute: {0}", attributeLocalName));
					}
				}
				else
				{
					EnqueueStatement(
						top.m_subjectType,
						top.m_name,
						attributeNamespaceURI + attributeLocalName,
						0,
						RdfObjectType.Literal,
						attributeValue,
						null);
				}
			}
		}

		private string GenerateAnonymousURI()
		{
			++lastAnonymousID;
			return "#" + lastAnonymousID;
		}

		private void InRDF_EndElement()
		{
			PopElement();
		}

		private void InDescription_Element()
		{
			string elementNamespaceURI = xmlReader.NamespaceURI;
			string elementLocalName = xmlReader.LocalName;

			PushElement(RdfXmlState.InProperty);
			top.m_name = elementNamespaceURI + elementLocalName;
			top.m_id = DiscoverStatementID();

			string parseType = null;

			// if the element is in the rdf namespace, try and find an unqualified parseType attribute.
			if (RdfSyntax.URI.Equals(elementNamespaceURI))
			{
				parseType = xmlReader.GetAttribute(RdfSyntax.ParseType);
			}

			// if the element is in some other namespace
			// or we didn't find an unqualified parseType attribute,
			// then look for a qualfied one.
			if (parseType == null)
			{
				parseType = xmlReader.GetAttribute(RdfSyntax.ParseType, RdfSyntax.URI);
			}

			if (RdfSyntax.ParseTypeResource.Equals(parseType)) // parseType="Resource"
			{
				top.m_state = RdfXmlState.InParseTypeResourceProperty;

				PushElement(RdfXmlState.InDescription);

				top.m_subjectType = RdfSubjectType.Anonymous;
				top.m_name = GenerateAnonymousURI();
			}

			else if (parseType != null) // assume parseType="Literal"
			{
				// DER - Parse the inner XML as the literal value.
				// Also consumes the closing tag
				String innerXML = xmlReader.ReadInnerXml();

				// Create the statement
				EnqueueStatement(
					top.m_parent.m_subjectType,
					top.m_parent.m_name,
					top.m_name,
					top.m_members,
					RdfObjectType.XML,
					innerXML,
					top.m_id);

				// Consume the InProperty node on the state stack
				PopElement();

			}
			else if (xmlReader.IsEmptyElement)
			{
				RdfSubjectType resourceType;
				string resource;

				DiscoverResourceURI(out resourceType, out resource);
#if DEBUG2
				Trace.WriteLine(resourceType+":   "+resource);
#endif
				if (MoveToFirstAttribute() || (resourceType == RdfSubjectType.URI)) 
				{

					PushElement(RdfXmlState.InResourceProperty);

					top.m_subjectType = resourceType;
					top.m_name = resource;

					HandlePropertyAttributes();

					EnqueueStatement(
						top.m_parent.m_parent.m_subjectType,
						top.m_parent.m_parent.m_name,
						top.m_parent.m_name,
						0,
						RdfObjectType.URI,
						resource,
						top.m_parent.m_id);

					PopElement();
					PopElement();
				} 
				else 
				{
					// Empty literal
					EnqueueStatement(
						top.m_parent.m_subjectType,
						top.m_parent.m_name,
						top.m_name,
						0,
						RdfObjectType.Literal,
						"",
						top.m_id);
					PopElement();
				}
						
			}
			else 
			{
				// This could be an expanded empty element (resource), or an element with a value.
				// If it has a resource or Property attributes then we have a resource
				RdfSubjectType resourceType;
				string resource;

				DiscoverResourceURI(out resourceType, out resource);
				// This was returning an empty anon node if there was a statement ID (and the object was a value)
				// Though I may get it trouble here if there is an empty anonymous node ??
				if (/*MoveToFirstAttribute() ||*/ (resourceType == RdfSubjectType.URI)) {
					PushElement(RdfXmlState.InResourceProperty);
					top.m_subjectType = resourceType;
					top.m_name = resource;
					HandlePropertyAttributes();
					EnqueueStatement(
						top.m_parent.m_parent.m_subjectType,
						top.m_parent.m_parent.m_name,
						top.m_parent.m_name,
						0,
						RdfObjectType.URI,
						resource,
						top.m_parent.m_id);					
					PopElement();

					// Set a state so we don't parse a value now.
					top.m_state = RdfXmlState.InResourceProperty;
				} 
				else 
				{
					// Continue: go parse the value
				}

			}
		}

		private void DiscoverResourceURI(out RdfSubjectType resourceType, out string resource)
		{
			resourceType = RdfSubjectType.URI;

			// should we only check for the unqualified resource if the element is
			// in the rdf namespace?
			resource = xmlReader.GetAttribute(RdfSyntax.Resource);

			if (resource == null)
			{
				resource = xmlReader.GetAttribute(RdfSyntax.Resource, RdfSyntax.URI);

				if (resource == null)
				{
					resourceType = RdfSubjectType.Anonymous;
					resource = GenerateAnonymousURI();
				}
			}
		}

		private string DiscoverStatementID()
		{
			string id = xmlReader.GetAttribute(RdfSyntax.ID);

			if (id == null)
			{
				id = xmlReader.GetAttribute(RdfSyntax.ID, RdfSyntax.URI);
				//Console.WriteLine("discover id = "+id);
			}
			return id;
		}

		private void InDescription_EndElement()
		{
			PopElement();
		}

		private void InProperty_Element()
		{
			top.m_state = RdfXmlState.InResourceProperty;

			PushElement(RdfXmlState.InDescription);

			bool isEmptyElement = xmlReader.IsEmptyElement;

			DiscoverDescriptionURI();

			HandlePropertyAttributes();

			EnqueueStatement(
				top.m_parent.m_parent.m_subjectType,
				top.m_parent.m_parent.m_name,
				top.m_parent.m_name,
				0,
				RdfObjectType.URI,
				top.m_name,
				top.m_parent.m_id);

			if (isEmptyElement)
			{
				PopElement();
			}
		}

		private void InProperty_Text()
		{
			top.Append(xmlReader.Value);
		}

		private void InProperty_EndElement()
		{
			// Changed this so if value == null, we use "" instead.
			//if (top.m_value != null)
			//{
				EnqueueStatement(
					top.m_parent.m_subjectType,
					top.m_parent.m_name,
					top.m_name,
					0,
					RdfObjectType.Literal,
					//top.m_value.ToString()
					top.m_value != null ? top.m_value.ToString() : "",
					top.m_id);
			
			//}
			

			PopElement();
		}

		private void InResourceProperty_EndElement()
		{
			PopElement();
		}

		private void InParseTypeResourceProperty_Element()
		{
			InDescription_Element();
		}

		private void InParseTypeResourceProperty_EndElement()
		{

			PopElement();
			PopElement();
		}

		private void InParseTypeLiteralProperty_EndElement()
		{
			String l = xmlReader.LocalName;	
			PopElement();
			PopElement();
		}

		enum RdfXmlState
		{
			OutsideRDF,
			InRDF,
			InDescription,
			InProperty,
			InLiteralProperty,
			InResourceProperty,
			InParseTypeLiteralProperty,
			InParseTypeResourceProperty
		}

		/// <summary>
		///    This is private.
		/// </summary>
		class RdfXmlElement
		{
			internal RdfXmlElement m_parent;
			internal RdfXmlState m_state;
			internal RdfSubjectType m_subjectType;
			internal string m_name;
			internal StringBuilder m_value;
			internal int m_members;
			internal string m_resource;
			internal string m_id;

			internal void Append(string value)
			{
				if (m_value == null)
				{
					m_value = new StringBuilder();
				}

				m_value.Append(value);
			}
		}

		public override int AttributeCount
		{
			get
			{
				return xmlReader.AttributeCount;
			}
		}

		public override string BaseURI
		{
			get
			{
				return xmlReader.BaseURI;
			}
		}

		public override int Depth
		{
			get
			{
				return xmlReader.Depth;
			}
		}

		public override bool EOF
		{
			get
			{
				return xmlReader.EOF;
			}
		}

		public override bool HasAttributes
		{
			get
			{
				return xmlReader.HasAttributes;
			}
		}

		public override bool HasValue
		{
			get
			{
				return xmlReader.HasValue;
			}
		}

		public override bool IsDefault
		{
			get
			{
				return xmlReader.IsDefault;
			}
		}

		public override bool IsEmptyElement
		{
			get
			{
				return xmlReader.IsEmptyElement;
			}
		}

		public override string this[int i]
		{
			get
			{
				return xmlReader[i];
			}
		}

		public override string this[string name]
		{
			get
			{
				return xmlReader[name];
			}
		}

		public override string this[string name, string namespaceURI]
		{
			get
			{
				return xmlReader[name, namespaceURI];
			}
		}

		public override string LocalName
		{
			get
			{
				return xmlReader.LocalName;
			}
		}

		public override string Name
		{
			get
			{
				return xmlReader.Name;
			}
		}

		public override string NamespaceURI
		{
			get
			{
				return xmlReader.NamespaceURI;
			}
		}

		public override XmlNameTable NameTable
		{
			get
			{
				return xmlReader.NameTable;
			}
		}

		public override XmlNodeType NodeType
		{
			get
			{
				return xmlReader.NodeType;
			}
		}

		public override string Prefix
		{
			get
			{
				return xmlReader.Prefix;
			}
		}

		public override char QuoteChar
		{
			get
			{
				return xmlReader.QuoteChar;
			}
		}

		public override ReadState ReadState
		{
			get
			{
				return xmlReader.ReadState;
			}
		}

		public override string Value
		{
			get
			{
				return xmlReader.Value;
			}
		}

		public override string XmlLang
		{
			get
			{
				string result = "";

				if (IsStatement)
				{
					CheckCurrent();
					result = current.XmlLang;
				}
				else
				{
					result = xmlReader.XmlLang;
				}

				return result;
			}
		}

		public override XmlSpace XmlSpace
		{
			get
			{
				return xmlReader.XmlSpace;
			}
		}

		public override void Close()
		{
			if (elements.Count > 0)
			{
				throw new Exception("elements stack not empty");
			}

			xmlReader.Close();
		}

		public override string GetAttribute(int i)
		{
			return xmlReader.GetAttribute(i);
		}

		public override string GetAttribute(string name)
		{
			return xmlReader.GetAttribute(name);
		}

		public override string GetAttribute(string name, string namespaceURI)
		{
			return xmlReader.GetAttribute(name, namespaceURI);
		}

		public override bool IsStartElement()
		{
			return xmlReader.IsStartElement();
		}

		public override bool IsStartElement(string name)
		{
			return xmlReader.IsStartElement(name);
		}

		public override bool IsStartElement(string name, string uri)
		{
			return xmlReader.IsStartElement(name, uri);
		}

		public override string LookupNamespace(string prefix)
		{
			return xmlReader.LookupNamespace(prefix);
		}

		public override void MoveToAttribute(int i)
		{
			xmlReader.MoveToAttribute(i);
		}

		public override bool MoveToAttribute(string name)
		{
			return xmlReader.MoveToAttribute(name);
		}

		public override bool MoveToAttribute(string name, string uri)
		{
			return xmlReader.MoveToAttribute(name, uri);
		}

		public override XmlNodeType MoveToContent()
		{
			return xmlReader.MoveToContent();
		}

		public override bool MoveToElement()
		{
			return xmlReader.MoveToElement();
		}

		public override bool MoveToFirstAttribute()
		{
			return xmlReader.MoveToFirstAttribute();
		}

		public override bool MoveToNextAttribute()
		{
			return xmlReader.MoveToNextAttribute();
		}

		// Read here?

		public override bool ReadAttributeValue()
		{
			return xmlReader.ReadAttributeValue();
		}

		public override string ReadInnerXml()
		{
			return xmlReader.ReadInnerXml();
		}

		public override string ReadOuterXml()
		{
			return xmlReader.ReadOuterXml();
		}

		public override string ReadString()
		{
			return xmlReader.ReadString();
		}

		public override void ResolveEntity()
		{
			xmlReader.ResolveEntity();
		}
	}
} // namespace XMarksTheDOT.RDF
