using System;
using BBN.RDF;

namespace BBN.RDF.Common
{
	public interface ResourceI
	{
		Resource Port(Model m);
		Resource GetEmbeddedResource();
	}
	/// <summary>
	/// Implementation for the Resource interface
	/// <author>Troy Self</author>
	/// </summary>
	public class ResourceImpl : Resource, ResourceI
	{
		protected string nameSpace;
		protected string localName;
		protected string uri;
		protected Model model;
		protected AnonId id;

		public ResourceImpl()
		{
			AssignId();
		}
		public ResourceImpl(string uri)
		{
			this.uri = uri;
			if(uri == null) this.uri = "";
			if (this.uri.Length < 1)
				AssignId();
			else SplitNameSpace();
		}
		public ResourceImpl(string nameSpace, string localName)
		{
			this.nameSpace = nameSpace;
			this.localName = localName;
			uri = nameSpace + localName;
		}
		public ResourceImpl(Model m)
		{
			model = m;
			AssignId();
		}
		public ResourceImpl(AnonId id)
		{
			this.id = id;
		}
		public ResourceImpl(AnonId id, Model m)
		{
			model = m;
			this.id = id;
		}
		public ResourceImpl(string uri, Model m)
		{
			this.uri = uri;
			this.model = m;
			if(this.uri == null || this.uri.Length < 1)
				AssignId();
			else SplitNameSpace();
		}
		public ResourceImpl(Resource r, Model m)
		{
			uri = r.GetURI();
			id = r.GetId();
			nameSpace = r.GetNameSpace();
			localName = r.GetLocalName();
			model = m;
		}
		public ResourceImpl(string nameSpace, string localName, Model m)
		{
			this.nameSpace = nameSpace;
			this.localName = localName;
			uri = nameSpace + localName;
			model = m;
		}
		protected void AssignId()
		{
			id = new AnonId();
		}
//		protected void SplitNameSpace()
//		{
//			int index = uri.LastIndexOf('#');
//			if(index < 0) // no # in uri
//			{
//				nameSpace = uri;
//				localName = "";
//			}
//			else
//			{
//				nameSpace = uri.Substring(0, index+1);
//				localName = uri.Substring(index+1);
//			}
//		}
		protected void SplitNameSpace() 
		{
			int splitPoint = Util.SplitNamespace(uri);
			if (splitPoint == 0) 
			{
				nameSpace = uri;
				localName = "";
			} 
			else 
			{
				nameSpace = uri.Substring(0, splitPoint);
				localName = uri.Substring(splitPoint);
			}
		}

		public AnonId GetId()
		{
			return id;
		}
		public string GetURI()
		{
			return uri;
		}
		public string GetNameSpace()
		{
			return nameSpace;
		}
		public string GetLocalName()
		{
			return localName;
		}
		public override string ToString()
		{
			if(uri == null)
				return "anon:" + id;
			else
				return GetURI();
		}
		public override bool Equals(object o)
		{
			if(o is Resource)
				return (uri.Equals(((Resource)o).GetURI()));
			else if (o is Statement)
				return this.Equals(((Statement)o).GetReificationId());
			else return false;
		}
		public override int GetHashCode()
		{
			if(IsAnon())
				return id.GetHashCode();
			else return uri.GetHashCode();
		}
		public bool IsAnon()
		{
			return (uri == null);
		}
		public Statement GetProperty(Property p)
		{
			if(model != null)
				return model.GetProperty(this, p);
			else throw new System.Exception("Resource not related to model");
		}
		public virtual StmtEnumerator ListProperties()
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.ListStatements(new SelectorImpl(this, null, (RDFNode)null));
		}
		public virtual StmtEnumerator ListProperties(Property p)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.ListStatements(new SelectorImpl(this, p, (RDFNode)null));		
		}
		public virtual StmtEnumerator ListIncomingProperties()
		{
			if (model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.ListStatements(new SelectorImpl(null, null, (RDFNode)this));
		}
		public virtual StmtEnumerator ListIncomingProperties(Property p)
		{
			if (model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.ListStatements(new SelectorImpl(null, p, (RDFNode)this));
		}
		public Resource AddProperty(Property p, bool o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, long o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, char o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, double o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, float o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, object o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, RDFNode o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, string o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o);
			return this;
		}
		public Resource AddProperty(Property p, string o, string l)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			model.Add(this, p, o, l);
			return this;
		}
		public bool HasProperty(Property p)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p);
		}
		public bool HasProperty(Property p, bool o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}

		public bool HasProperty(Property p, long o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public bool HasProperty(Property p, char o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public bool HasProperty(Property p, float o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public bool HasProperty(Property p, double o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public bool HasProperty(Property p, string o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public bool HasProperty(Property p, string o, string l)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o, l);
		}
		public bool HasProperty(Property p, object o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}

		public bool HasProperty(Property p, RDFNode o)
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			return model.Contains(this, p, o);
		}
		public virtual Resource RemoveProperties()
		{
			if(model == null)
			{
				throw new System.Exception("Resource not related to model");
			}
			StmtEnumerator iter = null;
			iter = model.ListStatements(new SelectorImpl(this, null, (RDFNode)null));
			while(iter.MoveNext())
			{
				model.Remove(iter.Current);
				//iter.Remove();
			}
			return this;
		}
		public Resource Begin()
		{
			if(model == null)
				throw new System.Exception("Resource not related to model");
			model.Begin();
			return this;
		}
		public Resource Abort()
		{
			if(model == null)
				throw new System.Exception("Resource not related to model");
			model.Abort();
			return this;
		}

		public Resource Commit()
		{
			if(model == null)
				throw new System.Exception("Resource not related to model");
			model.Commit();
			return this;
		}
		public Model GetModel()
		{
			return model;
		}
		public Resource GetEmbeddedResource()
		{
			return this;
		}
		public virtual Resource Port(Model m)
		{
			if(model == m)
				return this;
			else
				return new ResourceImpl(this, m);
		}
	}
#if DYNAMIC
	public class DynamicResourceImpl : ResourceImpl
	{
		public DynamicResourceImpl() : base()
		{
		}

		public DynamicResourceImpl(AnonId id) : base(id)
		{
		}
            
		public DynamicResourceImpl(AnonId id, Model m) : base(id, m)
		{
		}
            
		public DynamicResourceImpl(Model m) : base(m)
		{
		}
            
		public DynamicResourceImpl(Resource r, Model m) : base(r, m)
		{
		}
            
		public DynamicResourceImpl(string uri) : base(uri)
		{
		}
            
		public DynamicResourceImpl(string uri, Model m) : base(uri, m)
		{
		}
            
		public DynamicResourceImpl(string nameSpace, string localName) : base(nameSpace, localName)
		{
		}
            
		public DynamicResourceImpl(string nameSpace, string localName, Model m) : base(nameSpace, localName, m)
		{
		}

		/**
		 * load the page containing this URI if it hasn't been loaded already
		 */
		protected void MaybeLoad()
		{
			if (! ((DynamicModel)GetModel()).DynamicLoad())
				return;

			string uri = GetURI();
			int hash = uri.LastIndexOf('#');
			if (hash != (-1))
				uri = uri.Substring(0, hash);
			DynamicModel model = (DynamicModel) GetModel(); 
			if (! model.IsLoaded(uri))
			{
				model.Read(uri);
			}
		}

		public override StmtEnumerator ListProperties()
		{
			MaybeLoad();
			return base.ListProperties();
		}

		public override StmtEnumerator ListProperties(Property p)
		{
			MaybeLoad();
			return base.ListProperties(p);
		}

		public override Resource Port(Model m)
		{
			if (GetModel() == m)
				return this;
			else return new DynamicResourceImpl(GetURI(), m);
		}

	}
#endif
}
