using System;
using BBN.RDF;

namespace BBN.RDF.Common
{
	/// <summary>
	/// Summary description for StatementImpl.
	/// </summary>
	public class StatementImpl : ResourceImpl, Statement
	{
		protected Resource subject;
		protected Property predicate;
		protected RDFNode Object;
		protected Source source;
		protected Resource reificationID;

		public StatementImpl(Resource s, Property p, RDFNode o) : base()
		{
			if (s is Statement)
				subject = ((Statement)s).GetReificationId();
			else
				subject = s;
			predicate = p;
			if (s is Statement)
				Object = ((Statement)o).GetReificationId();
			else
				Object = o;
		}
		public StatementImpl(Resource s, Property p, RDFNode o, Model m)
			: base(m)
		{
			if (s is Statement) 
				subject = ((Statement)s).GetReificationId();
			else 
				subject = ((ModelI)m).Convert(s);
			predicate = ((ModelI)m).Convert(p);
			if (o is Statement)
				Object = ((Statement)o).GetReificationId();
			else
				Object = ((ModelI)m).Convert(o);
		}

		/// <summary>
		///  Create a new statement
		/// </summary>
		/// <param name="s">The subject of the statement</param>
		/// <param name="p">The property of the statement</param>
		/// <param name="o">The object of the statement</param>
		/// <param name="id">An anonymousID to assign to the statement</param>
		/// <param name="source">The source uri of the statement</param>
		/// <param name="m">The model for which the statement is being created</param>
		public StatementImpl(Resource s, Property p, RDFNode o, AnonId id, string source, Model m)
			: base(id, m)
		{
			if (s is Statement) 
				subject = ((Statement)s).GetReificationId();
			else 
				subject = ((ModelI)m).Convert(s);
			predicate = ((ModelI)m).Convert(p);
			if (o is Statement)
				Object = ((Statement)o).GetReificationId();
			else
				Object = ((ModelI)m).Convert(o);
			this.source = m.GetSource(source);
		}

		/// <summary>
		/// Create a new Statement
		/// </summary>
		/// <param name="s">The subject of the statement</param>
		/// <param name="p">The property of the statement</param>
		/// <param name="o">The object of the statement</param>
		/// <param name="uri">An rdf:ID to assign to the statement</param>
		/// <param name="source">The source uri of the statement</param>
		/// <param name="m">The model for which the statement is being created</param>
		public StatementImpl(Resource s, Property p, RDFNode o, string uri, string source, Model m)
			: base(uri, m)
		{
			if (s is Statement) 
				subject = ((Statement)s).GetReificationId();
			else 
				subject = ((ModelI)m).Convert(s);
			predicate = ((ModelI)m).Convert(p);
			if (o is Statement)
				Object = ((Statement)o).GetReificationId();
			else
				Object = ((ModelI)m).Convert(o);
			this.source = m.GetSource(source);
		}
		public override int GetHashCode()
		{
			return subject.GetHashCode() ^ predicate.GetHashCode() ^ Object.GetHashCode();
		}
		public Resource GetSubject()
		{
			return subject;
		}
		public Property GetPredicate()
		{
			return predicate;
		}
		public RDFNode GetObject()
		{
			return Object;
		}
		/// <summary>
		/// Get the source uri of this statement.
		/// </summary>
		/// <returns>Returns the source object associated with the statement</returns>
		public Source GetSource()
		{
			return source;
		}
		public new Statement GetProperty(Property p)
		{
			if(Object is Resource)
				return ((Resource)Object).GetProperty(p);
			else
				throw new System.Exception("Object not a resource");
		}
		public Statement GetStatementProperty(Property p)
		{
			return base.GetProperty(p);
		}
		public Resource GetResource()
		{
			if(Object is Resource)
				return (Resource)Object;
			else
				throw new System.Exception("Object not a Resource");
		}
		public Literal GetLiteral()
		{
			if(Object is Literal)
				return (Literal)Object;
			else
				throw new System.Exception("Object not a Literal");
		}
		public bool GetBoolean()
		{
			if(Object is Literal)
				return ((Literal)Object).GetBoolean();
			else
				throw new System.Exception("Object not a Literal");
		}
		public byte GetByte()
		{
			if(Object is Literal)
				return ((Literal)Object).GetByte();
			else
				throw new System.Exception("Object not a Literal");
		}
		public short GetShort()
		{
			if(Object is Literal)
				return ((Literal)Object).GetShort();
			else
				throw new System.Exception("Object not a Literal");
		}
		public int GetInteger()
		{
			if(Object is Literal)
				return ((Literal)Object).GetInteger();
			else
				throw new System.Exception("Object not a Literal");
		}
		public long GetLong()
		{
			if(Object is Literal)
				return ((Literal)Object).GetLong();
			else
				throw new System.Exception("Object not a Literal");
		}
		public char GetChar()
		{
			if(Object is Literal)
				return ((Literal)Object).GetChar();
			else
				throw new System.Exception("Object not a Literal");
		}
		public float GetFloat()
		{
			if(Object is Literal)
				return ((Literal)Object).GetFloat();
			else
				throw new System.Exception("Object not a Literal");
		}
		public double GetDouble()
		{
			if(Object is Literal)
				return ((Literal)Object).GetDouble();
			else
				throw new System.Exception("Object not a Literal");
		}
		public string GetString()
		{
			if(Object is Literal)
				return ((Literal)Object).ToString();
			else
				throw new System.Exception("Object not a Literal");
		}
		public Resource GetResource(ResourceF f)
		{
			if(Object is Resource)
			{
				try{return f.CreateResource(((ResourceI)Object).GetEmbeddedResource());}
				catch(System.Exception e)
				{
					throw e;
				}
			}
			else
				throw new System.Exception("Object not a Resource");
		}
		public object GetObject(ObjectF f)
		{
			try{return f.CreateObject(((Literal)Object).ToString());}
			catch(System.Exception e)
			{
				throw e;
			}
		}
		public string GetLanguage()
		{
			if(Object is Literal)
				return ((Literal)Object).GetLanguage();
			else
				throw new System.Exception("Object is wrong type");
		}
		public bool IsWellFormed()
		{
			if(Object is Literal)
				return ((Literal)Object).IsWellFormed();
			else
				throw new System.Exception("Object is wrong type");
		}
		public Statement Set(bool o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(long o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(char o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(float o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(double o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(string o)
		{
			Replace(new LiteralImpl(o));
			return this;
		}
		public Statement Set(string o, bool wellFormed)
		{
			Replace(new LiteralImpl(o, wellFormed));
			return this;
		}
		public Statement Set(string o, string l)
		{
			Replace(new LiteralImpl(o, l));
			return this;
		}
		public Statement Set(string o, string l, bool wellFormed)
		{
			Replace(new LiteralImpl(o, l, wellFormed));
			return this;
		}
		public Statement Set(RDFNode o)
		{
			Replace(o);
			return this;
		}
		public Statement Set(object o)
		{
			if(o is RDFNode)
				Replace((RDFNode)o);
			else
				Replace(new LiteralImpl(o.ToString()));
			return this;
		}
		protected void Replace(RDFNode n)
		{
			if(model != null)
			{
				model.Remove(this);
				model.Add(subject, predicate, n);
				Object = n;
			}
			else
				throw new System.Exception("Not related to model");
		}
		public override string ToString()
		{
			string result = "[" + subject.ToString() + ", " + predicate.ToString() + ", ";
			if(Object is Resource)
				result = result + "Resource<" + ((Resource)Object).ToString() + ">";
			else
				result = result + "Literal<" + ((Literal)Object).ToString() + ">";
			return result + "]";
		}
		public override bool Equals(object o)
		{
			if (o is Statement)
				return(subject.Equals(((Statement)o).GetSubject())
					&& (predicate.Equals(((Statement)o).GetPredicate()))
					&& (Object.Equals(((Statement)o).GetObject())) );
			else if (o is Resource)
				//return (uri.Equals(((Resource)o).GetURI()));
				return (this.GetReificationId().Equals(o));
			else 
				return false;
		}
		public override Resource RemoveProperties()
		{
			if(model != null)
			{
				base.RemoveProperties();
				return this;
			}
			else
				throw new System.Exception("Not related to model");
		}
		public Statement Remove()
		{
			if(model != null)
			{
				model.Remove(this);
				return this;
			}
			else
				throw new System.Exception("Not related to model");
		}
		public bool IsReified()
		{
			if(model != null)
				return model.IsReified(this);
			else
				throw new System.Exception("Not related to model");
		}
		public Resource GetReificationId()
		{
			if (reificationID != null)
			{
				return reificationID;
			}
			else if (model == null)
			{
				if (uri != null)
					return reificationID = new ResourceImpl(uri);
				else if (source == null)
					return reificationID = new ResourceImpl("#"+id.ToString());
				else 
					return reificationID = new ResourceImpl(source+"#"+id.ToString());
			}
			else
			{
				if (uri != null)
					return reificationID = model.GetResource(uri);
				else if (source == null)
					return reificationID = model.GetResource("#"+id.ToString());
				else
					return reificationID = model.GetResource(source+"#"+id.ToString());
			}
		}
		public new Resource Port(Model m)
		{
			if(model == m)
				return this;
			else
			{
				StatementImpl s;
				if (uri != null)
					s = new StatementImpl(subject, predicate, Object, uri, source.URI, m);
				else
					s = new StatementImpl(subject, predicate, Object, id, source.URI, m);
				return s;
			}
		}
	}
}
