using System;
using System.Collections;
using BBN.RDF;
using System.Runtime.CompilerServices; // for synchronizing methods

///
/// <summary
/// Contains the interfaces and classes necessary for storing an RDF model in memory
/// </summary>
/// <author>Troy Self</author>
namespace BBN.RDF.Common
{
	/// <summary>
	/// Interface for storing a model of statements
	/// </summary>
	public interface Store : IEnumerable
	{
		Resource CreateResource(Model m);
		Resource CreateResource(string uri, Model m);
		Property CreateProperty(string uri, Model m);
		Property CreateProperty(string nameSpace, string localPart, Model m);
		Statement CreateStatement(Resource s, Property p, RDFNode o, Model m);
		//Bag CreateBag(Model m);
		//Bag CreateBag(string uri, Model m);
		//Alt CreateAlt(Model m);
		//Alt CreateAlt(string uri, Model m);
		//Seq CreateSeq(Model m);
		//Seq CreateSeq(string uri, Model m);
		Resource GetResource(string uri, Model m);
		Property GetProperty(string uri, Model m);
		Property GetProperty(string nameSpace, string localPart, Model m);
		Property GetProperty(Property p);
		Statement GetStatement(Resource s, Property p, RDFNode o, Model m);
		IList GetStatementsBySubject(Resource s);
		IList GetStatementsByPredicate(Property p);
		IList GetStatementsByObject(RDFNode o);
		//Bag GetBag(string uri, Model m);
		//Bag GetBag(Resource r, Model m);
		//Alt GetAlt(string uri, Model m);
		//Alt GetAlt(Resource r, Model m);
		//Seq GetSeq(string uri, Model m);
		//Seq GetSeq(Resource r, Model m);
		void Add(Statement s);
		void Remove(Statement s);
		int Count{get;}
		bool Contains(Statement s);
		bool Contains(Resource s, Property p);
		IEnumerator List();
		IEnumerator List(Resource s, Property p, RDFNode o);
		IEnumerator ListBySubject(Resource s);
		IEnumerator ListByPredicate(Property p);
		IEnumerator ListByObject(RDFNode o);
		ResEnumerator ListSubjects();
		NsEnumerator ListNameSpaces();
		new IEnumerator GetEnumerator();
		void Close();
		
		/// <summary>
		/// Get the source object for a uri.
		/// </summary>
		Source GetSource(string uri);
		/// <summary>
		/// Enumerate over all the statements from a source.
		/// </summary>
		IEnumerator ListBySource(Source s);
		/// <summary>
		/// List the sources in the store.
		/// </summary>
		IEnumerator ListSources();
	}
	/// <summary>
	/// Implementation of the Store interface
	/// </summary>
	public abstract class StoreCom : Store
	{
		public StoreCom()
		{
		}
		// Abstract implementations for derived interface methods.
		// These should be implemented by the inheriting class
		public abstract void Add(Statement s);
		public abstract bool Contains(Resource r, Property p);
		public abstract bool Contains(Statement s);
		public abstract IEnumerator List();
		public abstract IEnumerator List(Resource r, Property p, RDFNode o);
		public abstract IEnumerator ListByObject(RDFNode o);
		public abstract IEnumerator ListByPredicate(Property p);
		public abstract IEnumerator ListBySubject(Resource r);
		/// <summary>
		/// List all statements associated with a source.
		/// </summary>
		/// <param name="s">The source of interest</param>
		/// <returns>An IEnumerator over all the statements associated with source</returns>
		public abstract IEnumerator ListBySource(Source s);
		public abstract NsEnumerator ListNameSpaces();
		public abstract ResEnumerator ListSubjects();
		/// <summary>
		/// List all sources in the store.
		/// </summary>
		/// <returns>An IEnumerator over all the sources in the store.</returns>
		public abstract IEnumerator ListSources();
		public abstract IEnumerator GetEnumerator();
		public abstract void Remove(Statement s);
		public abstract int Count{get;}
		public abstract IList GetStatementsBySubject(Resource s);
		public abstract IList GetStatementsByPredicate(Property p);
		public abstract IList GetStatementsByObject(RDFNode o);



		public virtual Resource CreateResource(Model m)
		{
			return new ResourceImpl(m);
		}
		public virtual Resource CreateResource(string uri, Model m)
		{
			return new ResourceImpl(uri, m);
		}
		public virtual Property CreateProperty(string uri, Model m)
		{
			return new PropertyImpl(uri, m);
		}
		public virtual Property CreateProperty(string nameSpace, string localPart, Model m)
		{
			return new PropertyImpl(nameSpace, localPart, m);
		}
		public virtual Statement CreateStatement(Resource s, Property p, RDFNode o, Model m)
		{
			return new StatementImpl(s, p, o, m);
		}
		/// <summary>
		/// Create a new source object for a uri.
		/// </summary>
		/// <param name="uri">The uri for the source</param>
		/// <returns>A new source object with the given uri</returns>
		protected virtual Source CreateSource(string uri)
		{
			return new SourceImpl(uri);
		}
		/*public virtual Bag CreateBag(Model m)
		{
			return new BagImpl(m);
		}*/
		/*public virtual Bag CreateBag(string uri, Model m)
		{
			return new BagImpl(uri, m);
		}*/
		/*public virtual Alt CreateAlt(Model m)
		{
			return new AltImpl(m);
		}*/
		/*public virtual Alt CreateAlt(string uri, Model m)
		{
			return new AltImpl(uri, m);
		}*/
		/*public virtual Seq CreateSeq(Model m)
		{
			return new SeqImpl(m);
		}*/
		/*public virtual Seq CreateSeq(string uri, Model m)
		{
			return new SeqImpl(uri, m);
		}*/
		public virtual Resource GetResource(string uri, Model m)
		{
			return new ResourceImpl(uri, m);
		}
		public virtual Property GetProperty(string uri, Model m)
		{
			return GetProperty(new PropertyImpl(uri, m));
		}
		public virtual Property GetProperty(string nameSpace, string localPart, Model m)
		{
			return GetProperty(new PropertyImpl(nameSpace, localPart, m));
		}
		public virtual Property GetProperty(Property p)
		{
			return p;
		}
		/// <summary>
		/// Get the source object for the uri from the store.
		/// </summary>
		/// <param name="uri">The uri of the source</param>
		/// <returns>A possibly new source object for the uri.</returns>
		public virtual Source GetSource(string uri)
		{
			return CreateSource(uri);
		}
		public virtual Statement GetStatement(Resource s, Property p, RDFNode o, Model m)
		{
			return new StatementImpl(s, p, o, m);
		}
		/*public virtual Bag GetBag(string uri, Model m)
		{
			return GetBag(GetResource(uri, m), m);
		}*/
		/*public virtual Bag GetBag(Resource r, Model m)
		{
			if(typeof(Resource).Equals(r.GetType()))
				return (Bag)r;
			return new BagImpl(r, m);
		}*/
		/*public virtual Alt GetAlt(string uri, Model m)
		{
			return GetAlt(GetResource(uri, m), m);
		}*/
		/*public virtual Alt GetAlt(Resource r, Model m)
		{
			if(typeof(Alt).Equals(r.GetType()))
				return (Alt)r;
			return new AltImpl(r, m);
		}*/
		/*public virtual Seq GetSeq(string uri, Model m)
		{
			return GetSeq(GetResource(uri, m), m);
		}*/
		/*public virtual Seq GetSeq(Resource r, Model m)
		{
			if(typeof(Seq).Equals(r.GetType()))
				return (Seq)r;
			return new SeqImpl(r, m);
		}*/
		public virtual void Close()
		{}
	}
}	
