using System;
using BBN.RDF;
using System.Collections;

namespace BBN.DAML
{
	/// <summary>
	/// Provides a registry that can map RDF values encoded using XML Schema type definitions
	/// into Java objects or scalars.  A basic set of mappings is provided, but other type
	/// mappings can be added by the user.
	/// </summary>
	public class XMLDatatypeRegistry
	{
		// Constants
		//////////////////////////////////

		/** The URI for XML schema datatypes (xsd) namespace */
		public static readonly string XSD_NAMESPACE_URI = "http://www.w3.org/2000/10/XMLSchema#";

		/** Resource to denote the type of an xsd integer */
		public static readonly Resource XSD_INTEGER = new BBN.RDF.Common.ResourceImpl( XSD_NAMESPACE_URI + "integer" );



		// Static variables
		//////////////////////////////////



		// Instance variables
		//////////////////////////////////

		/** Hashtable for type URI to translator object */
		protected Hashtable m_registry = new Hashtable();


		// Constructors
		//////////////////////////////////

		/**
		 * Constructor for the registry, which will register a basic set of mappings.
		 */
		public XMLDatatypeRegistry() 
		{
			RegisterDefaultTypes();
		}


		// External signature methods
		//////////////////////////////////

		/**
		 * Answer the datatype translator for a a resource whose <code>rdf:type</code>
		 * is the given URI.
		 *
		 * @param uri The uri of the type
		 * @return the corresponding translator, if one is registered, or null otherwise.
		 */
		public DatatypeTranslator GetTranslator( string uri ) 
		{
			return (DatatypeTranslator) m_registry[uri];
		}


		/**
		 * Register a translator for the type whose uri is as given.
		 *
		 * @param uri The uri of the type to be registered
		 * @param translator The translator that will map objects of the given type
		 *                   between Java and XML/RDF.
		 */
		public void RegisterTranslator( string uri, DatatypeTranslator translator ) 
		{
			m_registry[uri]= translator;
		}


		/**
		 * Answer an iteration over the types of URI's that are registered.
		 *
		 * @return an iteration, each element of which is a string representing a URI.
		 */
		public IEnumerator GetRegisteredTypes() 
		{
			return m_registry.Keys.GetEnumerator();
		}


		/**
		 * Answer true if the given type URI is registered in this registry
		 *
		 * @param typeURI The URI of the type to be tested, as a string.
		 * @return bool true if the type is registered.
		 */
		public bool IsRegisteredType( string typeURI ) 
		{
			return m_registry.ContainsKey( typeURI );
		}



		// Internal implementation methods
		//////////////////////////////////

		/**
		 * Add a basic (built-in) set of mappings between RDF/XML data and Java objects.
		 */
		protected void RegisterDefaultTypes() 
		{
			/* TBD
			// add a converter for integers
			m_registry[XSD_NAMESPACE_URI + "integer"] =
							new DatatypeTranslator() {
								// Get an Integer out of a node 
								public Object deserialize( RDFNode node ) {
									try {
										if (node is Literal) {
											// parse the string value into an integer
											return new Integer( ((Literal) node).GetString() );
										}
										else if (node is Resource) {
											// assume we have have resource whose rdf:value is the value of the node
											RDFNode val = ((Resource) node).GetProperty( RDF.value ).GetObject();
											return deserialize( val );
										}
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
									}

									return null;
								}

								// Convert an Integer to a node 
								public RDFNode serialize( Object value, DAMLModel model ) {
									try {
										return model.createLiteral( ((Integer) value).intValue() );
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
										return null;
									}
								}
							};

			// add a converter for strings
			m_registry[XSD_NAMESPACE_URI + "string"] =
							new DatatypeTranslator() {
								// Get an Integer out of a node 
								public Object deserialize( RDFNode node ) {
									try {
										if (node is Literal) {
											// parse the string value into an integer
											return ((Literal) node).GetString();
										}
										else if (node is Resource) {
											// assume we have have resource whose rdf:value is the value of the node
											RDFNode val = ((Resource) node).GetProperty( RDF.value ).GetObject();
											return deserialize( val );
										}
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
									}

									return null;
								}

								// Convert an Integer to a node 
								public RDFNode serialize( Object value, DAMLModel model ) {
									try {
										return model.createLiteral( (string) value );
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
										return null;
									}
								}
							};

			// add a converter for reals
			m_registry[XSD_NAMESPACE_URI + "real"] =
							new DatatypeTranslator() {
								// Get a Real out of a node 
								public Object deserialize( RDFNode node ) {
									try {
										if (node is Literal) {
											// parse the string value into an integer
											return new Float( ((Literal) node).GetString() );
										}
										else if (node is Resource) {
											// assume we have have resource whose rdf:value is the value of the node
											RDFNode val = ((Resource) node).GetProperty( RDF.value ).GetObject();
											return deserialize( val );
										}
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e);
									}

									return null;
								}

								// Convert an Integer to a node 
								public RDFNode serialize( Object value, DAMLModel model ) {
									try {
										return model.createLiteral( ((Float) value).floatValue() );
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
										return null;
									}
								}
							};

			// add a converter for decimals (we model them as long integers
			m_registry.put[XSD_NAMESPACE_URI + "decimal"] =
							new DatatypeTranslator() {
								// Get a Real out of a node 
								public Object deserialize( RDFNode node ) {
									try {
										if (node is Literal) {
											// parse the string value into an integer
											return new Long( ((Literal) node).GetString() );
										}
										else if (node is Resource) {
											// assume we have have resource whose rdf:value is the value of the node
											RDFNode val = ((Resource) node).GetProperty( RDF.value ).GetObject();
											return deserialize( val );
										}
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
									}

									return null;
								}

								// Convert an Integer to a node 
								public RDFNode serialize( Object value, DAMLModel model ) {
									try {
										return model.createLiteral( ((Long) value).longValue() );
									}
									catch (Exception e) {
										Console.Error.WriteLine( "RDF exception while converting datatype instance: " + e );
										return null;
									}
								}
							};
	*/
		}



		//==============================================================================
		// Inner class definitions
		//==============================================================================
	}
}
